The RFX (Rendered Effects System)
---------------------------------

Author: salsaman@xs4all.nl
Date: 23/7/2004
Version: 1.3

Changes
1.0 First version salsaman@xs4all.nl
1.1 Added input channels
1.2 Added "<author>" section
1.3 Added "<define>" section

Note:
The author grants you a non-exclusive, worldwide, royalty-free license to use the 
information disclosed in this Specification to make your software RFX-compliant; 
no other license, express or implied, is granted or intended hereby.



RFX is a system for generating plugins, e.g. rendered effects from 
scripts. It is fully extendible, and has features for:

- defining parameters
- sending layout hints to a GUI about how to present a parameter window
- allowing triggers when the parameter window is initialised, and/or when any 
     of the parameter values are altered
- allowing definition of code for pre/loop/post processing of blocks of frames

In the present RFX implementation for LiVES, pre/loop/post code and triggers 
must be written in LiVES-perl (language code 0xF0), which is based on Perl 
but with extra default variables and some subroutines to make frame (image) 
processing easier.

The following section describes a script file, and the rules which must be 
implemented in order to comply with the RFX script version 1.3




Script files
------------

Script files are the beginning point for all effects/tools/utilities in RFX.
From the script file, a plugin can be generated for a particular host 
application. The details of the plugin generation are left to the authors of 
the relevant host applications. RFX script files are abstracted in the sense 
that they are not tied to any particular widget set, or to any particular 
method of transferring data between host and plugin.

Here are the sections (some mandatory, some optional) which comprise an RFX 
script file. Each section in the script file is laid out as follows:

<keyword>
value1
value2
</keyword>

some keyword sections are optional, these are indicated below like:
<keyword> [optional]

some keyword sections take a single data value, some take multiple data values
Keywords which take multiple values are indicated thus:
<keyword> [list]

In the case of multiple values, each value should be terminated by a newline 
(\n) character.

Within values, the delimiter is '|'

Note:
As of RFX version 1.3, these default delimiters can be overwritten in the <define>
section. However, the default values are assumed below.


Here are the RFX sections:

<name>

<version> [optional]

<author> [optional]

<description>

<requires> [optional] [list]

<params> [optional] [list]

<param_window> [optional] [list]

<properties> [optional]

<language_code>

<pre> [optional] [list]

<loop> [optional] [list]

<post> [optional] [list]

<onchange> [optional] [list]

The contents of each section are described below.


<define>

This section contains the following entities:
field delimiter
line delimiter
RFX version

E.g.:

<define>
|
1.3
</define>

As noted, the following assumes the default field delimiter of | and the default 
line delimiter \n (newline).

As of the time of writing, LiVES will ignore the <define> section.



<name>

This section contains a single entry with the name of the effect. The name 
not contain spaces, newlines or the delimiter character (|).


e.g.

<name>
blank_frames
</name>




<version>

This is an optional section for a version string for the plugin/script.
Any format is acceptible, but it should not contain newlines or the 
delimiter character. Do not confuse this with the RFX version in the <define>
section.

e.g.

<version>
0.9.1.-pre2
</version>

<author>

This is an optional section for an "author" string for the plugin. 
Any format is acceptible, but it should not contain newlines or the 
delimiter character.

e.g.

<author>
somebody@somewhere.com
</author>



<description>

This section consists of a single entry with 4 fields:

menu text|action description|minimum frames|num_channels

e.g.

<description>
Edge Detect|Edge detecting|1|1
</description>

minimum frames is the minimum block length for frames that the plugin will 
accept. A value of 0 indicates a minimum of 1 frame (i.e. 0 is equivalent to 1)

num_channels indicates the number of input channels. (LiVES currently will only 
accept 0,1, or 2 input channels).


IMPORTANT NOTE
A value of -1 for minimum frames indicates that the plugin will not actually 
process frames. This is useful for plugins where only a parameter window is 
needed. In the case that minimum frames is -1, the host application should 
simply display the paramter window and allow the user to change the values, 
but no processing of frames should be performed. This is useful with trigger 
code for tools like calculators.




<requires> [optional] [list]

This section consists of a list of required binaries for the plugin to run. 
Each entry in the list should be separated by a newline.

e.g.

<requires>
convert
composite
</requires>

The host and/or plugin should check that the required binaries are present 
before attempting to process any frames.




<params> [optional] [list]

This section contains a list of parameters for the effect. Each entry in the 
list should be separated by a newline. Each entry in the list contains a 
variable number of fields depending on the parameter Type.

The types which are defined so far are:

num
bool
colRGB24

Each type is defined thus:

num: this is a numeric value. The 'num' type is immediately followed by a 
decimal number indicating the number of decimal places. An integer value is 
therefore declared as 'num0', while a numerical value with 6 decimal places of 
accuracy is declared as 'num6'.

The parameter format for a num is as follows:
parameter_name|group|type|default|minimum|maximum

The interpretation of group is left to the individual host application. 
Generally this is intended to work as follows; only one parameter in a non-zero
 group can take its non-off value, off value being 0 for booleans, and default 
for other parameter types.


IMPORTANT NOTE
An underscore in the parameter_name indicates that the following character is a
mnemonic accelerator for the parameter. In case the host does not support 
mnemonic accelerators, underscores should be stripped from the parameter_name.




e.g.

<params>
_Rotate Selection by|0|num0|360|-3600|3600|
</params>


bool: this indicates a boolean parameter
The format for this parameter type is:

parameter_name|group|type|default

e.g.

<params>
_Shrink rotated window to fit frame size|0|bool|0|
</params>


colRGB24: this indicates a 24 bit RGB value. The leftmost 8 bits correspond to 
red, the next 8 bits to green, and the rightmost 8 bits to blue.

The format for this parameter type is:
parameter_name|group|type|default_red|default_green|default_blue

e.g.
<params>
_Start Colour|0|colRGB24|255|255|255
</params>






<param_window> [optional] [list]

This section is used to provide layout hints to the host about how to draw a 
parameter window for this effect/tool.

The format is a list of layout hints separated by newlines. Each line in the 
list has a keyword followed by one or more subfields separated by "|"

Two keywords are currently supported:

layout
special


layout is hint to the host about how to lay out the parameter window. It is 
abstracted from any particular widget set. 

It is up to the host how each layout line in this section is actually 
interpreted.
 
One way to implement this is to assume each layout line describes a horizontal 
box. These horizontal boxes can be laid out within a vertical container. Any 
unassigned parameters can be added below by the host.

The layout keyword is followed by 0 or more fields which describe a layout row.
the fields can be:

px == position of parameter x on the screen
fill == fill with blank space
"label" == a label
hseparator == horizontal line separator


e.g:

<param_window>
layout|p1|
layout|p6|
layout|p2|
layout|p3|
layout|hseparator|
layout|p4|fill|
layout|p5|fill|
</param_window>



A second keyword which can be used in the param_window section is "special".
This indicates that the host can optionally add a widget to the window which 
links together some of the parameters in a special way.

Two examples are:

special|aspect|1|2
special|framedraw|rectdemask|1|2|3|4

It is up to individual host authors how they interpret the special keyword, 
but definitions from different host applications should not overlap.





<properties> [optional]

this is a bitmap field (currently 32 bit)
some bits are defined already

0x0001 == slow (hint to host)
0x0002 == may resize

If the "may resize" bit is set, _all_ frames in the processing block may be 
resized to a new width and height. The plugin should inform the host about 
the new frame width and height. The exact mechanism for this information to 
be passed back to the host is still to be decided.

0x8000 == reserved. This bit is reserved by the LiVES video editor.

If the value of properties is not defined, it is assumed to be 0x0000.



e.g.
<properties>
0x0001
</properties>




<language_code>

The language code indicates the scripting language for the <pre>, <loop>, 
<post> and <onchange> sections. This is a 32 bit unsigned integer.

Current values are:

0x000F == LiVES-perl




<pre> [optional] [list]

This section contains code which will be executed before the processing loop.

e.g. (in LiVES-perl):

<pre>
#calculate the number of frames
$length=$end-$start+1;
</pre>



<loop> [list]

IMPORTANT NOTE
This section is optional for non-processing scripts (where min frames is set 
to -1, see above)

This section contains code which will be executed for each frame in the 
processing block.

e.g. (in LiVES-perl):

<loop>
#negate each frame
system("$convert_command -negate $in $out");
</loop>


The rules for processing are as follows:

in each pass of the loop you should produce an output frame (possibly by using 
the input frame). For example, in LiVES-perl, the output frame is called $out.
The output frames should be created in numerical from the start of the 
processing block to the end in numerical order with no gaps. The image type 
should not be changed, nor should the palette. Frames may be resized provided 
the script has the "may resize" property bit set.

If no processing is to be done on a pass, then the input frame should be 
copied to the output frame.




<post> [optional] [list]

This section contains code which should be executed after all frames in the 
block have been processed.

e.g. (in LiVES-perl):

<post>
#clean up a temporary file
unlink "blank.jpg";
</post>




<onchange> [optional] [list]

This section contains the triggers for changing parameters either when the 
parameter window is first created, or when a parameter is changed. For example
 an init type trigger could set the maximum and minimum values for a parameter 
depending on the current frame width and height.

The format is:

event|code

event can be either "init" or a number representing a parameter. The first 
parameter is given the number 1, the second parameter 2, etc.
There can be multiple lines of code for each type of change. Code for the same 
trigger type is executed in the order in which it appears in this section.


e.g. (in LiVES-perl):

<onchange>
init|$p1_min=-$width+1;$p1_max=$width-1;$p2_min=-$height+1;$p2_max=$height-1;
1|if ($p1) {$p4=$p5*$p6;} elsif ($p2&&$p6>0.) {$p5=$p4/$p6;} elsif ($p5>0.) {$p6=$p4/$p5;}
</onchange>



