#include "plugins-livido.h"
#include <stdio.h>

livido_setup_f livido_setup;

livido_init_f alien_over_init;
livido_deinit_f alien_over_deinit;
livido_process_f alien_over_process;

static int inited;
static unsigned char *old_pixel_data;

/* declare our init function */
int alien_over_init(livido_instance_t *inst) {
  int rowstride=inst->in_channels[0].rowstrides[0];
  int size=inst->in_channels[0].height*rowstride;
  old_pixel_data=(unsigned char *)malloc(size);
  inited=0;
  return 0;
}


int alien_over_process
(livido_instance_t *env, livido_frame_t **in, livido_frame_t **out) {
  // host should have returned one input channel and one output channel
  // (which can be NULL, since we exported the property INPLACE)

  // we also set our input/output channels same as, so we can assume both frame
  // sizes and palettes match

  int i,j,k;
  int width=env->in_channels[0].width*3;
  int rowstride_i=env->in_channels[0].rowstrides[0];
  int size=env->in_channels[0].height*rowstride_i;

  unsigned char *curr_pixel_data=in[0]->pixel_data, *new_pixel_data, byte;
  int rowstride_o;

  int inplace=(out==NULL||out[0]==NULL);


  if (inplace) {
    new_pixel_data=curr_pixel_data;
    rowstride_o=rowstride_i;
  }
  else {
    new_pixel_data=out[0]->pixel_data;
    rowstride_o=env->out_channels[0].rowstrides[0];
  }


  for (i=0;i<size;i+=rowstride_i) {
    for (j=0;j<width;j++) {
      k=i/rowstride_i*rowstride_o;
      byte=curr_pixel_data[i+j];
      if (inited) {
	new_pixel_data[k+j]=((char)(byte=curr_pixel_data[i+j])+(char)old_pixel_data[i+j])/2;
      }
      else if (!inplace) new_pixel_data[k+j]=curr_pixel_data[i+j];
      old_pixel_data[i+j]=byte;
    }
  }
  inited=1;
  return 0;
}


int alien_over_deinit (livido_instance_t *inst) {
  if (inited) free(old_pixel_data);
  return 0; // success
}


// create 1 input and one matching output channel, no masks

// allow RGB888 and BGR888 palettes

livido_palette_t in_palettes[]={LVD_PALETTE_RGB888,LVD_PALETTE_BGR888,LVD_PALETTE_END};
livido_palette_t out_palettes[]={LVD_PALETTE_RGB888,LVD_PALETTE_BGR888,LVD_PALETTE_END};

livido_channel_template_t in_channels[]={{"In 1",0,in_palettes,0,0,0},LIVIDO_CHANNEL_TEMPLATE_END};

livido_channel_template_t out_channels[]={{"Out 1",LVD_CHNTMPL_SAME_AS_SIZE|LVD_CHNTMPL_SAME_AS_PALETTE,out_palettes,1,0,0},LIVIDO_CHANNEL_TEMPLATE_END};

livido_class_t fx[]={{"alien_overlay","salsaman","alien overlay",1,1.7,LVD_PROPERTY_REALTIME|LVD_PROPERTY_INPLACE|LVD_PROPERTY_STATIC_CHANNEL_SIZES,NULL,in_channels,out_channels,alien_over_init,alien_over_deinit,alien_over_process},LIVIDO_CLASS_END};


livido_class_t *livido_setup(void) {
  return fx;
}

