/* libvisual plugin for livido
   authors: Salsaman (G. Finch) <salsaman@xs4all.nl>
            Dennis Smit  <synap@yourbase.nl>
            Duilio J. Protti <dprotti@users.sourceforge.net>

   Released under the Lesser Gnu Public License (LGPL)

 (c) 2004, project authors
*/


#define LVD_ERROR_TOO_MANY_INSTANCES 1

// WARNING ! Only "jack" and "host audio" inputs are multi threaded

static int instances;

#include "plugins-livido.h"
#include <stdio.h>
#include <string.h>
#include <libvisual/lv_actor.h>
#include <libvisual/lv_input.h>
#include <libvisual/lv_libvisual.h>

livido_setup_f livido_setup;

livido_init_f libvis_init;
livido_deinit_f libvis_deinit;
livido_process_f libvis_process;

int libvis_host_audio_callback (VisInput *, VisAudio *, void *user_data);

typedef struct {
  VisVideo *video;
  VisActor *actor;
  VisInput *input;
  void *audio_data;
} livido_libvis_t;


/* declare our init function */
int libvis_init(livido_instance_t *inst) {
  
  livido_libvis_t *libvis;

  inst->internal=NULL;
  if (instances) return 1;

  libvis=(livido_libvis_t *)malloc (sizeof(livido_libvis_t));

  // TODO - get list of inputs, add LVD_PARAM_LIST with values for inputs

  //libvis->input = visual_input_new (NULL); // no audio input

  libvis->input = visual_input_new ("alsa");
  //libvis->input = visual_input_new ("esd");
  //libvis->input = visual_input_new ("jack");
  //libvis->input = visual_input_new ("mplayer");

  // host supplied audio buffers
  //visual_input_set_callback (libvis->input, libvis_host_audio_callback, (void *)libvis);

  if (libvis->input==NULL) { // wish this worked
    free (libvis);
    return LVD_ERROR_TOO_MANY_INSTANCES;
  }

  instances++;
  libvis->video = visual_video_new();

  // Force RGB 24
  visual_video_set_depth (libvis->video, VISUAL_VIDEO_DEPTH_24BIT);
  visual_video_set_dimension (libvis->video, inst->out_channels[0].width, inst->out_channels[0].height);

  // assume pitch is normal
  //visual_video_set_pitch (video, inst->out_);

  libvis->actor = visual_actor_new (inst->class->name);

  visual_actor_realize (libvis->actor);
  visual_actor_set_video (libvis->actor, libvis->video);
  visual_actor_video_negotiate (libvis->actor, 0, FALSE, FALSE);
  visual_input_realize (libvis->input);
  libvis->audio_data=NULL;

  inst->internal=(void *)libvis;
  return 0;
}


int libvis_process
(livido_instance_t *env, livido_frame_t **in, livido_frame_t **out) {
  // host should have returned one output channel and no input channels

  livido_libvis_t *libvis=(livido_libvis_t *)env->internal;
  visual_input_run (libvis->input);
  visual_video_set_buffer (libvis->video, out[0]->pixel_data);
  visual_actor_run (libvis->actor, libvis->input->audio);
  return 0;
}


int libvis_deinit (livido_instance_t *inst) {

  livido_libvis_t *libvis;
  if (inst->internal!=NULL) {
    libvis=(livido_libvis_t *)inst->internal;
    if (libvis->input!=NULL) visual_input_destroy (libvis->input);
    if (libvis->video!=NULL) visual_video_free (libvis->video);
    if (libvis->actor!=NULL) visual_actor_destroy (libvis->actor);
    free (libvis);
    inst->internal=NULL;
    }
  if (--instances<0) instances=0;
  return 0; // success
}



livido_palette_t out_palettes[]={LVD_PALETTE_RGB888,LVD_PALETTE_END};

livido_channel_template_t in_channels[]={LIVIDO_CHANNEL_TEMPLATE_END};
livido_channel_template_t out_channels[]={{"Out 1",0,out_palettes,0,0,0},LIVIDO_CHANNEL_TEMPLATE_END};

livido_class_t *livido_setup(void) {
  char *name=NULL;
  int count=0;
  livido_class_t *fx=NULL;

  instances=0;

  if (VISUAL_PLUGIN_API_VERSION>0) visual_log_set_verboseness (VISUAL_LOG_VERBOSENESS_NONE);

  if (visual_init (NULL,NULL)<0) {
    fprintf (stderr, "Libvis : Unable to init libvisual plugins\n");
    fx=malloc (sizeof(livido_class_t));
    fx->name=NULL;
    return fx;
  }

  for (;(name=visual_actor_get_next_by_name_nogl (name))!=NULL;count++);
  fx=malloc ((count+1)*sizeof(livido_class_t));
  for (count=0;(name=visual_actor_get_next_by_name_nogl (name))!=NULL;count++) {
    fx[count].name=strdup (name);
    fx[count].author=strdup ("libvisual");
    fx[count].description=strdup (name);
    fx[count].version=1;
    fx[count].livido_api_version=1.7;
    fx[count].properties=LVD_PROPERTY_REALTIME;
    fx[count].parameters=NULL;
    fx[count].in_channel_templates=in_channels;
    fx[count].out_channel_templates=out_channels;
    fx[count].init=&libvis_init;
    fx[count].deinit=&libvis_deinit;
    fx[count].process=&libvis_process;
  }
  fx[count].name=NULL;
  return fx;
}




///////////// callback so host can supply its own audio sammples //////////


int libvis_host_audio_callback (VisInput *input, VisAudio *audio, void *user_data) {
  // audio_data is 16bit signed, little endian
  // two channels non-interlaced, 512 samples in length
  // a rate of 44100Hz is recommended

  // since we have no in channels, we have to use the audio_data of the 
  // output channel (!)

  // return -1 on failure, 0 on success

  livido_libvis_t *libvis=(livido_libvis_t *)user_data;
  if (libvis->audio_data!=NULL) {
    memcpy (audio->plugpcm,libvis->audio_data,2048);
  }
  return 0;
}
