#include "plugins-livido.h"
#include <stdio.h>

livido_setup_f livido_setup;

livido_init_f mirror_init;
livido_deinit_f mirror_deinit;
livido_process_f mirrorx_process;
livido_process_f mirrory_process;
livido_process_f mirrorxy_process;



/* declare our init function */
int mirror_init(livido_instance_t *inst) {
  return 0;
}


int mirrorx_process
(livido_instance_t *env, livido_frame_t **in, livido_frame_t **out) {
  // host should have returned one input channel and one output channel
  // (which can be NULL, since we exported the property INPLACE)

  // we also set our input/output channels same as, so we can assume both frame
  // sizes and palettes match

  int i,j,offset=0;
  int rowstride=env->in_channels[0].rowstrides[0];
  int width=env->in_channels[0].width*3;
  int size=env->in_channels[0].height*rowstride;

  unsigned char *old_pixel_data=in[0]->pixel_data, *new_pixel_data;
  livido_palette_t palette=env->in_channels[0].palette;

  int inplace=(out==NULL||out[0]==NULL);

  if (palette==LVD_PALETTE_BGR888) offset=-1;

  if (inplace) {
    new_pixel_data=old_pixel_data;
  }
  else new_pixel_data=out[0]->pixel_data;


  for (i=0;i<size;i+=rowstride) {
    for (j=0;j<width/2;j++) {
      new_pixel_data[i+width-1-j+offset*2]=old_pixel_data[i+j];
      if (!inplace) new_pixel_data[i+j]=old_pixel_data[i+j];
      if (palette==LVD_PALETTE_BGR888) offset=++offset<2?offset:-1;
    }
  } 
  return 0;
}



int mirrory_process
(livido_instance_t *env, livido_frame_t **in, livido_frame_t **out) {
  // host should have returned one input channel and one output channel
  // (which can be NULL, since we exported the property INPLACE)

  // we also set our input/output channels same as, so we can assume both frame
  // sizes and palettes match

  int i,j;
  int rowstride=env->in_channels[0].rowstrides[0];
  int width=env->in_channels[0].width*3;
  int size=env->in_channels[0].height*rowstride;

  unsigned char *old_pixel_data=in[0]->pixel_data, *new_pixel_data;

  int inplace=(out==NULL||out[0]==NULL);

  if (inplace) {
    new_pixel_data=old_pixel_data;
  }
  else new_pixel_data=out[0]->pixel_data;

  for (i=0;i<size/2;i+=rowstride) {
    for (j=0;j<width;j++) {
      new_pixel_data[size-rowstride-i+j]=old_pixel_data[i+j];
      if (!inplace) new_pixel_data[i+j]=old_pixel_data[i+j];
    }
  } 
  return 0;
}



int mirrorxy_process
(livido_instance_t *env, livido_frame_t **in, livido_frame_t **out) {
  int ret=mirrorx_process(env,in,out);
  if (ret) return ret;
  return mirrory_process(env,in,out);
}






int mirror_deinit (livido_instance_t *inst) {
  return 0; // success
}


// create 1 input and one matching output channel, no masks

// allow RGB888 and BGR888 palettes

livido_palette_t in_palettes[]={LVD_PALETTE_RGB888,LVD_PALETTE_BGR888,LVD_PALETTE_END};
livido_palette_t out_palettes[]={LVD_PALETTE_RGB888,LVD_PALETTE_BGR888,LVD_PALETTE_END};

livido_channel_template_t in_channels[]={{"In 1",0,in_palettes,0,0,0},LIVIDO_CHANNEL_TEMPLATE_END};

livido_channel_template_t out_channels[]={{"Out 1",LVD_CHNTMPL_SAME_AS_SIZE|LVD_CHNTMPL_SAME_AS_PALETTE,out_palettes,1,0,0},LIVIDO_CHANNEL_TEMPLATE_END};

livido_class_t fx[]={{"mirrorx","salsaman","mirror example",1,1.7,LVD_PROPERTY_REALTIME|LVD_PROPERTY_INPLACE,NULL,in_channels,out_channels,mirror_init,mirror_deinit,mirrorx_process},{"mirrory","salsaman","mirror example",1,1.7,LVD_PROPERTY_REALTIME|LVD_PROPERTY_INPLACE,NULL,in_channels,out_channels,mirror_init,mirror_deinit,mirrory_process},{"mirrorxy","salsaman","mirror example",1,1.7,LVD_PROPERTY_REALTIME|LVD_PROPERTY_INPLACE,NULL,in_channels,out_channels,mirror_init,mirror_deinit,mirrorxy_process},LIVIDO_CLASS_END};


livido_class_t *livido_setup(void) {
  return fx;
}

