#include "plugins-livido.h"
#include <stdio.h>

livido_setup_f livido_setup;

livido_init_f noise_init;
livido_deinit_f noise_deinit;
livido_process_f noise_process;

static uint32_t fastrand_val;

inline uint32_t fastrand(void)
{
#define rand_a 1073741789L
#define rand_c 32749L

  return (fastrand_val= rand_a * fastrand_val + rand_c);
}

/* declare our init function */
int noise_init(livido_instance_t *inst) {
  fastrand_val=0; // TODO - seed with random seed
  return 0;
}


int noise_process
(livido_instance_t *env, livido_frame_t **in, livido_frame_t **out) {
  // host should have returned one input channel and one output channel
  // (which can be NULL, since we exported the property INPLACE)

  // we also set our input/output channels same as, so we can assume both frame
  // sizes and palettes match

  int i,j,k;
  int rowstride_i=env->in_channels[0].rowstrides[0];
  int width=env->in_channels[0].width*3;
  int size=env->in_channels[0].height*rowstride_i;
  int rowstride_o;

  unsigned char *old_pixel_data=in[0]->pixel_data, *new_pixel_data;

  int inplace=(out==NULL||out[0]==NULL);

  if (inplace) {
    new_pixel_data=old_pixel_data;
    rowstride_o=rowstride_i;
  }
  else {
    new_pixel_data=out[0]->pixel_data;
    rowstride_o=env->out_channels[0].rowstrides[0];      
  }
  for (i=0;i<size;i+=rowstride_i) {
    for (j=0;j<width;j++) {
      k=i/rowstride_i*rowstride_o;
      new_pixel_data[i+j]=old_pixel_data[i+j]+((fastrand()&0xF8000000)>>27)-16;
    }
  } 
  return 1;
}


int noise_deinit (livido_instance_t *inst) {
  return 1; // success
}


// create 1 input and one matching output channel, no masks

// allow RGB888 and BGR888 palettes

livido_palette_t in_palettes[]={LVD_PALETTE_RGB888,LVD_PALETTE_BGR888,LVD_PALETTE_END};
livido_palette_t out_palettes[]={LVD_PALETTE_RGB888,LVD_PALETTE_BGR888,LVD_PALETTE_END};

livido_channel_template_t in_channels[]={{"In 1",0,in_palettes,0,0,0},LIVIDO_CHANNEL_TEMPLATE_END};

livido_channel_template_t out_channels[]={{"Out 1",LVD_CHNTMPL_SAME_AS_SIZE|LVD_CHNTMPL_SAME_AS_PALETTE,out_palettes,1,0,0},LIVIDO_CHANNEL_TEMPLATE_END};

livido_class_t fx[]={{"noise","salsaman","random noise",1,1.7,LVD_PROPERTY_REALTIME|LVD_PROPERTY_INPLACE,NULL,in_channels,out_channels,noise_init,noise_deinit,noise_process},LIVIDO_CLASS_END};


livido_class_t *livido_setup(void) {
  return fx;
}

