#include "plugins-livido.h"
#include <stdio.h>

livido_setup_f livido_setup;

livido_init_f xeffect_init;
livido_deinit_f xeffect_deinit;
livido_process_f xeffect_process;



#define ABS(a)           (((a) < 0) ? -(a) : (a))


static inline unsigned int 
calc_luma (unsigned char *pixel) {
  return (unsigned int)(pixel[0]*30+pixel[1]*59+pixel[2]*11);
}

static inline void 
make_black (unsigned char *pixel) {
  pixel[0]=pixel[1]=pixel[2]=(unsigned char)0;
}

static inline void 
make_white (unsigned char *pixel) {
  pixel[0]=pixel[1]=pixel[2]=(unsigned char)255;
}


static inline void 
nine_fill (unsigned char *new_data, int i, int j, int rowstride, unsigned char *old_data, int x, int y) {
  // fill nine pixels with the centre colour
  new_data[i-rowstride+j-3]=new_data[i-rowstride]=new_data[i-rowstride+3]=new_data[i+j-3]=new_data[i+j]=new_data[i+j+3]=new_data[i+rowstride+j-3]=new_data[i+rowstride+j]=new_data[i+rowstride+j+3]=old_data[x+y];
  new_data[i-rowstride+j-2]=new_data[i-rowstride+1]=new_data[i-rowstride+4]=new_data[i+j-2]=new_data[i+j+1]=new_data[i+j+4]=new_data[i+rowstride+j-2]=new_data[i+rowstride+j+1]=new_data[i+rowstride+j+4]=old_data[x+y+1];
  new_data[i-rowstride+j-1]=new_data[i-rowstride+2]=new_data[i-rowstride+5]=new_data[i+j-1]=new_data[i+j+2]=new_data[i+j+5]=new_data[i+rowstride+j-1]=new_data[i+rowstride+j+2]=new_data[i+rowstride+j+5]=old_data[x+y+2];
}




/* declare our init function */
int xeffect_init(livido_instance_t *inst) {
  return 0;
}


int xeffect_process
(livido_instance_t *env, livido_frame_t **in, livido_frame_t **out) {
  // host should have returned one input channel and one output channel
  // (which can be NULL, since we exported the property INPLACE)

  // we also set our input/output channels same as, so we can assume both frame
  // sizes and palettes match

  unsigned int myluma;
  unsigned int threshold=10000;
  int nbr;
  register int i,j,k,l,o;

  int width=env->in_channels[0].width*3;
  int in_rowstride=env->in_channels[0].rowstrides[0];
  int out_rowstride=env->out_channels[0].rowstrides[0];
  int size=env->in_channels[0].height*in_rowstride;
  unsigned char *in_data=in[0]->pixel_data;
  unsigned char *out_data=out[0]->pixel_data;

  for (i=in_rowstride;i<size-in_rowstride;i+=in_rowstride) {
    o=i/in_rowstride*out_rowstride;
    for (j=3;j<width-3;j+=3) {
 
      myluma=calc_luma (&(in_data[i+j]));
      nbr=0;
      for (k=i-in_rowstride;k<=i+in_rowstride;k+=in_rowstride) {
	for (l=j-3;l<=j+3;l+=3) {
	  if ((k!=i||j!=l)&&ABS (calc_luma (&in_data[k+l])-myluma)>threshold) nbr++;
	}
      }
      if (nbr<2||nbr>5) {
	nine_fill (out_data,o,j,out_rowstride,in_data,i,j);
      }
      else {
	if (myluma<12500) {
	  make_black (&out_data[o+j]);
	}
	else{
	  if (myluma>20000) {
	    make_white (&out_data[o+j]);

	  }
	}
      }
    }
  }
  return 0;
}



int xeffect_deinit (livido_instance_t *inst) {
  return 0; // success
}


// create 1 input and one matching output channel, no masks

// allow RGB888 and BGR888 palettes

livido_palette_t in_palettes[]={LVD_PALETTE_RGB888,LVD_PALETTE_BGR888,LVD_PALETTE_END};
livido_palette_t out_palettes[]={LVD_PALETTE_RGB888,LVD_PALETTE_BGR888,LVD_PALETTE_END};

livido_channel_template_t in_channels[]={{"In 1",0,in_palettes,0,0,0},LIVIDO_CHANNEL_TEMPLATE_END};

livido_channel_template_t out_channels[]={{"Out 1",LVD_CHNTMPL_SAME_AS_SIZE|LVD_CHNTMPL_SAME_AS_PALETTE,out_palettes,1,0,0},LIVIDO_CHANNEL_TEMPLATE_END};

livido_class_t fx[]={{"graphic novel","salsaman","graphic novel",1,1.7,LVD_PROPERTY_REALTIME,NULL,in_channels,out_channels,xeffect_init,xeffect_deinit,xeffect_process},LIVIDO_CLASS_END};


livido_class_t *livido_setup(void) {
  return fx;
}

