// colourspace.h
// LiVES
// (c) G. Finch 2004 <salsaman@xs4all.nl>
// Released under the GPL 2.0
// see file ../COPYING for licensing details

// code for palette conversions

/*
 *  Some portions
 *  Copyright (C) 2001 Matthew J. Marjanovic <maddog@mir.com>
 *
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
// modifications to handle packed RGB24 by G. Finch (salsaman)


#include "main.h"


/* precomputed tables */

static gint Y_R[256];
static gint Y_G[256];
static gint Y_B[256];
static gint Cb_R[256];
static gint Cb_G[256];
static gint Cb_B[256];
static gint Cr_R[256];
static gint Cr_G[256];
static gint Cr_B[256];
static gint conv_RY_inited = 0;

static gint RGB_Y[256];
static gint R_Cr[256];
static gint G_Cb[256];
static gint G_Cr[256];
static gint B_Cb[256];
static gint conv_YR_inited = 0;


static gint myround(gdouble n)
{
  if (n >= 0) 
    return (gint)(n + 0.5);
  else
    return (gint)(n - 0.5);
}



static void init_RGB_to_YCbCr_tables(void)
{
  gint i;

  /*
   * Q_Z[i] =   (coefficient * i
   *             * (Q-excursion) / (Z-excursion) * fixed-pogint-factor)
   *
   * to one of each, add the following:
   *             + (fixed-pogint-factor / 2)         --- for rounding later
   *             + (Q-offset * fixed-pogint-factor)  --- to add the offset
   *             
   */
  for (i = 0; i < 256; i++) {
    Y_R[i] = myround(0.299 * (gdouble)i 
		     * 219.0 / 255.0 * (gdouble)(1<<FP_BITS));
    Y_G[i] = myround(0.587 * (gdouble)i 
		     * 219.0 / 255.0 * (gdouble)(1<<FP_BITS));
    Y_B[i] = myround((0.114 * (gdouble)i 
		      * 219.0 / 255.0 * (gdouble)(1<<FP_BITS))
		     + (gdouble)(1<<(FP_BITS-1))
		     + (16.0 * (gdouble)(1<<FP_BITS)));

    Cb_R[i] = myround(-0.168736 * (gdouble)i 
		      * 224.0 / 255.0 * (gdouble)(1<<FP_BITS));
    Cb_G[i] = myround(-0.331264 * (gdouble)i 
		      * 224.0 / 255.0 * (gdouble)(1<<FP_BITS));
    Cb_B[i] = myround((0.500 * (gdouble)i 
		       * 224.0 / 255.0 * (gdouble)(1<<FP_BITS))
		      + (gdouble)(1<<(FP_BITS-1))
		      + (128.0 * (gdouble)(1<<FP_BITS)));

    Cr_R[i] = myround(0.500 * (gdouble)i 
		      * 224.0 / 255.0 * (gdouble)(1<<FP_BITS));
    Cr_G[i] = myround(-0.418688 * (gdouble)i 
		      * 224.0 / 255.0 * (gdouble)(1<<FP_BITS));
    Cr_B[i] = myround((-0.081312 * (gdouble)i 
		       * 224.0 / 255.0 * (gdouble)(1<<FP_BITS))
		      + (gdouble)(1<<(FP_BITS-1))
		      + (128.0 * (gdouble)(1<<FP_BITS)));
  }
  conv_RY_inited = 1;
}




static void init_YCbCr_to_RGB_tables(void)
{
  gint i;

  gint startclamp_y=16;
  gint endclamp_y=235;

  gint startclamp_crcb=16;
  gint endclamp_crcb=240;

  /*
   * Q_Z[i] =   (coefficient * i
   *             * (Q-excursion) / (Z-excursion) * fixed-pogint-factor)
   *
   * to one of each, add the following:
   *             + (fixed-pogint-factor / 2)         --- for rounding later
   *             + (Q-offset * fixed-pogint-factor)  --- to add the offset
   *             
   */

  /* clip Y values under 16 */
    for (i = 0; i < startclamp_y; i++) {
    RGB_Y[i] = myround((1.0 * (gdouble)(16 - 16) 
		     * 255.0 / 219.0 * (gdouble)(1<<FP_BITS))
		    + (gdouble)(1<<(FP_BITS-1)));
		    }
  for (i = startclamp_y; i <= endclamp_y; i++) {
    RGB_Y[i] = myround((1.0 * (gdouble)(i - 16) 
		     * 255.0 / 219.0 * (gdouble)(1<<FP_BITS))
		    + (gdouble)(1<<(FP_BITS-1)));
  }
  /* clip Y values above 235 */
    for (i = endclamp_y+1; i < 256; i++) {
    RGB_Y[i] = myround((1.0 * (gdouble)(235 - 16) 
		     * 255.0 / 219.0 * (gdouble)(1<<FP_BITS))
		    + (gdouble)(1<<(FP_BITS-1)));
  }
  
  /* clip Cb/Cr values below 16 */	 
    for (i = 0; i < startclamp_crcb; i++) {
    R_Cr[i] = myround(1.402 * (gdouble)(-112)
		   * 255.0 / 224.0 * (gdouble)(1<<FP_BITS));
    G_Cr[i] = myround(-0.714136 * (gdouble)(-112)
		   * 255.0 / 224.0 * (gdouble)(1<<FP_BITS));
    G_Cb[i] = myround(-0.344136 * (gdouble)(-112)
		   * 255.0 / 224.0 * (gdouble)(1<<FP_BITS));
    B_Cb[i] = myround(1.772 * (gdouble)(-112)
		   * 255.0 / 224.0 * (gdouble)(1<<FP_BITS));
		   }
  for (i = startclamp_crcb; i <= endclamp_crcb; i++) {
    R_Cr[i] = myround(1.402 * (gdouble)(i - 128)
		   * 255.0 / 224.0 * (gdouble)(1<<FP_BITS));
    G_Cr[i] = myround(-0.714136 * (gdouble)(i - 128)
		   * 255.0 / 224.0 * (gdouble)(1<<FP_BITS));
    G_Cb[i] = myround(-0.344136 * (gdouble)(i - 128)
		   * 255.0 / 224.0 * (gdouble)(1<<FP_BITS));
    B_Cb[i] = myround(1.772 * (gdouble)(i - 128)
		   * 255.0 / 224.0 * (gdouble)(1<<FP_BITS));
  }
  /* clip Cb/Cr values above 240 */	 
    for (i = endclamp_crcb+1; i < 256; i++) {
    R_Cr[i] = myround(1.402 * (gdouble)(112)
		   * 255.0 / 224.0 * (gdouble)(1<<FP_BITS));
    G_Cr[i] = myround(-0.714136 * (gdouble)(112)
		   * 255.0 / 224.0 * (gdouble)(1<<FP_BITS));
    G_Cb[i] = myround(-0.344136 * (gdouble)(i - 128)
		   * 255.0 / 224.0 * (gdouble)(1<<FP_BITS));
    B_Cb[i] = myround(1.772 * (gdouble)(112)
		   * 255.0 / 224.0 * (gdouble)(1<<FP_BITS));
		   }
  conv_YR_inited = 1;
}


static inline void rgb2uyvy (guchar r0, guchar g0, guchar b0, guchar r1, guchar g1, guchar b1, uyvy_frame *uyvy) {
  register int a,b,c;
  uyvy->y0 = (Y_R[r0] + Y_G[g0]+ Y_B[b0]) >> FP_BITS;
  uyvy->y1 = (Y_R[r1] + Y_G[g1]+ Y_B[b1]) >> FP_BITS;
  if ((c=(((a=((Cb_R[r0]+Cb_G[g0]+Cb_B[b0])>>FP_BITS)-128)+(b=((Cb_R[r1]+Cb_G[g1]+Cb_B[b1])>>FP_BITS)-128))-((a*b)>>8))+128)>240) c=240;
  uyvy->v0=c<16?16:c;
  if ((c=(((a=((Cr_R[r0]+Cr_G[g0]+Cr_B[b0])>>FP_BITS)-128)+(b=((Cr_R[r1]+Cr_G[g1]+Cr_B[b1])>>FP_BITS)-128))-((a*b)>>8))+128)>240) c=240;
  uyvy->u0=c<16?16:c;
}

guchar 
calc_luma (guchar *pixel) {
  if (!conv_RY_inited) init_RGB_to_YCbCr_tables();
  return (Y_R[pixel[0]] + Y_G[pixel[1]]+ Y_B[pixel[2]]) >> FP_BITS;
}



void uyvy2rgb (uyvy_frame *uyvy, guchar *r0, guchar *g0, guchar *b0, guchar *r1, guchar *g1, guchar *b1) {
  register int r,g,b;

  if (!conv_YR_inited) init_YCbCr_to_RGB_tables();

    r = (RGB_Y[uyvy->y0] + R_Cr[uyvy->v0]) >> FP_BITS;
    g = (RGB_Y[uyvy->y0] + G_Cb[uyvy->u0]+ G_Cr[uyvy->v0]) >> FP_BITS;
    b = (RGB_Y[uyvy->y0] + B_Cb[uyvy->u0]) >> FP_BITS;

    *r0 = (r < 0) ? 0 : (r > 255) ? 255 : r ;
    *g0 = (g < 0) ? 0 : (g > 255) ? 255 : g ;
    *b0 = (b < 0) ? 0 : (b > 255) ? 255 : b ;

    r = (RGB_Y[uyvy->y1] + R_Cr[uyvy->v0]) >> FP_BITS;
    g = (RGB_Y[uyvy->y1] + G_Cb[uyvy->u0]+ G_Cr[uyvy->v0]) >> FP_BITS;
    b = (RGB_Y[uyvy->y1] + B_Cb[uyvy->u0]) >> FP_BITS;

    *r1 = (r < 0) ? 0 : (r > 255) ? 255 : r ;
    *g1 = (g < 0) ? 0 : (g > 255) ? 255 : g ;
    *b1 = (b < 0) ? 0 : (b > 255) ? 255 : b ;
}

// remaining code (c) salsaman






void 
bgr2uyvy_image(guchar *bgrdata, gint hsize, gint vsize, gint rowstride) {
  // for odd sized widths, cut the rightmost pixel
  gint hs3=(int)(hsize/2)*6;

  gint size=rowstride*vsize;
  uyvy_frame* u;
  register int i,j;

  mainw->noswitch=TRUE;
  // user could have switched clips while we were not looking...
  if (mainw->reinit_effects&&mainw->uyvy_frame!=NULL) {
    g_free(mainw->uyvy_frame);
    mainw->uyvy_frame=NULL;
  }
  mainw->noswitch=FALSE;

  if (mainw->uyvy_frame==NULL) {
    size_t uyvy_size=(size_t)(hs3/6*vsize);
    mainw->uyvy_frame=(uyvy_frame*)g_try_malloc (uyvy_size*sizeof(uyvy_frame));
    if (mainw->uyvy_frame==NULL) return;
  }

  u=mainw->uyvy_frame;

  hs3-=5;

  if (!conv_RY_inited) init_RGB_to_YCbCr_tables();

  for (i=0;i<size;i+=rowstride) {
    for (j=0;j<hs3;j+=6) {
      // convert 6 RGBRGB bytes to 4 UYVY bytes
      rgb2uyvy (bgrdata[i+j+2],bgrdata[i+j+1],bgrdata[i+j],bgrdata[i+j+5],bgrdata[i+j+4],bgrdata[i+j+3],u++);
    }
  }
}



void 
bgr2yuv420_image(guchar *bgrdata, gint hsize, gint vsize, gint rowstride) {
  // for odd sized widths, cut the rightmost pixel
  gint hs3=(int)(hsize/2)*6;

  guchar *y,*Cb,*Cr;
  gint size=hsize*vsize;
  uyvy_frame u;
  register int i,j;
  gint chroma_row=1;

  mainw->noswitch=TRUE;
  // user could have switched clips while we were not looking...
  if (mainw->reinit_effects&&mainw->uyvy_frame!=NULL) {
    g_free(mainw->yuv_frame);
    mainw->yuv_frame=NULL;
  }
  mainw->noswitch=FALSE;

  if (mainw->yuv_frame==NULL) {
    mainw->yuv_frame=(guchar *)g_try_malloc (size*1.5);
    if (mainw->yuv_frame==NULL) return;
  }
  y=mainw->yuv_frame;
  Cb=&(y[size]);
  Cr=&(y[size*5/4]);

  size=rowstride*vsize;

  if (!conv_RY_inited) init_RGB_to_YCbCr_tables();

  hs3-=5;
  for (i=0;i<size;i+=rowstride) {
    for (j=0;j<hs3;j+=6) {
      // convert 6 RGBRGB bytes to 4 UYVY bytes
      rgb2uyvy (bgrdata[i+j+2],bgrdata[i+j+1],bgrdata[i+j],bgrdata[i+j+5],bgrdata[i+j+4],bgrdata[i+j+3],&u);

      // TODO - make a UYVY frame here which can be passed to SDL
      // allow multiple pb plugins

      y[0]=u.y0;
      y[1]=u.y1;

      // halve the chroma samples for (lower quality) yuv420
      if (chroma_row) {
	Cb[0]=u.u0;
	Cr[0]=u.v0;
      }
      else {
	// average two rows
	Cb[0]=(Cb[0]+u.u0)>>1;
	Cr[0]=(Cr[0]+u.v0)>>1;
      }
      Cb++;
      Cr++;
      y+=2;
    }
    if (chroma_row) {
      Cb-=hsize/2;
      Cr-=hsize/2;
    }
    chroma_row=!chroma_row;
  }
}


void RGB2BGR_image (guchar *pixels, gint hsize, gint vsize, gint rowstride) {
  // inplace convert an RGB24 image to BGR24
  register int size=rowstride*vsize;
  register int i,j;
  register unsigned char  byte;

  for (i=0;i<size;i+=rowstride) {
    for (j=0;j<hsize;j+=3) {
      byte=pixels[i+j];
      pixels[i+j]=pixels[i+j+2];
      pixels[i+j+2]=byte;
    }
  }
}
