// effects.c
// LiVES (lives-exe)
// (c) G. Finch 2003
// Released under the GPL 2.0
// see file ../COPYING for licensing details

#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "main.h"
#include "effects.h"
#include "paramwindow.h"
#include "support.h"

//////////// Effects ////////////////


#ifdef HAVE_YUV4MPEG
extern GdkPixbuf *gdk_pixbuf_new_from_yuv4m (void *lives_yuv4m, GError **error);
#endif

#ifdef USE_LIVIDO
extern GdkPixbuf *gdk_pixbuf_new_from_generator (void *lives_yuv4m, GError **error);
extern void livido_swap_fg_bg (void);
#endif

//// some pixel utils


static inline void 
make_black (guchar *pixel) {
  pixel[0]=pixel[1]=pixel[2]=(guchar)0;
}

static inline void 
make_white (guchar *pixel) {
  pixel[0]=pixel[1]=pixel[2]=(guchar)255;
}


void 
nine_fill (guchar *new_data, gint i, gint j, guchar *old_data, gint x, gint y, gint rowstride) {
  // fill nine pixels with the centre colour
  new_data[i-rowstride+j-3]=new_data[i-rowstride]=new_data[i-rowstride+3]=new_data[i+j-3]=new_data[i+j]=new_data[i+j+3]=new_data[i+rowstride+j-3]=new_data[i+rowstride+j]=new_data[i+rowstride+j+3]=old_data[x+y];
  new_data[i-rowstride+j-2]=new_data[i-rowstride+1]=new_data[i-rowstride+4]=new_data[i+j-2]=new_data[i+j+1]=new_data[i+j+4]=new_data[i+rowstride+j-2]=new_data[i+rowstride+j+1]=new_data[i+rowstride+j+4]=old_data[x+y+1];
  new_data[i-rowstride+j-1]=new_data[i-rowstride+2]=new_data[i-rowstride+5]=new_data[i+j-1]=new_data[i+j+2]=new_data[i+j+5]=new_data[i+rowstride+j-1]=new_data[i+rowstride+j+2]=new_data[i+rowstride+j+5]=old_data[x+y+2];
}


///////////////////////////////////////////////////


// Rendered effects




gboolean  
do_effect(lives_rendered_effect_t *rfx) {
  // returns FALSE if the user cancelled
  gint oundo_start=cfile->undo_start;
  gint oundo_end=cfile->undo_start;
  gchar effectstring[128];
  gdouble old_pb_fps=cfile->pb_fps;

  gchar *text;
  gchar *fxcommand=NULL;
  gint current_file=mainw->current_file;

  if (rfx->num_in_channels!=2) {
    if (rfx->props&RFX_PROPS_MAY_RESIZE) {
      cfile->progress_start=cfile->undo_start=1;
      cfile->progress_end=cfile->undo_end=cfile->frames;
    }
    else {
      cfile->progress_start=cfile->undo_start=cfile->start;
      cfile->progress_end=cfile->undo_end=cfile->end;
    }
  }
  if (!mainw->internal_messaging) {
    gchar *pdefault;
    if (rfx->num_in_channels==2) {
      // transition has a few extra bits
      pdefault=g_strdup_printf ("%s %d %d %d %d %d %d \"%s/%s\"",cfile->handle,rfx->status,cfile->progress_start,cfile->progress_end,cfile->hsize,cfile->vsize,clipboard->start,prefs->tmpdir,clipboard->handle);
    }
    else {
      pdefault=g_strdup_printf ("%s %d %d %d %d %d",cfile->handle,rfx->status,cfile->progress_start,cfile->progress_end,cfile->hsize,cfile->vsize);
    }
    // and append params
    fxcommand=g_strconcat ("smogrify fxrender_",rfx->name," ", pdefault, param_marshall (rfx, FALSE), NULL);
    g_free (pdefault);
  }
  if (rfx->props&RFX_PROPS_MAY_RESIZE) {
    text=g_strdup_printf(_ ("%s all frames..."),_ (rfx->action_desc));
  }
  else if (rfx->num_in_channels==2) {
    text=g_strdup_printf(_ ("%s clipboard into frames %d to %d..."),_ (rfx->action_desc),cfile->progress_start,cfile->progress_end);
    } else {
    text=g_strdup_printf(_ ("%s frames %d to %d..."),_ (rfx->action_desc),cfile->start,cfile->end);
  }
  d_print(text);
  g_free(text);

  cfile->redoable=cfile->undoable=FALSE;
  gtk_widget_set_sensitive (mainw->redo, FALSE);
  gtk_widget_set_sensitive (mainw->undo, FALSE);

  cfile->undo_action=UNDO_EFFECT;

  if (rfx->props&RFX_PROPS_MAY_RESIZE) {
    cfile->ohsize=cfile->hsize;
    cfile->ovsize=cfile->vsize;
    mainw->resizing=TRUE;
    cfile->nokeep=TRUE;
  }

  mainw->effects_paused=FALSE;
  unlink(cfile->info_file);

  // 'play' as fast as we possibly can
  cfile->pb_fps=1000000;

  if (!mainw->internal_messaging) {
    system(fxcommand);
    g_free (fxcommand);
  }
  else {
#ifdef USE_LIVIDO
    if (mainw->num_tr_applied>0) {
#else
    if (mainw->rte&EFFECT_BLEND) {
#endif
      mainw->files[mainw->blend_file]->frameno=mainw->files[mainw->blend_file]->start-1;
      // sigh, start off forwards i suppose...:-)
      mainw->blend_file_step=1;
    }
  }

  if (rfx->num_in_channels==2) {
    g_snprintf (effectstring,128,_ ("%s clipboard with selection"),_ (rfx->action_desc));
  }
  else {
    g_snprintf (effectstring,128,_ ("%s frames %d to %d"),_ (rfx->action_desc),cfile->undo_start,cfile->undo_end);
  }
  if (do_progress_dialog(TRUE,TRUE,effectstring)==FALSE||mainw->error) {
    if (mainw->error) {
      do_error_dialog (mainw->msg);
      d_print (_ ("failed.\n"));
    }
    cfile->undo_start=oundo_start;
    cfile->undo_end=oundo_end;
    cfile->pb_fps=old_pb_fps;
    mainw->internal_messaging=FALSE;
    mainw->resizing=FALSE;
    cfile->nokeep=FALSE;
    return FALSE;
  }

  mainw->resizing=FALSE;
  cfile->nokeep=FALSE;
  cfile->changed=TRUE;
  gtk_widget_set_sensitive (mainw->undo, TRUE);
  cfile->undoable=TRUE;
  cfile->pb_fps=old_pb_fps;
  mainw->internal_messaging=FALSE;
  gtk_widget_set_sensitive (mainw->select_last, TRUE);
  set_undoable (_ (rfx->menu_text),TRUE);

  if (rfx->props&RFX_PROPS_MAY_RESIZE) {
    // get new frame size
    gchar **array;
    gint numtok=get_token_count (mainw->msg,'|');
    
    if (numtok>1) {
      array=g_strsplit(mainw->msg,"|",numtok);
      // [0] is "completed"
      cfile->hsize=atoi (array[1]);
      cfile->vsize=atoi (array[2]);
      g_strfreev(array);
    }
    if (cfile->hsize==cfile->ohsize&&cfile->vsize==cfile->ovsize) cfile->undo_action=UNDO_EFFECT;
    else {
      cfile->undo_action=UNDO_RESIZABLE;
      mainw->current_file=0; // force resize on switch_to_file()
    }
  }

  if (rfx->num_in_channels!=2) {
    switch_to_file (mainw->current_file,current_file);
    d_print(_ ("done\n"));
  }
  return TRUE;
}



// realtime fx


  

gint 
realfx_progress (gboolean reset) {
  static int i;
  GError *error=NULL;
  gchar fname[256];
  gchar oname[256];
  GdkPixbuf *image;
  gchar *com;

  // this is called periodically from do_processing_dialog for internal effects

  if (reset==TRUE) {
    i=cfile->start;
    clear_mainw_msg();
    return 1;
  }

  if (mainw->effects_paused) return 1;

  // sig_progress...
  g_snprintf (mainw->msg,256,"%d",i);
  // load, effect, save frame

  g_snprintf(fname,256,"%s/%s/%08d.%s",prefs->tmpdir,cfile->handle,i,prefs->image_ext);
  g_snprintf(oname,256,"%s/%s/%08d.mgk",prefs->tmpdir,cfile->handle,i);

  image=gdk_pixbuf_new_from_file(fname,&error);

  if (image==NULL) {
    g_snprintf (mainw->msg,256,"error|missing image %d",i);
    return 1;
  }

  on_rte_apply (mainw->rte,image);

  if (!strcmp (prefs->image_ext,"jpg")) {
    gdk_pixbuf_save (image, oname, "jpeg", &error,"quality", "100", NULL);
  }
  else if (!strcmp (prefs->image_ext,"png")) {
    gdk_pixbuf_save (image, oname, "png", &error, NULL);
  }
  else {
    //gdk_pixbuf_save_to_callback(...);
  }
  gdk_pixbuf_unref (image);
  
  if (++i>cfile->end) {
    com=g_strdup_printf ("smogrify mv_mgk %s %d %d",cfile->handle,cfile->start,cfile->end);
    system (com);
    g_free (com);
    mainw->internal_messaging=FALSE;
  }
  return 1;
}



void
on_realfx_activate                   (GtkMenuItem     *menuitem,
				      gpointer         user_data)
{
  mainw->internal_messaging=TRUE;
  mainw->progress_fn=&realfx_progress;
  mainw->progress_fn (TRUE);
  invalidate_pixel_buffers(guess_rowstride (mainw->current_file));
  do_effect (&mainw->rendered_fx[0]);
}


gint guess_rowstride (gint idx) {
  if (mainw->files[idx]->type==CLIP_TYPE_DISK&&mainw->image272!=NULL) {
    return gdk_pixbuf_get_rowstride(GDK_PIXBUF (gtk_image_get_pixbuf (GTK_IMAGE (mainw->image272))))*3;
  }
  else return cfile->hsize*3;
}

void 
on_rte_apply (gint eff_mask, GdkPixbuf *pixbuf) {
// realtime effects

  int rowstride;
  int height;

#ifdef USE_LIVIDO
  guchar *in_pixel_data[2];
#else
  guchar *new_pixel_data;
  guchar byte;
  int i,j;

  gint size=height*rowstride;
#endif
  int width3=gdk_pixbuf_get_width(pixbuf)*3;

  GdkPixbuf *blendr=NULL;

  if (pixbuf==NULL) return;

  if (mainw->foreign) return;

  rowstride=gdk_pixbuf_get_rowstride(pixbuf);
  height=gdk_pixbuf_get_height(pixbuf);


  // assume RGB 24 bit
  if (gdk_pixbuf_get_bits_per_sample(pixbuf)!=8||gdk_pixbuf_get_n_channels(pixbuf)!=3) return;

#ifdef USE_LIVIDO
  in_pixel_data[0]=gdk_pixbuf_get_pixels(pixbuf);
  if (mainw->num_tr_applied&&mainw->blend_file!=mainw->current_file&&mainw->blend_file!=-1) {
    blendr=get_blend_frame(width3/3,height);
  }
  if (blendr!=NULL) {
    if (gdk_pixbuf_get_rowstride (blendr)!=width3) {
      // bleeeuuughhh choke barf
      in_pixel_data[1]=NULL;
    }
    else in_pixel_data[1]=gdk_pixbuf_get_pixels(blendr);
  }
  else in_pixel_data[1]=NULL;
  livido_apply_effects(height,rowstride,in_pixel_data,EXTP_RGB24LE);
  if (blendr!=NULL) gdk_pixbuf_unref(blendr);
  return;
#else
  new_pixel_data=gdk_pixbuf_get_pixels(pixbuf);

  // real time effects
  if (mainw->rte&EFFECT_MIRRORX) {
    for (i=0;i<size;i+=rowstride) {
      for (j=0;j<width3/2;j++) {
	new_pixel_data[i+width3-1-j]=new_pixel_data[i+j];
      }
    }
  }
    
  if (mainw->rte&EFFECT_MIRRORY) {
    gdouble mratio=1.;
    if (mainw->rte_keys==EFFECT_MIRRORY) {
      mratio=256./(0xff^mainw->blend_factor);
    }
    for (i=0;i<size*mratio/(mratio+1);i+=rowstride) {
      for (j=0;j<rowstride;j++) {
	new_pixel_data[size-(int)(i/mratio/rowstride+1)*rowstride+j]=new_pixel_data[i+j];
      }
    }
  }

  if (mainw->rte&EFFECT_ALPHA) {
    if (mainw->old_pixel_data==NULL) {
      mainw->old_pixel_data=g_malloc0(size);
    }
    for (i=0;i<size;i+=rowstride) {
      for (j=0;j<width3;j++) {
	byte=new_pixel_data[i+j];
	new_pixel_data[i+j]=((char)byte+(char)mainw->old_pixel_data[i+j])/2;
	mainw->old_pixel_data[i+j]=byte;
      }
    }
  }
    
  if (mainw->rte&EFFECT_NOISE) {
    for (i=0;i<size;i+=rowstride) {
      for (j=0;j<width3;j++) {
	new_pixel_data[i+j]+=((fastrand()&0xF8000000)>>27)-16;
      }
    }
  }
    
  if (mainw->rte&EFFECT_NEGATE) {
    for (i=0;i<size;i+=rowstride) {
      for (j=0;j<width3;j++) {
	new_pixel_data[i+j]=new_pixel_data[i+j]^0xFF;
      }
    }
  }
    
  if (mainw->rte&EFFECT_EDGE) {
    guchar myluma;
    guchar threshold=15;
    gint nbr;
    int k,l;

    if (mainw->old_pixel_data==NULL) {
      mainw->old_pixel_data=g_malloc0(size);
    }
    memcpy(mainw->old_pixel_data,new_pixel_data,size);
    for (i=rowstride;i<size-rowstride;i+=rowstride) {
      for (j=3;j<rowstride-3;j+=3) {
	myluma=calc_luma (&mainw->old_pixel_data[i+j]);
	nbr=0;
	for (k=i-rowstride;k<=i+rowstride;k+=rowstride) {
	  for (l=j-3;l<=j+3;l+=3) {
	    if ((k!=i||j!=l)&&ABS (calc_luma (&mainw->old_pixel_data[k+l])-myluma)>threshold) nbr++;
	  }
	}
	if (nbr<2||nbr>5) {
	  nine_fill (new_pixel_data,i,j,mainw->old_pixel_data,i,j,rowstride);
	}
	else {
	  if (myluma<80) {
	    new_pixel_data[i+j]=mainw->old_pixel_data[i+j]/2;
	    new_pixel_data[i+j+1]=mainw->old_pixel_data[i+j+1]/2;
	    new_pixel_data[i+j+2]=mainw->old_pixel_data[i+j+2]/2;
	  }
	  else{
	    if (myluma>170) {
	      make_white (&new_pixel_data[i+j]);
	    }
	  }
	}
      }
    }
  }
  
  if (mainw->rte&EFFECT_POSTER) {
    for (i=0;i<size;i+=rowstride) {
      for (j=0;j<width3;j++) {
	new_pixel_data[i+j]=new_pixel_data[i+j]>128?255:0;
      }
    }
  }
    
  if (mainw->rte&EFFECT_REMOVE_BG) {
    // TODO - use HSI palette
      guchar luma;
      guchar av_luma;
      gint cpos;
      guchar luma_threshold;
      luma_threshold=(mainw->rte_keys==EFFECT_REMOVE_BG?mainw->blend_factor:235);

      if (mainw->av_luma_data==NULL) {
	mainw->av_luma_data=(guchar *)g_malloc0(cfile->hsize*cfile->vsize*3*sizint);
	mainw->av_count=0;
      }
      
      mainw->av_count++;
      
      for (i=0;i<size;i+=rowstride) {
	cpos=i/rowstride*width3;
	for (j=0;j<width3;j+=3) {
	  luma=calc_luma (&new_pixel_data[i+j]);
	  if (mainw->av_count==1) {
	    mainw->av_luma_data[cpos+j/3]=luma;
	  }
	  av_luma=(guint)((gdouble)(mainw->av_luma_data[cpos+j/3]*mainw->av_count)+(gdouble)luma)/((gdouble)mainw->av_count+1.);
	  mainw->av_luma_data[cpos+j/3]=av_luma;
	  if (ABS (luma-av_luma)>(luma_threshold)) {
	    switch (mainw->rte_mode*(mainw->rte_keys==EFFECT_REMOVE_BG)) {
	    case 1:
	      // fire-ish effect
	      new_pixel_data[i+j]=((guchar)(((fastrand()>>24)+(new_pixel_data[i+j+1]=((guchar) (fastrand()>>25))))&0xff));     //R & G
	      new_pixel_data[i+j+2]=(guchar)0;                       //B
	      break;
	    //
	    case 2:
	      break;
	    case 0:
	      // make moving things black
	      make_black (&new_pixel_data[i+j]);
	    }
	  }
	  else {
	    // blue glow
	    if (mainw->rte_mode==2&&mainw->rte_keys==EFFECT_REMOVE_BG) {
	      new_pixel_data[i+j]=new_pixel_data[i+j+1]=((guchar)(fastrand()>>24));                                                 //R&G
	      new_pixel_data[i+j+2]=(guchar)255;  //B
	    }}}}
  }

  if (mainw->rte&EFFECT_BLEND) {
    if (mainw->current_file!=mainw->blend_file&&mainw->blend_file!=-1) {
      blendr=get_blend_frame(width3/3,height);
      if (blendr!=NULL) {
	guchar *blend_data=gdk_pixbuf_get_pixels (blendr);
	for (i=0;i<size;i+=rowstride) {
	  for (j=0;j<rowstride;j+=3) {
	    // avg
	    switch (mainw->rte_mode*(mainw->rte_keys==EFFECT_BLEND)) {
	    case 0:
	      // chroma blend
	      new_pixel_data[i+j]=(guchar)((new_pixel_data[i+j]*mainw->blend_factor+blend_data[i+j]*(0xFF^mainw->blend_factor))>>8);
	      new_pixel_data[i+j+1]=(guchar)((new_pixel_data[i+j+1]*mainw->blend_factor+blend_data[i+j+1]*(0xFF^mainw->blend_factor))>>8);
	      new_pixel_data[i+j+2]=(guchar)((new_pixel_data[i+j+2]*mainw->blend_factor+blend_data[i+j+2]*(0xFF^mainw->blend_factor))>>8);
	      break;
	    case 1:
	      // threshold luma blend (background)
	      if (calc_luma (&new_pixel_data[i+j])<(mainw->blend_factor)) {
		new_pixel_data[i+j]=blend_data[i+j];
		new_pixel_data[i+j+1]=blend_data[i+j+1];
		new_pixel_data[i+j+2]=blend_data[i+j+2];
	      }
	      break;	
	    case 2:
	      // threshold luma blend (foreground)
	      if (calc_luma (&blend_data[i+j])>(mainw->blend_factor)) {
		new_pixel_data[i+j]=blend_data[i+j];
		new_pixel_data[i+j+1]=blend_data[i+j+1];
		new_pixel_data[i+j+2]=blend_data[i+j+2];
	      }
	      break;
	    }
	  }
	}
	gdk_pixbuf_unref(blendr);
      }
    }
  }
#endif // USE_LIVIDO
}


void 
invalidate_pixel_buffers (gint rowstride) {
  // invalidate all pixel buffers
  // this should call [stop()]/start() for all fx plugins

  // we call this when we start playing, and when we switch to a different sized clip

#ifdef USE_LIVIDO
  if (mainw->blend_file>-1&&mainw->num_tr_applied&&(mainw->files[mainw->blend_file]==NULL||(mainw->files[mainw->blend_file]->clip_type==CLIP_TYPE_DISK&&(!mainw->files[mainw->blend_file]->frames||!mainw->files[mainw->blend_file]->is_loaded)))) {
    // invalid blend file
    mainw->blend_file=mainw->current_file;
  }
#else
  if (mainw->blend_file>-1&&mainw->rte&EFFECT_BLEND&&(mainw->files[mainw->blend_file]==NULL||!mainw->files[mainw->blend_file]->frames||!mainw->files[mainw->blend_file]->is_loaded)) {
    // invalid blend file
    mainw->blend_file=mainw->current_file;
  }
#endif

  if (!(mainw->uyvy_frame==NULL)) {
    g_free(mainw->uyvy_frame);
    mainw->uyvy_frame=NULL;
  }
  if (!(mainw->yuv_frame==NULL)) {
    g_free(mainw->yuv_frame);
    mainw->yuv_frame=NULL;
  }

#ifdef USE_LIVIDO
  livido_init_resize(rowstride);
  return;
#endif


  if (!(mainw->old_pixel_data==NULL)) {
    g_free(mainw->old_pixel_data);
    mainw->old_pixel_data=NULL;
  }

  if (!(mainw->av_luma_data==NULL)) {
    g_free(mainw->av_luma_data);
    mainw->av_luma_data=NULL;
  }
}

GdkPixbuf *get_blend_frame(gint hsize, gint vsize) {
  GError *error=NULL;
  GdkPixbuf *blendr=NULL, *blendp=NULL;
  file *blend_file;

  if (mainw->blend_file==-1||mainw->files[mainw->blend_file]==NULL) return NULL;
  blend_file=mainw->files[mainw->blend_file];
  
  blend_file->frameno+=mainw->blend_file_step;
  
  if (blend_file->frameno>blend_file->end) {
    if (mainw->ping_pong) {
      blend_file->frameno=blend_file->end;
      mainw->blend_file_step=-1;
    }
    else {
      blend_file->frameno=blend_file->start;
      mainw->blend_file_step=1;
    }
  }
  else if (blend_file->frameno<blend_file->start) {
    if (mainw->ping_pong) {
      blend_file->frameno=blend_file->start;
      mainw->blend_file_step=1;
    }
    else {
      blend_file->frameno=blend_file->end;
      mainw->blend_file_step=-1;
    }
  }
  
  switch (blend_file->clip_type) {
  case CLIP_TYPE_DISK:
    blendp=gdk_pixbuf_new_from_file(g_strdup_printf ("%s/%s/%08d.%s",prefs->tmpdir,mainw->files[mainw->blend_file]->handle,mainw->files[mainw->blend_file]->frameno,prefs->image_ext),&error);
    break;
#ifdef HAVE_YUV4MPEG
  case CLIP_TYPE_YUV4MPEG:
    blendp=gdk_pixbuf_new_from_yuv4m (blend_file->ext_src,&error);
    break;
#endif
#ifdef USE_LIVIDO
  case CLIP_TYPE_GENERATOR:
    // special handling for clips where host controls size
    if (mainw->reinit_effects) {
      // this is a different frame size from the last
      mainw->reinit_effects=FALSE;
      
      // only valid for RGB24 - TODO
      invalidate_pixel_buffers(hsize*3);
    }
    blendp=gdk_pixbuf_new_from_generator (blend_file->ext_src,&error);
    break;
#endif

  }
  if (error==NULL&&blendp!=NULL) {
    blendr=gdk_pixbuf_scale_simple(blendp,hsize,vsize,GDK_INTERP_NEAREST);
    gdk_pixbuf_unref(blendp);
  }
  else {
    if (error!=NULL) g_error_free(error); // returns NULL on error
  }
  return blendr;
}


////////////////////////////////////////////////////////////////////
// keypresses


gboolean rte_on_off_callback (GtkAccelGroup *group, GObject *obj, guint keyval, GdkModifierType mod, gpointer user_data)
{
// this is the callback which happens when a rte is keyed
  gint key=GPOINTER_TO_INT(user_data);
  guint new_rte=1<<(key-1);

  if (key==EFFECT_NONE) {
    // switch up/down keys to default (fps change)
#ifdef USE_LIVIDO
    livido_deinit_all();
#else
    mainw->rte=EFFECT_NONE;
#endif
    mainw->rte_keys=mainw->last_grabable_effect=EFFECT_NONE;
    mainw->num_tr_applied=0;
  }
  else {
    mainw->rte^=new_rte;
    if (mainw->rte&new_rte) {
      // switch is ON
#ifdef USE_LIVIDO
      if (!(livido_init_effect(key-1,mainw->frame_image==NULL?0:gdk_pixbuf_get_rowstride(mainw->frame_image)))) {
	// ran out of instance slots, no effect assigned, or some other error
	mainw->rte^=new_rte;
      }
      return TRUE;
#else
      switch (new_rte) {
      case EFFECT_REMOVE_BG:
	if (mainw->av_luma_data!=NULL) {
	  g_free (mainw->av_luma_data);
	  mainw->av_luma_data=NULL;
	}
	// allow switch of up/down/x keys
	mainw->rte_mode=0;
	mainw->rte_mode_max=2;
	mainw->last_grabable_effect=EFFECT_REMOVE_BG;
	break;
      case EFFECT_BLEND:
	mainw->blend_file=mainw->current_file;
	mainw->blend_file_step=cfile->pb_fps<0?-1:1;
	
	// allow switch of up/down/x keys
	mainw->rte_mode=0;
	mainw->rte_mode_max=2;
	mainw->last_grabable_effect=EFFECT_BLEND;
	mainw->num_tr_applied++;
	break;
      case EFFECT_MIRRORY:
	// allow switch of up/down/x keys
	mainw->rte_mode=0;
	mainw->rte_mode_max=0;
	mainw->last_grabable_effect=EFFECT_MIRRORY;
	break;
      }
#endif
    }
    else {
      // effect is OFF
#ifdef USE_LIVIDO
      livido_deinit_effect(key-1);
#else
      switch (new_rte) {
      case EFFECT_BLEND:
	if (mainw->rte_keys==EFFECT_BLEND) {
	  mainw->num_tr_applied--;
	}
      }
#endif
    }
  }
  if (mainw->current_file>0&&cfile->play_paused) {
    load_frame_image (cfile->frameno,cfile->last_frameno);
  }
  return TRUE;
}


gboolean grabkeys_callback (GtkAccelGroup *group, GObject *obj, guint keyval, GdkModifierType mod, gpointer user_data) {
  // assign the keys to the last key-grabable effect 
  mainw->rte_keys=mainw->last_grabable_effect;
#ifdef USE_LIVIDO
  mainw->blend_factor=livido_get_blend_factor(mainw->rte_keys);
#else
  mainw->blend_factor=50; // TODO - rename this
#endif
  return TRUE;
}


gboolean rtemode_callback (GtkAccelGroup *group, GObject *obj, guint keyval, GdkModifierType mod, gpointer user_data) {
  if (mainw->rte_keys==-1) return TRUE;
#ifdef USE_LIVIDO
  livido_key_setmode(0,-1);
#else
  mainw->rte_mode=((mainw->rte_mode==mainw->rte_mode_max)?0:++mainw->rte_mode);
#endif
  return TRUE;
}



gboolean swap_fg_bg_callback (GtkAccelGroup *group, GObject *obj, guint keyval, GdkModifierType mod, gpointer user_data) {
  gint old_file=mainw->current_file;
  gint blend_step=mainw->blend_file_step;

  if (mainw->playing_file<0||!mainw->num_tr_applied||mainw->noswitch||mainw->blend_file==-1||mainw->blend_file==mainw->current_file||mainw->preview||mainw->noswitch) {
    return TRUE;
  }
  mainw->blend_file_step=cfile->pb_fps>0?1:-1;
  do_quick_switch (mainw->blend_file);
  if ((cfile->pb_fps>0&&blend_step<0)||(cfile->pb_fps<0&&blend_step>0)) cfile->pb_fps*=-1;
  mainw->blend_file=old_file;

#ifdef USE_LIVIDO
  livido_swap_fg_bg();
#endif
  return TRUE;

  // TODO - something with fps matching
  // **TODO - for livido, invert all transition parameters for any active effects
}






//////////////////////////////////////////////////////////////



