// main.h
// LiVES
// (c) G. Finch (salsaman@xs4all.nl)
// see file ../COPYING for licensing details


// begin legal warning
/*
    NO WARRANTY

      BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
      FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
      PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
      OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
      PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
REPAIR OR CORRECTION.

      IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
      WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
      REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,
      INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING
      OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED
TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY
YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER
      PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGES.
*/
// end legal warning


// Have fun, and let's fight for Free Speech, Open Media and True Creativity !
// - Salsaman


#ifndef HAS_MAIN_H
#define HAS_MAIN_H

#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>

// don't change this unless the backend is changed as well
// i.e. $GUI_BOOTSTRAP_FILE in smogrify
#define BOOTSTRAP_NAME "/tmp/.smogrify"

#define MAX_FILES 512

// this must match AC_PREFIX_DEFAULT in configure.in
// TODO - when lives-plugins is a separate package, use pkg-config to get PREFIX and remove PREFIX_DEFAULT
#ifndef PREFIX_DEFAULT
#define PREFIX_DEFAULT "/usr"
#endif

// if --prefix= was not set, this is set to "NONE"
#ifndef PREFIX
#define PREFIX PREFIX_DEFAULT
#endif

#define TEST_SCALE   // TODO - check what this does
#define USE_EXT_RENDER // test external rendered effects

#define DOC_DIR "/share/doc/LiVES/"

#define THEME_DIR "/share/lives/themes/"
#define PLUGIN_DIR "/share/lives/plugins/"
#define ICON_DIR "/share/lives/icons/"

// insert methods
#define AFTER_END 1
#define BEFORE_START 0

// LiVES will show a warning if this (MBytes) is exceeded on load
// (can be overridden in prefs)
#define WARN_FILE_SIZE 500

// maximum fps we will allow (gdouble)
// TODO - make pref
#define FPS_MAX 200.

#define MAX_RECORD_FRAMES 100000

#define ENABLE_DVD_GRAB

// >test for YUV4MPEG input
// : if this is enabled, manually copy Makefile.yuv4m.am to Makefile.am
// and re-run automake (also requires mjpegtools-devel)
#define YUV4MPEG_TEST

#define FP_BITS 16 // max fp bits [apparently 16 is faster]


#define USE_LIVIDO

#define AUTO_RESAMPLE_RESIZE

// events - we have a relative time and a value; each event type currently has its own array of these
// reltime is only used to quantise events; after that it is discarded...heh..

// TODO - change long doubles to just long
typedef struct {
  gint value;
  long double reltime;
} event;

typedef struct {
  GtkWidget *processing;
  GtkWidget *progressbar;
  GtkWidget *label;
  GtkWidget *label2;
  GtkWidget *stop_button;
  GtkWidget *pause_button;
  GtkWidget *preview_button;
  GtkWidget *cancel_button;
  guint frames_done;
} process;


typedef struct {                // corresponds to one clip in the GUI
  // cfile - one clip
  //
  // basic info (saved during backup)
  gint bpp;
  gdouble fps;
  gint hsize;
  gint vsize;
  gint arps; // audio sample rate
  guint signed_endian; // bitfield

  #define AFORM_UNSIGNED 1
  #define AFORM_BIG_ENDIAN 2
  #define AFORM_UNKNOWN 65536

  gint arate; // audio playback rate
  gdouble reserved;    // no longer used
  gint achans;
  gint asampsize;
  /////////////////
  gint frames;
  gchar title[256];  // max length seems to be 250 for mencoder
  gchar author[256];
  gchar comment[256];
  ////////////////

#if defined USE_ALSA
  static snd_pcm_t *playback_handle;
#endif

  // extended info (not saved)

  gint rowstride;

  // the processing window
  process *proc_ptr;

  gchar handle[256];
  gint ohsize;
  gint ovsize;
  gdouble f_size;
  guint afilesize;
  gint old_frames; // for deordering, etc.
  gchar file_name[256];
  gchar info_file[256];
  gchar name[256];  // the display name
  gchar type[40];
  gint start;
  gint end;
  gint insert_start;
  gint insert_end;
  gint progress_start;
  gint progress_end;
  gboolean changed;
  GtkWidget *menuentry;
  gboolean orig_file_name;
  gboolean was_renamed;;
  gboolean is_untitled;
  gdouble pb_fps;
  gdouble freeze_fps;
  gboolean play_paused;
  
  //opening/restoring status
  gboolean opening;
  gboolean opening_audio;
  gboolean opening_only_audio;
  gboolean opening_loc;
  gboolean restoring;
  gboolean is_loaded;  // should we continue loading if we come back to this clip

  // don't show preview/pause buttons on processing
  gboolean nopreview;

  // don't show the 'keep' button - e.g. for operations which resize frames
  gboolean nokeep;

  // various times; total time is calculated as the longest of video, laudio and raudio
  gdouble total_time;
  gdouble video_time;
  gdouble laudio_time;
  gdouble raudio_time;
  gdouble pointer_time;

  // current and last played index frames for internal player
  gint frameno;
  gint last_frameno;

  // performance play/record
  gint perf_counter;  // <0 = playing, >0 = recording, 0 = no events

  // these two mark the event envelope
  gint perf_start;
  gint perf_end;

  // these are employed when we resample with variable fps
  gint ev_fps_start;
  gint ev_fps_end;
  long double ev_origticks;

  // these events help with automating performance playback
  gshort event_pending;
# define EVENT_AUTO -2
# define EVENT_NONE -1

  // events stored in events arrays
# define EVENT_FRAMES 0
# define EVENT_FPS 1

# define LAST_EVENT EVENT_FPS

# define CREATE_EV_FRAMES (guint64) (1<<EVENT_FRAMES)
# define CREATE_EV_FPS (guint64) (1<<EVENT_FPS)

  // events
  event *events[LAST_EVENT+1];

  //undo
  gshort undo_action;
#define UNDO_NONE 0
#define UNDO_EFFECT 1
#define UNDO_RESIZABLE 2
#define UNDO_MERGE 3
#define UNDO_REORDER 4
#define UNDO_RESAMPLE 5
#define UNDO_TRIM_AUDIO 6
#define UNDO_CHANGE_SPEED 7
#define UNDO_AUDIO_RESAMPLE 8
#define UNDO_APPEND_AUDIO 9
#define UNDO_INSERT 10
#define UNDO_CUT 11
#define UNDO_DELETE 12
#define UNDO_DELETE_AUDIO 13
#define UNDO_INSERT_SILENCE 14
#define UNDO_NEW_AUDIO 15
#define UNDO_ATOMIC_RESAMPLE_RESIZE 16

  gint undo_start;
  gint undo_end;
  gchar undo_text[32];
  gchar redo_text[32];
  gboolean undoable;
  gboolean redoable;

  // used for storing undo values
  gint undo1_int;
  gint undo2_int;
  gint undo3_int;
  gint undo4_int;
  guint undo1_uint;
  gdouble undo1_dbl;
  gdouble undo2_dbl;
  gboolean undo1_boolean;
  gboolean undo2_boolean;
  gboolean undo3_boolean;

  gint undo_arate; // audio playback rate
  guint undo_signed_endian;
  gint undo_achans;
  gint undo_asampsize;
  gint undo_arps; // audio sample rate

  gshort clip_type;
#define CLIP_TYPE_DISK 0
#define CLIP_TYPE_YUV4MPEG 1
#define CLIP_TYPE_GENERATOR 2
  void *ext_src; // points to opaque source for non-disk types
  gint disk_frames; // count for recording to disk

  gboolean ratio_fps; // if the fps was set by a ratio

  // TODO - change to lives_clip_t
} file;

size_t sizhead; // length of basic details

typedef struct {
  // the following can be assumed TRUE, they are checked on startup
  gboolean has_smogrify;
  gboolean smog_version_correct;
  gboolean can_read_from_config;
  gboolean can_write_to_config;
  gboolean can_write_to_tmp;
  gboolean can_write_to_tempdir;

  // the following may need checking before use
  gboolean has_xmms;
  gboolean has_sox;
  gboolean has_mplayer;
  gboolean has_convert;
  gboolean has_composite;
  gboolean has_cdda2wav;
  gboolean has_midistartstop;

  // home directory - default location for config file
  gchar home_dir[256];

  // used for returning startup messages from the backend
  gchar startup_msg[256];

  // plugins
  gboolean has_encoder_plugins;

  // extensions
  gboolean has_veejay;

} capability;


typedef struct {
  gchar name[51];
  gshort audio_codec;
  // must match with the "acodec" GList in interface.c
#define AUDIO_CODEC_MP3 0
#define AUDIO_CODEC_PCM 1
#define AUDIO_CODEC_MP2 2
#define AUDIO_CODEC_VORBIS 3
#define AUDIO_CODEC_MAX 31
  //
#define AUDIO_CODEC_NONE 32
#define AUDIO_CODEC_UNKNOWN 33

  guint capabilities;
#define CAN_ENCODE_WITH_FRAME_CHANGES 1<<0
#define CAN_ENCODE_WITH_FPS_CHANGES 1<<1
#define CAN_ENCODE_PNG 1<<2
#define ENCODER_NON_NATIVE 1<<3

  // current output format
  gchar of_name[51];
  gchar of_desc[128];
  gint of_allowed_acodecs;
  gchar of_restrict[1024];
  gchar of_def_ext[16];
}
_encoder;



typedef struct {
  gchar bgcolour[256];
  gboolean stop_screensaver;
  gboolean open_maximised;
  gchar theme[64];  // the theme name

  gshort video_player;
#define VID_PLAYER_INT 1
#define VID_PLAYER_MPLAYER 2
#define VID_PLAYER_VEEJAY 3

  gshort pb_quality;
#define PB_QUALITY_LOW 1
#define PB_QUALITY_MED 2  // default
#define PB_QUALITY_HIGH 3

  _encoder encoder;

  gshort audio_player;
#define AUD_PLAYER_SOX 1
#define AUD_PLAYER_MPLAYER 2

  // frame quantisation type
  gshort q_type;
#define Q_FILL 1
#define Q_SMOOTH 1

  gchar tmpdir[256];

  // TODO - also use cur_vid_load_dir, etc.
  gchar def_vid_load_dir[256];
  gchar def_vid_save_dir[256];
  gchar def_audio_dir[256];
  gchar def_image_dir[256];
  gchar def_proj_dir[256];

  gchar prefix_dir[256];
  gchar image_ext[16];

  unsigned int warning_mask;
  // if these bits are set, we do not show the warning
#define WARN_MASK_FPS 1<<0
#define WARN_MASK_FSIZE 1<<1
#define WARN_MASK_SAVE_QUALITY 1<<2
#define WARN_MASK_SAVE_SET 1<<3

  gchar effect_command[256];
  gchar video_play_command[256];
  gchar video_open_command[256];
  gchar audio_play_command[256];
  gchar cdplay_device[256];
  gdouble default_fps;
  gint bar_height;
  gboolean pause_effect_during_preview;
  gboolean open_decorated;
  int sleep_time;
  gboolean pause_during_pb;
  gboolean fileselmax;
  gboolean show_recent;
  gint warn_file_size;
  gboolean pause_xmms;
  gboolean midisynch;
  gint dl_bandwidth;
  gboolean conserve_space;
  gboolean ins_resample;
  gboolean show_tool;
  gshort sepwin_type;
  gboolean show_player_stats;
  gboolean show_framecount;
  gint default_height;
  gint default_width;
  gboolean loop_recording;
  gboolean discard_tv;
  gboolean save_directories;
  guint max_record_frames;
  gboolean safer_preview;
  gboolean is_pal; // TRUE == PAL, FALSE == NTSC
  gboolean debug_encoders;
  gboolean no_bandwidth;
  gboolean osc_udp_started;
  guint osc_udp_port;
#define OSC_DEFAULT_UDP_PORT 9999

#ifdef ENABLE_OSC
  gboolean omc_noisy;
#endif

  gshort firsttime;
  const gchar *wm; //window manager name
  gint ocp; // open_compression_percent : get/set in prefs

  // whether the 'blend_file' (second file in a transition) is fg or bg
  gboolean antialias;

  // TODO - move into mainw, make space there first
  gint reorder_width;
  gint reorder_height;
  gboolean reorder_leave_back;

  gboolean ignore_tiny_fps_diffs;

} _prefs;




typedef struct {
  gchar tmpdir[256];
  gchar theme[64];
  gchar vpp_name[64]; // new video playback plugin
  _encoder encoder;
  gboolean show_recent;
  gboolean show_tool;
  guint osc_udp_port;
} _future_prefs;



// some shared structures
capability *capable;
_prefs *prefs;
_future_prefs *future_prefs;

// common defs for mainwindow
#include "plugins.h"
#include "colourspace.h"
#include "mainwindow.h"
#include "keyboard.h"
#include "preferences.h"
mainwindow *mainw;
_prefsw *prefsw;


// internal player clock
#include <sys/time.h>
struct timeval tv;


//networking code
#include "network.h"


// type sizes
size_t sizint, sizdbl, sizlong, sizfloat;


// the name of the executable
gchar myname[256];


// some useful functions

//gui.c
void  create_LiVES (void);
void enable_record (void);
void toggle_record (void);
void disable_record (void);

// interface.c
process* create_processing (const gchar *text);
void add_to_winmenu(void);
void remove_from_winmenu(void);
void make_play_window(void);
void resize_play_window (void);
void kill_play_window (void);
void make_preview_box (void);
void add_to_playframe (void);

// dialogs.c
gboolean do_progress_dialog(gboolean visible, gboolean cancellable, const gchar *text);
gint do_warning_dialog(const gchar *text);
gint do_warning_dialog_with_check(const gchar *text, gint warn_mask_number);
void do_error_dialog(const gchar *text);
void do_blocking_error_dialog(const gchar *text);
void do_memory_error_dialog (void);
void too_many_files(void);
void tempdir_warning (void);

gint do_clipboard_fps_warning(void);
void perf_mem_warning(void);
gboolean do_comments_dialog (void);
void do_auto_dialog(const gchar *text);
void do_encoder_acodec_error (void);
void do_encoder_fps_error(gdouble restricted_fps);
void do_encoder_arate_error(gint best_arate);
void do_encoder_size_error(gint width, gint height);
void do_encoder_sox_error(void);
gboolean do_encoder_aspect_error(gint width, gint height, gboolean allow_aspect_override);

void do_keys_window (void);
void do_messages_window (void);
void do_firstever_dialog (void);
void do_upgrade_error_dialog(void);

// general
void do_text_window (const gchar *title, const gchar *text);

// unused
gboolean do_socket_dialog (lives_socket *v, const gchar *text, const char *host, int port);

#ifdef AUTO_RESAMPLE_RESIZE
gboolean do_encoder_restrict_dialog (gint width, gint height, gdouble fps, gint fps_num, gint fps_denom, gint arate, gboolean anr);
#endif

// saveplay.c
void add_file_info(const gchar *check_handle);
void save_file_comments (void);
void reget_afilesize (file *cfile);
void deduce_file(gchar *filename, gdouble start_time, gint end);
void open_file (const gchar *filename);
void open_file_sel(const gchar *file_name,gdouble start_time, gint frames);
void open_fw_device (void);
gboolean get_new_handle(gint index, gchar *name);
gboolean get_temp_handle(gint index);
void get_handle_from_info_file(void);
void create_cfile(void);
void save_file (gboolean existing, gchar *n_file_name);
void play_file (void);
gboolean save_frame(gint frame, const gchar *filename);
void wait_for_stop (const gchar *stop_command);

// saveplay.c backup
void backup_file(const gchar *filename);
void backup_event_frames(void);
void backup_event_fps(void);
void save_event_frames(void);
void save_event_fps(void);
void write_headers (file *file);

// saveplay.c restore
void restore_file(const gchar *filename);
gboolean read_headers(const gchar *file_name);
gboolean load_event_frames (void);
gboolean load_event_fps (void);
void restore_event_frames(void);
void restore_event_fps(void);

// saveplay.c sets
void open_set_file (gchar *set_name, gint clipnum);

// saveplay.c reorder/deorder
gint reorder_frames(void);
gint deorder_frames(gint old_framecount);



// main.c
gboolean pre_init(void);
void init(void);
void catch_sigint(int signum);
gboolean startup_message_fatal(gpointer msg);
gboolean startup_message_nonfatal(gpointer msg);
capability *get_capabilities(void);
void load_start_image(gint frame);
void load_end_image(gint frame);
void load_preview_image(gboolean update_always);
void load_frame_image(gint frame, gint last_frame);
void sensitize(void);
void desensitize(void);
void procw_desensitize(void);
void close_current_file (void);
void get_next_free_file(void);
void switch_to_file(gint old_file, gint new_file);
void do_quick_switch (gint new_file);
void resize (gdouble scale);
void read_file_details(const gchar *file_name);
gboolean do_start_messages(gpointer);
void set_palette_colours (void);
void set_main_title(const gchar *filename, gint or_untitled_number);
void set_record (void);

//gui.c
void fade_background(void);
void unfade_background(void);
void block_expose (void);
void unblock_expose (void);

// utils.c
void get_dirname(gchar *filename);
void get_basename(gchar *filename);
void d_print(const gchar *text);
void get_pref(const gchar *key, gchar *val, gint maxlen);
void get_pref_default(const gchar *key, gchar *val, gint maxlen);
gboolean get_boolean_pref(const gchar *key);
gdouble get_double_pref(const gchar *key);
gint get_int_pref(const gchar *key);
void get_location(const gchar *exe, gchar *val, gint maxlen);
void set_pref (const gchar *key, const gchar *value);
void set_boolean_pref(const gchar *key, gboolean value);
void set_double_pref(const gchar *key, gdouble value);
void set_int_pref(const gchar *key, gboolean value);
gboolean apply_prefs(gboolean skip_warnings);
void save_future_prefs(void);
void set_menu_text(GtkWidget *menu, const gchar *text, gboolean use_mnemonic);
void get_menu_text(GtkWidget *menu, gchar *text);
void reset_clip_menu (void);
void get_play_times(void);
void get_total_time (file *file);
guint get_signed_endian (int asigned, int endian);
void fullscreen_internal(void);
void unhide_cursor(GdkWindow *window);
void hide_cursor(GdkWindow *window);
void set_alwaysontop(GtkWidget *window, gboolean ontop); // TODO - use for playwin
void colour_equal(GdkColor *c1, const GdkColor *c2);
void switch_to_int_player(void);
void switch_to_mplayer(void);
void switch_aud_to_sox(void);
void switch_aud_to_mplayer(void);
void prepare_to_play_foreign(void);
gboolean after_foreign_play(void);
gboolean check_file(const gchar *file_name);  // check if file exists
gboolean check_dir_access (gchar *dir);
guint get_file_size(int fd);
guint sget_file_size(gchar *name);
void quantise_events (gint rec_start);
gdouble calc_variable_time (gint frame);
gint calc_variable_frame (gdouble time, gshort calc_method);
#define CALC_LAST 0
#define CALC_NEAREST 1

gboolean create_event_space(gint length_in_events, guint64 which_events);
void add_to_recent(const gchar *filename, gdouble start, gint frames);
gint verhash (gchar *version);
void set_undoable (const gchar *what, gboolean sensitive);
void set_redoable (const gchar *what, gboolean sensitive);
void zero_spinbuttons (void);
void draw_little_bars (void);
void set_sel_label (void);
void clear_mainw_msg (void);
gint get_token_count (gchar *string, int delim);
GdkPixmap* gdk_pixmap_copy (GdkPixmap *pixmap);
void get_border_size (GtkWidget *win, gint *bx, gint *by);
gchar *g_strappend (gchar *string, gint len, const gchar *new);
void find_when_to_stop (void);
void minimise_aspect_delta (gdouble allowed_aspect,gint hblock,gint vblock,gint hsize,gint vsize,gint *width,gint *height);

void combo_set_popdown_strings (GtkCombo *combo, GList *list);

gint hextodec (const gchar *string);
gint get_hex_digit (gchar *c);

guint32 fastrand(void);
void fastsrand(guint32 seed);


// effects.c
void invalidate_pixel_buffers (gint rowstride);

// plugins.c
gboolean check_encoder_restrictions (gboolean get_extension);

//callbacks.c
void lives_exit (void);

// osc.c
#ifdef ENABLE_OSC
gboolean lives_osc_init(guint osc_udp_port);
gint lives_osc_poll(gpointer data);
void lives_osc_end(void);
#endif

// inlines
#define cfile mainw->files[mainw->current_file]
#define clipboard mainw->files[0]


#endif // #ifndef HAS_MAIN_H
