// network.c from LiVES
// see file ../COPYING for license
// All LiVES code (c) G. Finch (salsaman) 

// general network architecture
// other projects can reserve a socket type (in network.h)


 
#include <errno.h>
#include <fcntl.h>
#include <gtk/gtk.h>

#include "main.h"












static lives_socket *sockets[MAX_SOCKETS];

void 
network_init (void) {
  // called on LiVES startup
  int i;
  for (i=0;i<MAX_SOCKETS;i++) {
    sockets[i]=NULL;
  }
}

void 
network_end (void) {
  // called on LiVES end
  int i;
  for (i=0;i<MAX_SOCKETS;i++) {
    if (!(sockets[i]==NULL)) {
      close (sockets[i]->handle);
    }
  }
}


gboolean 
socket_connect( lives_socket *v, const char *name, int port_id ) {
  // called to connect a socket v to port port_id on host name
  v->he = gethostbyname(name);
  v->handle = socket( AF_INET, SOCK_STREAM, 0);
  v->server_addr.sin_family = AF_INET;
  v->server_addr.sin_port = htons(port_id);
  v->server_addr.sin_addr = *( (struct in_addr*) v->he->h_addr);
  if (connect(v->handle, (struct sockaddr*) &v->server_addr,sizeof(struct sockaddr))==-1) {
    return FALSE; /* error */
  }
  return TRUE;
}

gboolean  
socket_read(lives_socket *v) {
  // if there is no data, return FALSE
  // else read socket v, and set the result in mainw->msg
  // mainw->error should be checked after
  int nbytes,error;
  error=ioctl(v->handle, FIONREAD, &nbytes, 0);
  if (error==0) {
    if(nbytes == 0) return FALSE; // nothing read
    if (nbytes>511) {
      // read too many bytes
      mainw->error=TRUE;
      g_print ("Network overflow.\n");
      g_snprintf (mainw->msg,512,"Network buffer overflow.");
      return FALSE;
    }
    recv( v->handle, mainw->msg, nbytes,0);
    memset (mainw->msg+nbytes,0,1);
    return TRUE;
  }
  if (error==ETIMEDOUT) {
    g_print ("Network timed out.\n");
    g_snprintf (mainw->msg,512,"Network timed out");
  }
  g_print ("Network error.\n");
  g_snprintf (mainw->msg,512,"Network error");
  return FALSE;
}


gboolean 
socket_write (lives_socket *v, gchar *msg) {
  // write msg to socket v
  if ((send(v->handle, msg, strlen(msg)+1, 0))==-1) {
    g_print("Error sending %s\n",msg);
    return FALSE;
  }
#ifdef VJTEST_DEBUG
  g_print("Network Sent: %s\n",msg);
#endif
  return TRUE;
}







// client functions

// start up veejay and create 2 sockets for it
gboolean 
network_add_vj_client (void) {
  // TODO - check if pid file is present for veejay
  // TODO - start up veejay
  gchar *com=g_strdup ("veejay-dummy");
  gchar *msg;
  mainw->error=FALSE;

  if (!(sockets[VEEJAY_SOCKET_SEND]==NULL&&sockets[VEEJAY_SOCKET_RECV]==NULL)) {
    d_print ("Error: already connected to Veejay\n");
    return FALSE;
  }

  sockets[VEEJAY_SOCKET_SEND] = (lives_socket*)malloc(sizeof(lives_socket));
  sockets[VEEJAY_SOCKET_RECV] = (lives_socket*)malloc(sizeof(lives_socket));

  system (com);
  g_free (com);

  // keep trying to connect to the send socket...timeout after SOCKET_TIMEOUT
  if (do_socket_dialog(sockets[VEEJAY_SOCKET_SEND],"Connecting to Veejay...",VEEJAY_HOST,VEEJAY_PORT)==FALSE) {
    msg=g_strdup_printf("Cannot connect to Veejay at host '%s' on port %d\n",VEEJAY_HOST,VEEJAY_PORT);
    d_print (msg);
    g_free (msg);
    mainw->error=TRUE;
    return FALSE;
  }
  
  if (do_socket_dialog(sockets[VEEJAY_SOCKET_SEND],"Connecting to Veejay...",VEEJAY_HOST,VEEJAY_PORT+1)==FALSE) {
    msg=g_strdup_printf("Cannot connect to Veejay at host '%s' on port %d\n",VEEJAY_HOST,VEEJAY_PORT+1);
    d_print (msg);
    g_free (msg);
    mainw->error=TRUE;
    return FALSE;
  }
  return TRUE;
}



void 
vj_client_send(const char *msg) {
  gchar *buf;
  gchar veejay_msg[MAXDATASIZE+1];
  mainw->error=FALSE;
  
  // TODO - timeout if read takes too long
#ifdef VJTEST_DEBUG
  g_print ("Waiting for message from VJ\n");
#endif
  while (socket_read(sockets[VEEJAY_SOCKET_RECV])==FALSE) {
    if (mainw->error==TRUE) {
      return;
    }
  }
#ifdef VJTEST_DEBUG
  g_print ("Got message from VJ:%s\n",mainw->msg);
#endif

  // parse any status message from veejay here

  //


  // translate our message
  // TODO - use actual message numbers
  if (!(strcmp (msg,"WID0"))) {
    g_snprintf (veejay_msg,MAXDATASIZE,"999 0");
  }
  else if (!(strcmp (msg,"WIDx"))) {
    g_snprintf (veejay_msg,MAXDATASIZE,"999 %x",mainw->xwin);
  }
  else if (!(strcmp (msg,"RESIZE"))) {
    g_snprintf (veejay_msg,MAXDATASIZE,"200 %d %d",mainw->pwidth,mainw->pheight);
  }
  else if (!(strcmp (msg,"RENDER_PIXBUF"))) {
    g_snprintf (veejay_msg,MAXDATASIZE,"1234 %lu %d %d %s",gdk_pixbuf_get_pixels (mainw->frame_image),cfile->hsize,cfile->vsize,"RGB24");
  }

  // send the message
  buf=g_strdup_printf ("V%03dD%s",strlen(veejay_msg),veejay_msg);
	       socket_write (sockets[VEEJAY_SOCKET_SEND],buf);
  g_free (buf);
}








