// paramwindow.c
// LiVES
// (c) G. Finch 2004 <salsaman@xs4all.nl>
// see file COPYING for licensing details

// dynamic window generation from parameter arrays :-)



// NOTE: this is now the "house style" for all LiVES widgets, there should be a consistent look across all 
// windows.




#include "main.h"
#include "paramwindow.h"
#include "callbacks.h"
#include "support.h"
#include "paramspecial.h"

extern void do_effect (lives_rendered_effect_t *);

// TODO -
// use list of these in case we have multiple windows open
// right now this is single threaded because of this
GSList *usrgrp_to_livesgrp=NULL; // ordered list of lives_widget_group_t


void
on_paramwindow_ok_clicked (GtkButton *button, lives_rendered_effect_t *rfx) {
  if (button!=NULL) {
    gtk_widget_destroy(gtk_widget_get_toplevel(GTK_WIDGET(button)));
  }
  if (usrgrp_to_livesgrp!=NULL) g_slist_free (usrgrp_to_livesgrp);
  usrgrp_to_livesgrp=NULL;
  on_render_fx_activate (NULL,rfx);
  special_cleanup();
}


void
on_paramwindow_cancel_clicked (GtkButton *button, lives_rendered_effect_t *rfx) {
  if (button!=NULL) {
    gtk_widget_destroy(gtk_widget_get_toplevel(GTK_WIDGET(button)));
  }
  if (usrgrp_to_livesgrp!=NULL) g_slist_free (usrgrp_to_livesgrp);
  usrgrp_to_livesgrp=NULL;
  special_cleanup();
}


void
on_render_fx_activate (GtkMenuItem *menuitem, lives_rendered_effect_t *rfx) {
  GList *onchange=NULL;
  gchar **array;
  gchar *dir;
  int i;

  // do onchange|init
  if (menuitem!=NULL) {
    switch (rfx->status) {
    case RFX_STATUS_BUILTIN:
      dir=g_strdup(PLUGIN_RENDERED_EFFECTS_BUILTIN);
      break;
    case RFX_STATUS_CUSTOM:
      dir=g_strdup(PLUGIN_RENDERED_EFFECTS_CUSTOM);
      break;
    default:
      dir=g_strdup(PLUGIN_RENDERED_EFFECTS_TEST);
      break;
    }
    if ((onchange=plugin_request_by_line (dir,rfx->name,"get_onchange"))!=NULL) {
      for (i=0;i<g_list_length (onchange);i++) {
	array=g_strsplit (g_list_nth_data (onchange,i),"|",-1);
	if (!strcmp (array[0],"init")) {
	  // onchange is init
	  // create dummy object with data
	  GtkWidget *dummy_widget=gtk_label_new(NULL);
	  g_object_set_data (G_OBJECT (dummy_widget),"param_number",GINT_TO_POINTER (-1));
	  do_onchange (G_OBJECT (dummy_widget),rfx);
	  gtk_widget_destroy (dummy_widget);
	}
	g_strfreev (array);
      }
      g_list_free (onchange);
    }
    g_free (dir);
  }
  if (rfx->min_frames>-1) {
    do_effect(rfx);
  }
}


void
on_render_fx_pre_activate (GtkMenuItem *menuitem, lives_rendered_effect_t *rfx) {
  GtkWidget *fx_dialog;
  GtkWidget *top_dialog_vbox;
  GtkWidget *dialog_action_area;
  GtkWidget *cancelbutton;
  GtkWidget *okbutton;

  gchar *txt;
  gboolean no_process=FALSE;

  if (rfx->min_frames<0) no_process=TRUE;
  else {
    // check we have a real clip open
    if (mainw->current_file<=0) return;
    if (cfile->end-cfile->start+1<rfx->min_frames) {
      txt=g_strdup_printf (_("\nYou must select at least %d frames to use this effect.\n\n"),rfx->min_frames);
      do_blocking_error_dialog (txt);
      g_free (txt);
      return;
    }
    
    // here we invalidate cfile->ohsize, cfile->ovsize
    cfile->ohsize=cfile->hsize;
    cfile->ovsize=cfile->vsize;
    
    if (cfile->undo_action==UNDO_RESIZABLE) {
      set_undoable(NULL,FALSE);
    }
  }

  fx_dialog = gtk_dialog_new ();
  txt=g_strdup_printf ("LiVES: - %s",_(rfx->menu_text));
  gtk_window_set_title (GTK_WINDOW (fx_dialog), txt);
  g_free (txt);

  gtk_container_set_border_width (GTK_CONTAINER (fx_dialog), 20);
  gtk_window_set_position (GTK_WINDOW (fx_dialog), GTK_WIN_POS_CENTER);
  gtk_window_set_modal (GTK_WINDOW (fx_dialog), TRUE);
  gtk_window_set_transient_for(GTK_WINDOW(fx_dialog),GTK_WINDOW(mainw->LiVES));

  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(fx_dialog, GTK_STATE_NORMAL, &palette->normal_back);
    gtk_dialog_set_has_separator(GTK_DIALOG(fx_dialog),FALSE);
  }

  top_dialog_vbox = GTK_DIALOG (fx_dialog)->vbox;

  make_param_box(GTK_VBOX (top_dialog_vbox), rfx);

  dialog_action_area = GTK_DIALOG (fx_dialog)->action_area;
  gtk_container_set_border_width (GTK_CONTAINER (dialog_action_area), 80);
  gtk_box_set_spacing (GTK_BOX (dialog_action_area),10);

  cancelbutton = gtk_button_new_from_stock ("gtk-cancel");
  GTK_WIDGET_SET_FLAGS (cancelbutton, GTK_CAN_DEFAULT);
  if (!no_process) {
    gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area), GTK_BUTTONBOX_END);
    gtk_dialog_add_action_widget (GTK_DIALOG (fx_dialog), cancelbutton, GTK_RESPONSE_CANCEL);
    okbutton = gtk_button_new_from_stock ("gtk-ok");
    gtk_dialog_add_action_widget (GTK_DIALOG (fx_dialog), okbutton, GTK_RESPONSE_OK);
  }
  else {
    okbutton = gtk_button_new_from_stock ("gtk-ok");
    gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area), GTK_BUTTONBOX_SPREAD);
    gtk_box_pack_start (GTK_BOX (dialog_action_area), okbutton, TRUE, TRUE, 0);
    gtk_button_set_relief (GTK_BUTTON (okbutton), GTK_RELIEF_NONE);
    if (palette->style&STYLE_1) {
      gtk_widget_modify_bg(okbutton, GTK_STATE_SELECTED|GTK_STATE_ACTIVE, &palette->white);
    }
  }

  GTK_WIDGET_SET_FLAGS (okbutton, GTK_CAN_DEFAULT);
  gtk_widget_grab_default (okbutton);

  gtk_widget_show_all (fx_dialog);

  gtk_window_add_accel_group (GTK_WINDOW (fx_dialog), mainw->accel_group);
  gtk_widget_add_accelerator (cancelbutton, "activate", mainw->accel_group,
                              GDK_Escape, 0, 0);


  g_signal_connect (GTK_OBJECT (cancelbutton), "clicked",
		    G_CALLBACK (on_paramwindow_cancel_clicked),
		    NULL);

  if (no_process) {
    g_signal_connect (GTK_OBJECT (okbutton), "clicked",
		      G_CALLBACK (on_paramwindow_cancel_clicked),
		      (gpointer)rfx);
  }
  else {
    g_signal_connect (GTK_OBJECT (okbutton), "clicked",
		      G_CALLBACK (on_paramwindow_ok_clicked),
		      (gpointer)rfx);
  }
  g_signal_connect (GTK_OBJECT (fx_dialog), "delete_event",
                      G_CALLBACK (return_true),
                      NULL);

}
  




void make_param_box(GtkVBox *top_vbox, lives_rendered_effect_t *rfx) {
  // make a dynamic parameter window
  GtkWidget *param_vbox;
  GtkWidget *top_hbox;
  GtkWidget *hbox=NULL;

  gchar **array;
  gboolean used[rfx->num_params];
  gboolean has_box=FALSE;
  gchar *line;
  gchar *dir;
  GList *hints=NULL;
  GList *onchange=NULL;
  GSList *layout=NULL;
  int i,j,k,pnum;
  int length;

  gint num_tok;
  gchar label_text[256]; // max length of a label in layout hints

  if (mainw->current_file>0) {
    // initialise special widgets
    init_special();
  }

  usrgrp_to_livesgrp=NULL;

  // paramwindow start, everything goes in top_hbox
  // TODO - make this hbox embeddable, use it for 'merge'

  top_hbox = gtk_hbox_new (FALSE, 10);
  gtk_box_pack_start (GTK_BOX (top_vbox), top_hbox, TRUE, TRUE, 0);

  // dialog vbox holds the dynamic parameters
  param_vbox = gtk_vbox_new (TRUE, 10);
  gtk_box_pack_start (GTK_BOX (top_hbox), param_vbox, TRUE, TRUE, 0);
  gtk_box_set_spacing (GTK_BOX (param_vbox), 20);
  switch (rfx->status) {
  case RFX_STATUS_BUILTIN:
    dir=g_strdup(PLUGIN_RENDERED_EFFECTS_BUILTIN);
    break;
  case RFX_STATUS_CUSTOM:
    dir=g_strdup(PLUGIN_RENDERED_EFFECTS_CUSTOM);
    break;
  default:
    dir=g_strdup(PLUGIN_RENDERED_EFFECTS_TEST);
    break;
  }

  // do onchange|init
  if ((onchange=plugin_request_by_line (dir,rfx->name,"get_onchange"))!=NULL) {
    for (i=0;i<g_list_length (onchange);i++) {
      array=g_strsplit (g_list_nth_data (onchange,i),"|",-1);
      if (!strcmp (array[0],"init")) {
	// onchange is init
	// create dummy object with data
	GtkWidget *dummy_widget=gtk_label_new(NULL);
	g_object_set_data (G_OBJECT (dummy_widget),"param_number",GINT_TO_POINTER (-1));
	do_onchange (G_OBJECT (dummy_widget),rfx);
	gtk_widget_destroy (dummy_widget);
      }
      else {
	// note other onchanges so we don't have to keep parsing the list
	gint which=atoi (array[0]);
	if (which>0&&which<=rfx->num_params) {
	  rfx->params[which-1].onchange=TRUE;
	}
      }
      g_strfreev (array);
    }
    g_list_free (onchange);
  }

  // do param window hints
  if ((hints=plugin_request_by_line (dir,rfx->name,"get_param_window"))!=NULL) {
    for (i=0;i<g_list_length (hints);i++) {
      if (!strncmp (g_list_nth_data (hints,i),"layout|",7)) {
	layout=g_slist_append (layout,(gchar *)g_list_nth_data (hints,i)+7);
      }
      else if (!strncmp (g_list_nth_data (hints,i),"special|",8)&&mainw->current_file>0) {
	add_to_special((gchar *)g_list_nth_data (hints,i)+8,rfx);
      }
    }
    g_list_free (hints);  // no longer needed
  }
  g_free(dir);

  for (i=0;i<rfx->num_params;i++) {
    used[i]=FALSE;
    for (j=0;j<MAX_PARAM_WIDGETS;rfx->params[i].widgets[j++]=NULL);
  }

  // use layout hints to build as much as we can
  for (i=0;i<g_slist_length (layout);i++) {
    has_box=FALSE;
    line=g_slist_nth_data (layout,i);
    num_tok=get_token_count (line,'|');
    // ignore | inside strings
    array=g_strsplit (line,"|",num_tok);
    for (j=0;j<num_tok;j++) {
      if (!strncmp (array[j],"p",1)&&(pnum=atoi ((gchar *)array[j]+1))>0&&pnum<=rfx->num_params&&!used[pnum-1]) {
	// parameter, eg. p1
	if (!has_box) {
	  hbox = gtk_hbox_new (FALSE, 0);
	  gtk_box_pack_start (GTK_BOX (param_vbox), hbox, TRUE, TRUE, 0);
	  has_box=TRUE;
	}
	add_param_to_box (GTK_BOX (hbox),rfx,pnum-1);
	used[pnum-1]=TRUE;
      }
      else if (!j&&!strcmp (array[j],"hseparator")) {
	hbox = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (param_vbox), hbox, TRUE, TRUE, 0);
	add_hsep_to_box (GTK_BOX (hbox));
	j=num_tok;  // ignore anything after hseparator
      }
      else if (!strncmp (array[j],"fill",4)) {
	// can be filln
	if (strlen (array[j])==4||!(length=atoi (array[j]+4))) length=1;
	if (!has_box) {
	  hbox = gtk_hbox_new (FALSE, 0);
	  gtk_box_pack_start (GTK_BOX (param_vbox), hbox, TRUE, TRUE, 0);
	  has_box=TRUE;
	}
	for (k=0;k<length;k++) {
	  add_fill_to_box (GTK_BOX (hbox));
	}
      }
      else if (!strncmp (array[j],"\"",1)) {
	// label
	if (!has_box) {
	  hbox = gtk_hbox_new (FALSE, 0);
	  gtk_box_pack_start (GTK_BOX (param_vbox), hbox, TRUE, TRUE, 0);
	  has_box=TRUE;
	}
	g_snprintf (label_text,256,"%s",array[j]+1);
	while (strcmp (array[j]+strlen (array[j])-1,"\"")&&j<num_tok-1) {
	  g_strappend (label_text,256,array[++j]);
	}
	if (strlen (label_text)>1) {
	  if (!strcmp (label_text+strlen (label_text)-1,"\"")) {
	    memset (label_text+strlen (label_text)-1,0,1);
	  }
	  add_label_to_box (GTK_BOX (hbox),label_text);
	}}}
    g_strfreev (array);
  }
  if (layout!=NULL) g_slist_free (layout);

  // add any unused parameters
  for (i=0;i<rfx->num_params;i++) {
    if (!used[i]) {
      hbox = gtk_hbox_new (FALSE, 0);
      gtk_box_pack_start (GTK_BOX (param_vbox), hbox, TRUE, TRUE, 0);
      add_param_to_box (GTK_BOX (hbox),rfx,i);
    }
  }

  // add a padding box
  hbox = gtk_hbox_new (FALSE, 10);
  gtk_box_pack_start (GTK_BOX (param_vbox), hbox, TRUE, TRUE, 0);
  add_fill_to_box (GTK_BOX (hbox));
}






void add_param_to_box (GtkBox *box, lives_rendered_effect_t *rfx, gint pnum) {
  // box here is vbox inside top_hbox inside top_dialog
  GtkWidget *label;
  GtkWidget *labelcname;
  GtkWidget *checkbutton;
  GtkWidget *radiobutton;
  GtkWidget *spinbutton;
  GtkWidget *spinbutton_red;
  GtkWidget *spinbutton_green;
  GtkWidget *spinbutton_blue;
  GtkWidget *button;
  GtkWidget *hbox;
  GtkObject *spinbutton_adj;
  GtkObject *spinbutton_red_adj;
  GtkObject *spinbutton_green_adj;
  GtkObject *spinbutton_blue_adj;

  lives_param_t *param;
  lives_widget_group_t *group=NULL;
  gint maxlen;
  gulong satfunc,spinfunc;
  gchar *name;
  gchar *txt;

  lives_colRGB24_t rgb;
  lives_colRGB24_t rgb_def;
  gboolean not_default=FALSE;

  if (pnum>=rfx->num_params) {
    add_label_to_box (box,(_("Invalid parameter")));
    return;
  }

  param=&rfx->params[pnum];
  name=g_strdup_printf ("%s",param->name);
  switch (param->type) {
  case LIVES_PARAM_BOOL :
    if (!param->group) {
      checkbutton = gtk_check_button_new ();
      label=gtk_label_new_with_mnemonic (_ (name));
      gtk_label_set_mnemonic_widget (GTK_LABEL (label),checkbutton);
      if (palette->style&STYLE_1) {
	gtk_widget_modify_fg(label, GTK_STATE_NORMAL, &palette->normal_fore);
      }
      hbox = gtk_hbox_new (FALSE, 0);
      gtk_box_pack_start (GTK_BOX (box), hbox, FALSE, FALSE, 10);
      gtk_box_pack_start (GTK_BOX (hbox), checkbutton, FALSE, FALSE, 0);
      gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 10);
      GTK_WIDGET_SET_FLAGS (checkbutton, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkbutton), get_bool_param(param->value));
      g_signal_connect_after (GTK_OBJECT (checkbutton), "toggled",
			      G_CALLBACK (after_boolean_param_toggled),
			      (gpointer)rfx);
      
      // store parameter so we know whose trigger to use
      g_object_set_data (G_OBJECT (checkbutton),"param_number",GINT_TO_POINTER (pnum+1));
      param->widgets[0]=checkbutton;
    }
    else {
      if ((group=livesgrp_from_usrgrp (usrgrp_to_livesgrp, param->group))==NULL) {
	radiobutton = gtk_radio_button_new (NULL);
	usrgrp_to_livesgrp=add_usrgrp_to_livesgrp (usrgrp_to_livesgrp, gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton)), param->group);
	group=livesgrp_from_usrgrp (usrgrp_to_livesgrp, param->group);
      }
      else {
	radiobutton = gtk_radio_button_new (group->rbgroup);
	group->rbgroup=gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton));
      }
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radiobutton), get_bool_param(param->value));
      if (get_bool_param(param->value)) {
	group->active_param=pnum+1;
      }
      GTK_WIDGET_SET_FLAGS (radiobutton, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
      hbox = gtk_hbox_new (FALSE, 0);
      gtk_box_pack_start (GTK_BOX (box), hbox, FALSE, FALSE, 10);
      gtk_box_pack_start (GTK_BOX (hbox), radiobutton, FALSE, FALSE, 0);
      
      label=gtk_label_new_with_mnemonic (_ (name));
      gtk_label_set_mnemonic_widget (GTK_LABEL (label),radiobutton);
      if (palette->style&STYLE_1) {
	gtk_widget_modify_fg(label, GTK_STATE_NORMAL, &palette->normal_fore);
      }
      gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 10);
      
      g_signal_connect_after (GTK_OBJECT (radiobutton), "toggled",
			      G_CALLBACK (after_boolean_param_toggled),
			      (gpointer)rfx);
      // store parameter so we know whose trigger to use
      g_object_set_data (G_OBJECT (radiobutton),"param_number",GINT_TO_POINTER (pnum+1));
      param->widgets[0]=radiobutton;
      }
    break;
    
  case LIVES_PARAM_NUM :
    // TODO - set page size etc
    label = gtk_label_new_with_mnemonic (_ (name));
    if (palette->style&STYLE_1) {
      gtk_widget_modify_fg(label, GTK_STATE_NORMAL, &palette->normal_fore);
    }
    if (param->dp) {
      spinbutton_adj = gtk_adjustment_new ((gdouble)(get_float_param(param->value)), (gdouble)param->min, (gdouble)param->max, 1, 10, 10);
    }
    else {
      spinbutton_adj = gtk_adjustment_new ((gdouble)(get_int_param(param->value)), (gdouble)param->min, (gdouble)param->max, 1, 10, 10);
    }

    spinbutton = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton_adj), 1, param->dp);
    gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (spinbutton),TRUE);
    txt=g_strdup_printf ("%d",(int)param->max);
    maxlen=strlen (txt);
    g_free (txt);
    txt=g_strdup_printf ("%d",(int)param->min);
    if (strlen (txt)>maxlen) maxlen=strlen (txt);
    g_free (txt);

    gtk_entry_set_width_chars (GTK_ENTRY (spinbutton),maxlen+param->dp+param->dp?4:1);
    GTK_WIDGET_SET_FLAGS (spinbutton, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
    gtk_entry_set_activates_default (GTK_ENTRY ((GtkEntry *)&(GTK_SPIN_BUTTON (spinbutton)->entry)), TRUE);
    gtk_label_set_mnemonic_widget (GTK_LABEL (label), spinbutton);
    gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (spinbutton),GTK_UPDATE_ALWAYS);
    
    if (param->group) {
      if ((group=livesgrp_from_usrgrp (usrgrp_to_livesgrp, param->group))==NULL) {
	radiobutton = gtk_radio_button_new (NULL);
	usrgrp_to_livesgrp=add_usrgrp_to_livesgrp (usrgrp_to_livesgrp, gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton)), param->group);
	group=livesgrp_from_usrgrp (usrgrp_to_livesgrp, param->group);
      }
      else {
	radiobutton = gtk_radio_button_new (group->rbgroup);
	group->rbgroup=gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton));
      }
      if (param->dp) {
	if (get_float_param (param->value)!=get_float_param (param->def)) not_default=TRUE;
      }
      else {
	if (get_int_param (param->value)!=get_int_param (param->def)) not_default=TRUE;
      }
      if (!group->active_param||not_default) {
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radiobutton), TRUE);
	group->active_param=pnum+1;
      }
      else gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radiobutton), FALSE);

      // store parameter so we know whose default to use
      g_object_set_data (G_OBJECT (radiobutton),"param",(gpointer)param);
      
      GTK_WIDGET_SET_FLAGS (radiobutton, 0);
      gtk_box_pack_start (GTK_BOX (box), radiobutton, FALSE, FALSE, 0);
      
      // activate this radiobutton when value is changed
      // and when deactivated, set back to default
      satfunc=g_signal_connect_after (GTK_OBJECT (spinbutton), "value_changed",
			      G_CALLBACK (spin_act_toggle),
			      radiobutton);
      
      add_gulong_data_to_widget(GTK_WIDGET(spinbutton),"satfunc",satfunc);

      g_signal_connect_after (GTK_OBJECT (radiobutton), "toggled",
			      G_CALLBACK (toggle_set_default_value),
			      spinbutton);
      
    }
    hbox = gtk_hbox_new (TRUE, 0);
    gtk_box_pack_start (GTK_BOX (box), hbox, TRUE, TRUE, 10);
    gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 10);
    gtk_box_pack_start (GTK_BOX (hbox), spinbutton, TRUE, TRUE, 0);
    
    spinfunc=g_signal_connect_after (GTK_OBJECT (spinbutton), "value_changed",
			    G_CALLBACK (after_param_value_changed),
			    (gpointer)rfx);
    add_gulong_data_to_widget(GTK_WIDGET(spinbutton),"spinfunc",spinfunc);
    
    // store parameter so we know whose trigger to use
    g_object_set_data (G_OBJECT (spinbutton),"param_number",GINT_TO_POINTER (pnum+1));
    param->widgets[0]=spinbutton;
    break;
    
  case LIVES_PARAM_COLRGB24 :
    get_colRGB24_param(param->value,&rgb);

    spinbutton_red_adj = gtk_adjustment_new (rgb.red, 0, 255, 1, 10, 10);
    spinbutton_red = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton_red_adj), 1, 0);
    spinbutton_green_adj = gtk_adjustment_new (rgb.green, 0, 255, 1, 10, 10);
    spinbutton_green = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton_green_adj), 1, 0);
    spinbutton_blue_adj = gtk_adjustment_new (rgb.blue, 0, 255, 1, 10, 10);
    spinbutton_blue = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton_blue_adj), 1, 0);
    
    gtk_entry_set_width_chars (GTK_ENTRY (spinbutton_red),4);
    gtk_entry_set_width_chars (GTK_ENTRY (spinbutton_green),4);
    gtk_entry_set_width_chars (GTK_ENTRY (spinbutton_blue),4);
    GTK_WIDGET_SET_FLAGS (spinbutton_red, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
    GTK_WIDGET_SET_FLAGS (spinbutton_green, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
    GTK_WIDGET_SET_FLAGS (spinbutton_blue, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
    gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (spinbutton_red),GTK_UPDATE_ALWAYS);
    gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (spinbutton_green),GTK_UPDATE_ALWAYS);
    gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (spinbutton_blue),GTK_UPDATE_ALWAYS);
    gtk_entry_set_activates_default (GTK_ENTRY ((GtkEntry *)&(GTK_SPIN_BUTTON (spinbutton_red)->entry)), TRUE);
    gtk_entry_set_activates_default (GTK_ENTRY ((GtkEntry *)&(GTK_SPIN_BUTTON (spinbutton_green)->entry)), TRUE);
    gtk_entry_set_activates_default (GTK_ENTRY ((GtkEntry *)&(GTK_SPIN_BUTTON (spinbutton_blue)->entry)), TRUE);
    
    if (param->group) {
      if ((group=livesgrp_from_usrgrp (usrgrp_to_livesgrp, param->group))==NULL) {
	radiobutton = gtk_radio_button_new (NULL);
	usrgrp_to_livesgrp=add_usrgrp_to_livesgrp (usrgrp_to_livesgrp, gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton)), param->group);
	group=livesgrp_from_usrgrp (usrgrp_to_livesgrp, param->group);
      }
      else {
	radiobutton = gtk_radio_button_new (group->rbgroup);
	group->rbgroup=gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton));
      }

      get_colRGB24_param (param->value,&rgb);
      get_colRGB24_param (param->def,&rgb_def);
      if (rgb.red!=rgb_def.red||rgb.green!=rgb_def.green||rgb.blue!=rgb_def.blue) not_default=TRUE;

      if (!group->active_param||not_default) {
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radiobutton), TRUE);
	group->active_param=pnum+1;
      }
      else gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radiobutton), FALSE);
      GTK_WIDGET_SET_FLAGS (radiobutton, 0);
      gtk_box_pack_start (GTK_BOX (box), radiobutton, FALSE, FALSE, 0);
      
      // store parameter so we know whose default to use
      g_object_set_data (G_OBJECT (radiobutton),"param",(gpointer)param);
      
      // activate this radiobutton when value is changed
      // and when deactivated, set back to default
      satfunc=g_signal_connect_after (GTK_OBJECT (spinbutton_red), "value_changed",
				      G_CALLBACK (spin_act_toggle),
				      radiobutton);
      add_gulong_data_to_widget(GTK_WIDGET(spinbutton_red),"satfunc",satfunc);

      satfunc=g_signal_connect_after (GTK_OBJECT (spinbutton_green), "value_changed",
				      G_CALLBACK (spin_act_toggle),
				      radiobutton);
      add_gulong_data_to_widget(GTK_WIDGET(spinbutton_green),"satfunc",satfunc);
      satfunc=g_signal_connect_after (GTK_OBJECT (spinbutton_blue), "value_changed",
				      G_CALLBACK (spin_act_toggle),
				      radiobutton);
      add_gulong_data_to_widget(GTK_WIDGET(spinbutton_red),"satfunc",satfunc);
      
      g_signal_connect_after (GTK_OBJECT (radiobutton), "toggled",
			      G_CALLBACK (toggle_set_default_red),
			      spinbutton_red);
      g_signal_connect_after (GTK_OBJECT (radiobutton), "toggled",
			      G_CALLBACK (toggle_set_default_green),
			      spinbutton_green);
      g_signal_connect_after (GTK_OBJECT (radiobutton), "toggled",
			      G_CALLBACK (toggle_set_default_blue),
			      spinbutton_blue);
    }
    hbox = gtk_hbox_new (FALSE, 0);
    gtk_box_pack_start (GTK_BOX (box), hbox, FALSE, FALSE, 10);
    
    labelcname=gtk_label_new_with_mnemonic (_ (param->name));
    gtk_label_set_justify (GTK_LABEL (labelcname), GTK_JUSTIFY_LEFT);
    if (palette->style&STYLE_1) {
      gtk_widget_modify_fg(labelcname, GTK_STATE_NORMAL, &palette->normal_fore);
    }
    gtk_box_pack_start (GTK_BOX (hbox), labelcname, FALSE, FALSE, 10);
    
    label=gtk_label_new (_("Red"));
    gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
    if (palette->style&STYLE_1) {
      gtk_widget_modify_fg(label, GTK_STATE_NORMAL, &palette->normal_fore);
    }
    gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 10);
    gtk_box_pack_start (GTK_BOX (hbox), spinbutton_red, FALSE, FALSE, 0);
    
    label=gtk_label_new (_("Green"));
    gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
    if (palette->style&STYLE_1) {
      gtk_widget_modify_fg(label, GTK_STATE_NORMAL, &palette->normal_fore);
    }
    gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 10);
    
    gtk_box_pack_start (GTK_BOX (hbox), spinbutton_green, FALSE, FALSE, 0);
    label=gtk_label_new (_("Blue"));
    gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
    if (palette->style&STYLE_1) {
      gtk_widget_modify_fg(label, GTK_STATE_NORMAL, &palette->normal_fore);
    }
    gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 10);
    
    gtk_box_pack_start (GTK_BOX (hbox), spinbutton_blue, FALSE, FALSE, 0);
    
    
    // colsel button : TODO - use gtk_color_button
    button = gtk_button_new_from_stock ("gtk-select-color");
    g_object_set_data (G_OBJECT (button),"param_number",GINT_TO_POINTER (pnum+1));
    gtk_label_set_mnemonic_widget (GTK_LABEL (labelcname),button);
    gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 20);
    g_signal_connect (GTK_OBJECT (button), "clicked",
		      G_CALLBACK (on_pwcolourbutton_clicked),
		      (gpointer)rfx);
    
    g_signal_connect_after (GTK_OBJECT (spinbutton_red), "value_changed",
			    G_CALLBACK (after_param_red_changed),
			    (gpointer)rfx);
    g_signal_connect_after (GTK_OBJECT (spinbutton_green), "value_changed",
			    G_CALLBACK (after_param_green_changed),
			    (gpointer)rfx);
    g_signal_connect_after (GTK_OBJECT (spinbutton_blue), "value_changed",
			    G_CALLBACK (after_param_blue_changed),
			    (gpointer)rfx);
    
    // store parameter so we know whose trigger to use
    g_object_set_data (G_OBJECT (spinbutton_red),"param_number",GINT_TO_POINTER (pnum+1));
    g_object_set_data (G_OBJECT (spinbutton_green),"param_number",GINT_TO_POINTER (pnum+1));
    g_object_set_data (G_OBJECT (spinbutton_blue),"param_number",GINT_TO_POINTER (pnum+1));
    
    param->widgets[0]=spinbutton_red;
    param->widgets[1]=spinbutton_green;
    param->widgets[2]=spinbutton_blue;
    break;
  }
  
  if (mainw->current_file>0) {
    // see if there were any 'special' hints
    check_for_special (param,pnum,box);
  }
  g_free (name);
}

void add_hsep_to_box (GtkBox *box) {
  GtkWidget *hseparator = gtk_hseparator_new ();
  gtk_box_pack_start (box, hseparator, TRUE, TRUE, 0);
}

void add_fill_to_box (GtkBox *box) {
  GtkWidget *blank_label = gtk_label_new ("");
  gtk_box_pack_start (box, blank_label, TRUE, TRUE, 0);
}

void add_label_to_box (GtkBox *box, const gchar *text) {
  GtkWidget *label = gtk_label_new_with_mnemonic (_ (text));
  gtk_box_pack_start (box, label, TRUE, TRUE, 10);
  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
    if (palette->style&STYLE_1) {
      gtk_widget_modify_fg(label, GTK_STATE_NORMAL, &palette->normal_fore);
    }
}

GSList *add_usrgrp_to_livesgrp (GSList *u2l, GSList *rbgroup, gint usr_number) {
  lives_widget_group_t *wgroup=g_malloc (sizeof(lives_widget_group_t));
  wgroup->usr_number=usr_number;
  wgroup->rbgroup=rbgroup;
  wgroup->active_param=0;
  u2l=g_slist_append (u2l, (gpointer)wgroup);
  return u2l;
}




lives_widget_group_t *livesgrp_from_usrgrp (GSList *u2l, gint usrgrp) {
  int i;
  lives_widget_group_t *group;

  for (i=0;i<g_slist_length (u2l);i++) {
    group=g_slist_nth_data (u2l,i);
    if (group->usr_number==usrgrp) return group;
  }
  return NULL;
}



void
after_boolean_param_toggled        (GtkToggleButton *togglebutton,
				    lives_rendered_effect_t *         rfx)
{
  gint param_number=GPOINTER_TO_INT (g_object_get_data (G_OBJECT (togglebutton),"param_number"));
  lives_param_t *param=&rfx->params[param_number-1];
  gboolean old_value=get_bool_param(param->value);

  set_bool_param(param->value,gtk_toggle_button_get_active (togglebutton));

  if (get_bool_param(param->value)!=old_value&&param->onchange) {
    do_onchange (G_OBJECT (togglebutton), rfx);
  }
}



void
after_param_value_changed           (GtkSpinButton   *spinbutton,
				     lives_rendered_effect_t *rfx)
{
  gint param_number=GPOINTER_TO_INT (g_object_get_data (G_OBJECT (spinbutton),"param_number"));
  lives_param_t *param=&rfx->params[param_number-1];
  gfloat old_float=0.;
  gint old_int=0;

  if (param->dp) {
    old_float=get_float_param(param->value);
  }
  else {
    old_int=get_int_param(param->value);
  }

  if (param->dp) {
    set_float_param(param->value,(gfloat)gtk_spin_button_get_value(GTK_SPIN_BUTTON(spinbutton)));
  }
  else {
    set_int_param(param->value,gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbutton)));
  }
  if (((param->dp&&(get_float_param(param->value)!=old_float))||(!param->dp&&(get_int_param(param->value)!=old_int)))&&param->onchange) {
    do_onchange (G_OBJECT (spinbutton), rfx);
  }
}

void
after_param_red_changed           (GtkSpinButton   *spinbutton,
				   lives_rendered_effect_t *rfx) {
  //  after_param_value_changed (spinbutton, rfx);
  gint param_number=GPOINTER_TO_INT (g_object_get_data (G_OBJECT (spinbutton),"param_number"));
  lives_param_t *param=&rfx->params[param_number-1];
  lives_colRGB24_t old_value;
  gint new_red;

  get_colRGB24_param(param->value,&old_value);
  new_red=gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbutton));
  set_colRGB24_param(param->value,new_red,old_value.green,old_value.blue);
  if (new_red!=old_value.red&&param->onchange) {
    do_onchange (G_OBJECT (spinbutton), rfx);
  }
}

void
after_param_green_changed           (GtkSpinButton   *spinbutton,
				     lives_rendered_effect_t *rfx) {
  //  after_param_value_changed (spinbutton, rfx);
  gint param_number=GPOINTER_TO_INT (g_object_get_data (G_OBJECT (spinbutton),"param_number"));
  lives_param_t *param=&rfx->params[param_number-1];
  lives_colRGB24_t old_value;
  gint new_green;

  get_colRGB24_param(param->value,&old_value);
  new_green=gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbutton));
  set_colRGB24_param(param->value,old_value.red,new_green,old_value.blue);
  if (new_green!=old_value.green&&param->onchange) {
    do_onchange (G_OBJECT (spinbutton), rfx);
  }
}


void
after_param_blue_changed           (GtkSpinButton   *spinbutton,
				    lives_rendered_effect_t *rfx) {
  gint param_number=GPOINTER_TO_INT (g_object_get_data (G_OBJECT (spinbutton),"param_number"));
  lives_param_t *param=&rfx->params[param_number-1];
  lives_colRGB24_t old_value;
  gint new_blue;

  get_colRGB24_param(param->value,&old_value);
  new_blue=gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbutton));
  set_colRGB24_param(param->value,old_value.red,old_value.green,new_blue);
  if (new_blue!=old_value.blue&&param->onchange) {
    do_onchange (G_OBJECT (spinbutton), rfx);
  }
}
 


void
after_param_alpha_changed           (GtkSpinButton   *spinbutton,
				     lives_rendered_effect_t *rfx) {
  gint param_number=GPOINTER_TO_INT (g_object_get_data (G_OBJECT (spinbutton),"param_number"));
  lives_param_t *param=&rfx->params[param_number-1];
  lives_colRGBA32_t old_value;

  gint new_alpha=gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbutton));

  get_colRGBA32_param(param->value,&old_value);
  
  set_colRGBA32_param(param->value,old_value.red,old_value.green,old_value.blue,new_alpha);
  if (new_alpha!=old_value.alpha&&param->onchange) {
    do_onchange (G_OBJECT (spinbutton), rfx);
  }
}


gchar *param_marshall (lives_rendered_effect_t *rfx, gboolean with_min_max) {
  // this function will marshall all parameters into a space separated string
  // in case of string parameters, these will be surrounded by " and all 
  // quotes will be escaped \"

  // the returned string should be g_free()'ed after use
  gchar *old_return=g_strdup (""),*new_return=g_strdup ("");
  gchar *return_pattern;
  lives_colRGB24_t rgb;
  int i;
 
  for (i=0;i<rfx->num_params;i++) {
    switch (rfx->params[i].type) {
    case LIVES_PARAM_COLRGB24:
      get_colRGB24_param(rfx->params[i].value,&rgb);
      if (!with_min_max) {
	new_return=g_strdup_printf ("%s %d",old_return,(((rgb.red<<8)+rgb.green)<<8)+rgb.blue);
      }
      else {
	new_return=g_strdup_printf ("%s %d %d %d",old_return,rgb.red,rgb.green,rgb.blue);
      }
      old_return=new_return;
      break;
    default:
      if (rfx->params[i].dp) {
	return_pattern=g_strdup_printf ("%%s %%.%df",rfx->params[i].dp);
	new_return=g_strdup_printf (return_pattern,old_return,get_float_param(rfx->params[i].value));
	if (with_min_max) {
	  g_free (old_return);
	  old_return=new_return;
	  new_return=g_strdup_printf (return_pattern,old_return,rfx->params[i].min);
	  g_free (old_return);
	  old_return=new_return;
	  new_return=g_strdup_printf (return_pattern,old_return,rfx->params[i].max);
	}
	g_free (return_pattern);
      }
      else {
	new_return=g_strdup_printf ("%s %d",old_return,get_int_param(rfx->params[i].value));
	if (with_min_max&&rfx->params[i].type!=LIVES_PARAM_BOOL) {
	  g_free (old_return);
	  old_return=new_return;
	  new_return=g_strdup_printf ("%s %d",old_return,(int)rfx->params[i].min);
	  g_free (old_return);
	  old_return=new_return;
	  new_return=g_strdup_printf ("%s %d",old_return,(int)rfx->params[i].max);
	}
      }
      g_free (old_return);
      old_return=new_return;
    }
  }
  if (mainw->current_file>0&&with_min_max) {
    if (rfx->num_in_channels<2) {
      new_return=g_strdup_printf ("%s %d %d %d %d %d",old_return,cfile->hsize,cfile->vsize,cfile->start,cfile->end,cfile->frames);
    }
    else {
      // for transitions, change the end to indicate the merge section
      // this is better for length calculations
      gint cb_frames=clipboard->frames;
      gint start=cfile->start,end=cfile->end,ttl;

      if (prefs->ins_resample==TRUE&&clipboard->fps!=cfile->fps) {
	if ((cb_frames=(gint) (clipboard->frames/clipboard->fps*cfile->fps+.49999))<1) cb_frames=1;
      }

      if (cfile->end-cfile->start+1>(cb_frames*(ttl=gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON (merge_opts->spinbutton_loops))))&&!merge_opts->loop_to_fit) {
	end=cb_frames*ttl;
	if (!merge_opts->align_start) {
	  start=cfile->end-end+1;
	  end=cfile->end;
	}
	else {
	  start=cfile->start;
	  end+=start-1;
	}
      }
      new_return=g_strdup_printf ("%s %d %d %d %d %d %d %d",old_return,cfile->hsize,cfile->vsize,start,end,cfile->frames,clipboard->hsize,clipboard->vsize);
    }
  }
  else {
    new_return=g_strdup (old_return);
  }
  g_free (old_return);
  return new_return;
}


void param_demarshall (lives_rendered_effect_t *rfx, GList *plist, gboolean upd) {
  int i;
  gint pnum=0;
  gint red,green,blue;
  lives_param_t *param;

  // param->widgets[n] are only valid if upd==TRUE

  if (plist==NULL) return;

  for (i=0;i<rfx->num_params;i++) {
    param=&rfx->params[i];
    switch (param->type) {
    case LIVES_PARAM_BOOL:
      set_bool_param(param->value,~~(atoi (g_list_nth_data (plist,pnum++))));
      if (upd) {
	if (param->widgets[0]&&GTK_IS_TOGGLE_BUTTON (param->widgets[0])) {
	  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (param->widgets[0]),get_bool_param(param->value));
	}
      }
      break;
    case LIVES_PARAM_NUM:
      if (param->dp) {
	gfloat float_val=(gfloat)g_strtod (g_list_nth_data (plist,pnum++),NULL);
	param->min=(gfloat)g_strtod (g_list_nth_data (plist,pnum++),NULL);
	param->max=(gfloat)g_strtod (g_list_nth_data (plist,pnum++),NULL);
	if (float_val<param->min) float_val=param->min;
	if (float_val>param->max) float_val=param->max;
	set_float_param(param->value,float_val);
	if (upd) {
	  if (param->widgets[0]&&GTK_IS_SPIN_BUTTON (param->widgets[0])) {
	    gulong spinfunc=get_gulong_data_from_widget(GTK_WIDGET(param->widgets[0]),"spinfunc");
	    g_signal_handler_block(param->widgets[0],spinfunc);
	    gtk_spin_button_set_range (GTK_SPIN_BUTTON (param->widgets[0]),(gdouble)param->min,(gdouble)param->max);
	    gtk_spin_button_update(GTK_SPIN_BUTTON(param->widgets[0]));
	    g_signal_handler_unblock(param->widgets[0],spinfunc);
	    gtk_spin_button_set_value (GTK_SPIN_BUTTON (param->widgets[0]),(gdouble)get_float_param(param->value));
	    gtk_spin_button_update(GTK_SPIN_BUTTON(param->widgets[0]));
	  }
	}
      }
      else {
	gint int_value=atoi (g_list_nth_data (plist,pnum++));
	gint int_min=atoi (g_list_nth_data (plist,pnum++));
	gint int_max=atoi (g_list_nth_data (plist,pnum++));
	if (int_value<int_min) int_value=int_min;
	if (int_value>int_max) int_value=int_max;
	param->min=(gfloat)int_min;
	param->max=(gfloat)int_max;
	set_int_param(param->value,int_value);

	if (upd) {
	  if (param->widgets[0]&&GTK_IS_SPIN_BUTTON (param->widgets[0])) {
	    gulong spinfunc=get_gulong_data_from_widget(GTK_WIDGET(param->widgets[0]),"spinfunc");
	    g_signal_handler_block(param->widgets[0],spinfunc);
	    gtk_spin_button_set_range (GTK_SPIN_BUTTON (param->widgets[0]),(gdouble)param->min,(gdouble)param->max);
	    gtk_spin_button_update(GTK_SPIN_BUTTON(param->widgets[0]));
	    g_signal_handler_unblock(param->widgets[0],spinfunc);
	    gtk_spin_button_set_value (GTK_SPIN_BUTTON (param->widgets[0]),(gdouble)get_int_param(param->value));
	    gtk_spin_button_update(GTK_SPIN_BUTTON(param->widgets[0]));
	  }
	}
      }
      break;
    case LIVES_PARAM_COLRGB24:
      red=atoi (g_list_nth_data (plist,pnum++)); // red
      green=atoi (g_list_nth_data (plist,pnum++)); // green
      blue=atoi (g_list_nth_data (plist,pnum++)); // blue
      set_colRGB24_param(param->value,red,green,blue);

	if (upd) {
	  if (param->widgets[0]&&GTK_IS_SPIN_BUTTON (param->widgets[0])) {
	    gtk_spin_button_set_value (GTK_SPIN_BUTTON (param->widgets[0]),(gdouble)red);
	  }
	  if (param->widgets[1]&&GTK_IS_SPIN_BUTTON (param->widgets[1])) {
	    gtk_spin_button_set_value (GTK_SPIN_BUTTON (param->widgets[1]),(gdouble)green);
	  }
	  if (param->widgets[2]&&GTK_IS_SPIN_BUTTON (param->widgets[2])) {
	    gtk_spin_button_set_value (GTK_SPIN_BUTTON (param->widgets[2]),(gdouble)blue);
	  }
	}
	break;
    }
  }
}


void toggle_set_default_value (GtkToggleButton *button, GtkSpinButton *spinbutton) {
  lives_param_t *param;
  gulong satfunc;
  
  if (gtk_toggle_button_get_active (button)) return;
  param=(lives_param_t *)g_object_get_data (G_OBJECT (button),"param");
  satfunc=get_gulong_data_from_widget(GTK_WIDGET(spinbutton),"satfunc");
  g_signal_handler_block(spinbutton,satfunc);
  if (param->dp) {
    gtk_spin_button_set_value (GTK_SPIN_BUTTON (spinbutton),(gdouble)(get_float_param(param->def)));
  }
  else {
    gtk_spin_button_set_value (GTK_SPIN_BUTTON (spinbutton),(gdouble)(get_int_param(param->def)));
  }
  g_signal_handler_unblock(spinbutton,satfunc);
}

void toggle_set_default_red (GtkToggleButton *button, GtkSpinButton *spinbutton) {
  lives_param_t *param;
  lives_colRGB24_t rgb;
  gulong satfunc;

  if (gtk_toggle_button_get_active (button)) return;

  param=(lives_param_t *)g_object_get_data (G_OBJECT (button),"param");
  get_colRGB24_param(param->def,&rgb);
  satfunc=get_gulong_data_from_widget(GTK_WIDGET(spinbutton),"satfunc");
  g_signal_handler_block(spinbutton,satfunc);
  gtk_spin_button_set_value (GTK_SPIN_BUTTON (spinbutton),(gdouble)(rgb.red));
  g_signal_handler_unblock(spinbutton,satfunc);
}

void toggle_set_default_green (GtkToggleButton *button, GtkSpinButton *spinbutton) {
  lives_param_t *param;
  lives_colRGB24_t rgb;
  gulong satfunc;

  if (gtk_toggle_button_get_active (button)) return;

  param=(lives_param_t *)g_object_get_data (G_OBJECT (button),"param");
  get_colRGB24_param(param->def,&rgb);
  satfunc=get_gulong_data_from_widget(GTK_WIDGET(spinbutton),"satfunc");
  g_signal_handler_block(spinbutton,satfunc);
  gtk_spin_button_set_value (GTK_SPIN_BUTTON (spinbutton),(gdouble)(rgb.green));
  g_signal_handler_unblock(spinbutton,satfunc);
}

void toggle_set_default_blue (GtkToggleButton *button, GtkSpinButton *spinbutton) {
  lives_param_t *param;
  lives_colRGB24_t rgb;
  gulong satfunc;

  if (gtk_toggle_button_get_active (button)) return;

  param=(lives_param_t *)g_object_get_data (G_OBJECT (button),"param");
  get_colRGB24_param(param->def,&rgb);
  satfunc=get_gulong_data_from_widget(GTK_WIDGET(spinbutton),"satfunc");
  g_signal_handler_block(spinbutton,satfunc);
  gtk_spin_button_set_value (GTK_SPIN_BUTTON (spinbutton),(gdouble)(rgb.blue));
  g_signal_handler_unblock(spinbutton,satfunc);
}

void toggle_set_default_alpha (GtkToggleButton *button, GtkSpinButton *spinbutton) {
  lives_param_t *param;
  lives_colRGBA32_t rgba;
  gulong satfunc;

  if (gtk_toggle_button_get_active (button)) return;

  param=(lives_param_t *)g_object_get_data (G_OBJECT (button),"param");
  get_colRGBA32_param(param->def,&rgba);
  satfunc=get_gulong_data_from_widget(GTK_WIDGET(spinbutton),"satfunc");
  g_signal_handler_block(spinbutton,satfunc);
  gtk_spin_button_set_value (GTK_SPIN_BUTTON (spinbutton),(gdouble)(rgba.alpha));
  g_signal_handler_unblock(spinbutton,satfunc);
}


void do_onchange (GObject *object, lives_rendered_effect_t *rfx) {
  gint which=GPOINTER_TO_INT (g_object_get_data (object,"param_number"));
  gchar *com;
  GList *retvals;
  void **param_store=NULL;

  if (which<=0) {
    // backup param values, if any change here we will also set defaults
    param_store=store_rfx_params (rfx);
    com=g_strdup_printf ("smogrify fxinit_%s %d%s",rfx->name,rfx->status,param_marshall (rfx,TRUE));
    retvals=plugin_request_by_space (NULL,NULL,com);
  }
  else {
    com=g_strdup_printf ("onchange_%d%s",which,param_marshall (rfx,TRUE));
    switch (rfx->status) {
    case RFX_STATUS_BUILTIN:
      retvals=plugin_request_by_space (PLUGIN_RENDERED_EFFECTS_BUILTIN,rfx->name,com);
      break;
    case RFX_STATUS_CUSTOM:
      retvals=plugin_request_by_space (PLUGIN_RENDERED_EFFECTS_CUSTOM,rfx->name,com);
      break;
    default:
      retvals=plugin_request_by_space (PLUGIN_RENDERED_EFFECTS_TEST,rfx->name,com);
    }
  }
  if (retvals!=NULL) {
    param_demarshall (rfx,retvals,which>0);
    g_list_free (retvals);
  }
  else {
    if (which<=0&&mainw->error) {
      mainw->error=FALSE;
      do_blocking_error_dialog (g_strdup_printf ("\n\n%s\n\n",mainw->msg));
    }
  }
  g_free (com);
  if (param_store!=NULL) {
    set_rfx_params_from_store (rfx,param_store);
    rfx_params_store_free (rfx,param_store);
    g_free (param_store);
  }
}



void
on_pwcolourbutton_clicked              (GtkButton       *button,
					lives_rendered_effect_t *rfx)
{
  GtkWidget *selector;
  GtkWidget *csdialog;
  GtkWidget *okbutton;
  GtkWidget *cancelbutton;
  gint pnum=GPOINTER_TO_INT (g_object_get_data (G_OBJECT (button),"param_number"));

  csdialog = gtk_color_selection_dialog_new (_("LiVES: - Select Colour"));
  gtk_window_set_position (GTK_WINDOW (csdialog), GTK_WIN_POS_CENTER);
  gtk_window_set_transient_for(GTK_WINDOW(csdialog),GTK_WINDOW(mainw->LiVES));

  gtk_window_set_modal (GTK_WINDOW (csdialog), TRUE);
  if (palette->style&STYLE_3||palette->style==STYLE_PLAIN) {
    gtk_widget_modify_fg (csdialog, GTK_STATE_NORMAL, &palette->normal_fore);
    gtk_widget_modify_bg (csdialog, GTK_STATE_NORMAL, &palette->normal_back);
  }
  else {
    gtk_widget_modify_bg (csdialog, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }
  okbutton = GTK_COLOR_SELECTION_DIALOG (csdialog)->ok_button;
  GTK_WIDGET_SET_FLAGS (okbutton, GTK_CAN_DEFAULT);
  g_object_set_data (G_OBJECT (okbutton),"param_number",GINT_TO_POINTER (pnum));

  cancelbutton = GTK_COLOR_SELECTION_DIALOG (csdialog)->cancel_button;
  GTK_WIDGET_SET_FLAGS (cancelbutton, GTK_CAN_DEFAULT);

  selector = GTK_COLOR_SELECTION_DIALOG (csdialog)->colorsel;
  gtk_color_selection_set_has_opacity_control (GTK_COLOR_SELECTION (selector), FALSE);

  g_signal_connect (GTK_OBJECT (cancelbutton), "clicked",
		    G_CALLBACK (on_pwcolsel_cancel_clicked),
		    NULL);
  g_signal_connect (GTK_OBJECT (okbutton),
		    "clicked",
		    G_CALLBACK (on_pwcolsel_ok_clicked),
		    (gpointer)rfx);
  gtk_widget_show_all(csdialog);
}


void
on_pwcolsel_ok_clicked (GtkButton *button, lives_rendered_effect_t *rfx)
{
  GdkColor selected;
  gint pnum=GPOINTER_TO_INT (g_object_get_data (G_OBJECT (button),"param_number"));
  lives_param_t *param=&rfx->params[pnum-1];

  gtk_color_selection_get_current_color(GTK_COLOR_SELECTION (GTK_COLOR_SELECTION_DIALOG (gtk_widget_get_toplevel(GTK_WIDGET(button)))->colorsel),&selected);

  gtk_spin_button_set_value(GTK_SPIN_BUTTON(param->widgets[0]),(gdouble)(gint)((selected.red+128)/257));
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(param->widgets[1]),(gdouble)(gint)((selected.green+128)/257));
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(param->widgets[2]),(gdouble)(gint)((selected.blue+128)/257));
  gtk_widget_destroy(gtk_widget_get_toplevel(GTK_WIDGET(button)));
}


void
on_pwcolsel_cancel_clicked (GtkButton *button, gpointer user_data) {
  gtk_widget_destroy(gtk_widget_get_toplevel(GTK_WIDGET(button)));
}


// moved from callbacks.c

void
spin_act_toggle           (GtkSpinButton   *spinbutton,
			   GtkToggleButton         *togglebutton) {
  // this function allows a spinbutton to activate a togglebutton
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (togglebutton), TRUE);
}


// store gulong (function pointer) data in a widget
void add_gulong_data_to_widget(GtkWidget *widget ,gchar *data_name, gulong value) {
  int i;
  gchar *dnfull;

  for (i=0;i<(sizlong/sizint);i++) {
    dnfull=g_strdup_printf("%s%d",data_name,i);
    g_object_set_data (G_OBJECT (widget),dnfull,GINT_TO_POINTER ((gint)(value&INT_MAX)));
    value>>=sizint*8;
    g_free(dnfull);
  }
}



// retrieve gulong (function pointer) data from a widget
gulong get_gulong_data_from_widget(GtkWidget *widget ,gchar *data_name) {
  int i;
  gchar *dnfull;
  gulong result=0;
  
  for (i=0;i<(sizlong/sizint);i++) {
    dnfull=g_strdup_printf("%s%d",data_name,i);
    result<<=sizint*8;
    result+=GPOINTER_TO_INT (g_object_get_data (G_OBJECT (widget),dnfull));
    g_free(dnfull);
  }
  return result;
}
