// plugins.c
// LiVES
// (c) G. Finch 2003 <salsaman@xs4all.nl>
// see file ../COPYING for licensing details

#include <dlfcn.h>

#include "main.h"

#ifdef AUTO_RESAMPLE_RESIZE
#include "resample.h"
#endif


#include "support.h"

static gboolean list_plugins;


GList *
get_plugin_list (gchar *plugin_type, gboolean allow_nonex) {
  // returns a GList * of plugins of type plugin_type
  // returns empty list if there are no plugins of that type
  gchar *com;
  if (plugin_type==PLUGIN_THEMES) {
    com=g_strdup_printf ("smogrify list_plugins 0 1 \"%s\" 0",g_filename_from_utf8 (THEME_DIR,-1,NULL,NULL,NULL));
  }
  else if (!strcmp (plugin_type,PLUGIN_RENDERED_EFFECTS_CUSTOM_SCRIPTS)||!strcmp (plugin_type,PLUGIN_RENDERED_EFFECTS_TEST_SCRIPTS)||!strcmp (plugin_type,PLUGIN_RENDERED_EFFECTS_CUSTOM)||!strcmp (plugin_type,PLUGIN_RENDERED_EFFECTS_TEST)) {
    // look in home
    com=g_strdup_printf ("smogrify list_plugins %d 0 \"%s\" 1",allow_nonex,plugin_type);
  }
  else {
    com=g_strdup_printf ("smogrify list_plugins %d 0 \"%s%s\" 0",allow_nonex,PLUGIN_DIR,plugin_type);
  }
  list_plugins=TRUE;
  return get_plugin_result (com,"|",FALSE);
}


// this is an internal function
GList *
get_plugin_result (gchar *command, gchar *delim, gboolean allow_blanks) {
  gchar **array;
  gint bytes=0,pieces;
  int outfile_fd,i;
  gchar *msg,*buf;
  gint error;
  GList *list=NULL;
  gint count=30000000/prefs->sleep_time;  // timeout of 30 seconds
  // TODO - capture error text and display it
  gchar *outfile=g_strdup_printf ("%s/.smogplugin.%d",prefs->tmpdir,getpid());
  gchar *com=g_strconcat (command," >",outfile,NULL);
  gchar buffer[65536];

  mainw->error=FALSE;

  if ((error=system (com))!=0&&error!=126*256&&error!=256) {
    if (list_plugins==FALSE) {
      gchar *msg2;
      g_free (com);
      if (mainw->is_ready) {
	if ((outfile_fd=open(outfile,O_RDONLY))) {
	  bytes=read (outfile_fd,&buffer,65535);
	  close (outfile_fd);
	  unlink (outfile);
	  memset (buffer+bytes,0,1);
	}
	msg=g_strdup_printf (_("\nPlugin error: %s failed with code %d"),command,error/256);
	if (bytes) {
	  msg2=g_strconcat (msg,g_strdup_printf (_ (" : message was %s\n"),buffer),NULL);
	}
	else {
	  msg2=g_strconcat (msg,"\n",NULL);
	}
	d_print (msg2);
	g_free (msg2);
	g_free (msg);
      }
    }
    g_free (outfile);
    return list;
  }
  g_free (com);
  
  if (!g_file_test (outfile, G_FILE_TEST_EXISTS)) {
    g_free (outfile);
    g_free (command);
    return NULL;
  }


  while (!(outfile_fd=open(outfile,O_RDONLY))&&(count-->0||list_plugins==TRUE)) {
    g_usleep (prefs->sleep_time);
  }
  
  if (!count) {
    g_printerr (_("Plugin timed out on message %s\n"),command);
    g_free (command);
    g_free (outfile);
    g_free(buffer);
    return list;
  }
  g_free (command);
  
  bytes=read (outfile_fd,&buffer,65535);
  close (outfile_fd);
  unlink (outfile);
  g_free (outfile);
  memset (buffer+bytes,0,1);

#ifdef DEBUG_PLUGINS
  g_print("plugin msg: %s %d\n",buffer,error);
#endif
  
  if (error==256) {
    mainw->error=TRUE;
    g_snprintf (mainw->msg,512,buffer);
    return list;
  }
  
  pieces=get_token_count (buffer,delim[0]);
  array=g_strsplit(buffer,delim,pieces);
  for (i=0;i<pieces;i++) {
    if ((buf=g_strdup(array[i]))!=NULL) {
      buf=(g_strchomp (g_strchug(buf)));
      if (strlen (buf)||allow_blanks) {
	list=g_list_append (list, buf);
      }
    }
  }
  g_strfreev (array);
  return list;
}


GList *
plugin_request_with_blanks (const gchar *plugin_type, const gchar *plugin_name, const gchar *request) {
  // allow blanks in a list
  return plugin_request_common(plugin_type, plugin_name, request, "|", TRUE);
}

GList *
plugin_request (const gchar *plugin_type, const gchar *plugin_name, const gchar *request) {
  return plugin_request_common(plugin_type, plugin_name, request, "|", FALSE);
}
GList *
plugin_request_by_line (const gchar *plugin_type, const gchar *plugin_name, const gchar *request) {
  return plugin_request_common(plugin_type, plugin_name, request, "\n", FALSE);
}
GList *
plugin_request_by_space (const gchar *plugin_type, const gchar *plugin_name, const gchar *request) {
  return plugin_request_common(plugin_type, plugin_name, request, " ", FALSE);
}



GList *
plugin_request_common (const gchar *plugin_type, const gchar *plugin_name, const gchar *request, gchar *delim, gboolean allow_blanks) {
  // returns a GList of responses to -request, or NULL on error
  // by_line says whether we split on '\n' or on '|'
  GList *empty_list=NULL;
  gchar *com;

  if (plugin_type!=NULL) {
    // some types live in home directory...
    if (!strcmp (plugin_type,PLUGIN_RENDERED_EFFECTS_CUSTOM)||!strcmp (plugin_type,PLUGIN_RENDERED_EFFECTS_TEST)) {
      com=g_strdup_printf ("%s/%s/%s %s",capable->home_dir,plugin_type,plugin_name,request);
    }
    else {
      com=g_strdup_printf ("%s%s%s/%s %s",prefs->prefix_dir,PLUGIN_DIR,plugin_type,plugin_name,request);
    }
    if (plugin_name==NULL||!strlen(plugin_name)) {
      return empty_list;
    }
  }
  else com=g_strdup (request);

  list_plugins=FALSE;
  return get_plugin_result (com,delim,allow_blanks);
}




void 
do_plugin_error(const gchar *plugin_type, const gchar *plugin_name) {
  gchar *msg;

  if (plugin_name==NULL) {
    msg=g_strdup_printf(_("LiVES was unable to find its %s plugins. Please make sure you have the plugins installed in\n%s%s%s"),plugin_type,prefs->prefix_dir,PLUGIN_DIR,plugin_type);
    do_error_dialog(msg);
    g_free(msg);
    return;
  }
  msg=g_strdup_printf(_("LiVES did not receive a response from the %s plugin called '%s'.\nPlease make sure you have that plugin installed correctly in\n%s%s%s\nor switch to another plugin using Tools|Preferences|Encoding\n"),plugin_type,plugin_name,prefs->prefix_dir,PLUGIN_DIR,plugin_type);
  do_error_dialog(msg);
  g_free(msg);
}






gboolean 
open_vid_playback_plugin (const gchar *name) {
  // this is called on startup or when the user selects a new playback plugin

  // TODO - check for fixed framerates

  gchar *plugname=g_strdup_printf ("%s%s%s/%s",prefs->prefix_dir,PLUGIN_DIR,PLUGIN_VID_PLAYBACK,name);
  guint cap;
  GModule *GM=g_module_open (plugname,0);
  gboolean OK=TRUE;
  gchar *msg;

  const gchar *(*get_fps_list)(void);
  gboolean (*set_fps)(gdouble fps);

  if (GM==NULL) {
    gchar *msg=g_strdup_printf (_("\n\nFailed to open playback module %s\nError was %s\n"),plugname,g_module_error());
    set_pref ("vid_playback_plugin","none");
    do_error_dialog (msg);
    g_free (msg);
    return FALSE;
  }

  if (vid_playback_plugin!=NULL) {
    g_module_close (vid_playback_plugin->gmodule);
    g_free (vid_playback_plugin);
    mainw->fixed_fps=-1.;
  }

  vid_playback_plugin=(_vid_playback_plugin *) g_malloc (sizeof(_vid_playback_plugin));

  if (g_module_symbol (GM,"version",(gpointer)&vid_playback_plugin->version)==FALSE) {
    OK=FALSE;
  }
  if (g_module_symbol (GM,"set_palette",(gpointer)&vid_playback_plugin->set_palette)==FALSE) {
    OK=FALSE;
  }
  if (g_module_symbol (GM,"set_keyfun",(gpointer)&vid_playback_plugin->set_keyfun)==FALSE) {
    OK=FALSE;
  }
  if (g_module_symbol (GM,"get_capabilities",(gpointer)&vid_playback_plugin->get_capabilities)==FALSE) {
    OK=FALSE;
  }
  if (g_module_symbol (GM,"init_screen",(gpointer)&vid_playback_plugin->init_screen)==FALSE) {
    OK=FALSE;
  }
  if (g_module_symbol (GM,"render_frame",(gpointer)&vid_playback_plugin->render_frame)==FALSE) {
    OK=FALSE;
  }
  if (g_module_symbol (GM,"free_frame",(gpointer)&vid_playback_plugin->free_frame)==FALSE) {
    OK=FALSE;
  }
  if (g_module_symbol (GM,"send_keycodes",(gpointer)&vid_playback_plugin->send_keycodes)==FALSE) {
    OK=FALSE;
  }
  if (g_module_symbol (GM,"exit_screen",(gpointer)&vid_playback_plugin->exit_screen)==FALSE) {
    OK=FALSE;
  }

  if (OK==FALSE) {
    g_free (vid_playback_plugin);
    vid_playback_plugin=NULL;
    g_module_close (GM);
    return FALSE;
  }

  // try different palettes until we get TRUE
  if ((*vid_playback_plugin->set_palette)("UYVY")==FALSE) {
    if ((*vid_playback_plugin->set_palette)("RGB24LE")==FALSE) {
      if ((*vid_playback_plugin->set_palette)("YUV420")==FALSE) {
	g_free (vid_playback_plugin);
	vid_playback_plugin=NULL;
	g_module_close (GM);
	d_print (_("failed to agree a palette\n"));
	return FALSE;
      }
      mainw->ext_palette=EXTP_YUV420;
    }
    else {
      mainw->ext_palette=EXTP_RGB24LE;
    }
  }
  else {
    mainw->ext_palette=EXTP_UYVY;
  }
  cap=(*vid_playback_plugin->get_capabilities)();
  vid_playback_plugin->can_resize=cap&1;
  if (!vid_playback_plugin->can_resize&&mainw->ext_palette==EXTP_UYVY) {
    // UYVY must resize
    if ((*vid_playback_plugin->set_palette)("RGB24LE")==FALSE) {
      g_free (vid_playback_plugin);
      vid_playback_plugin=NULL;
      g_module_close (GM);
      d_print (_("UYVY must resize\n"));
      return FALSE;
    }
    mainw->ext_palette=EXTP_RGB24LE;
    cap=(*vid_playback_plugin->get_capabilities)();
    vid_playback_plugin->can_resize=cap&1;
  }
  vid_playback_plugin->fwidth=vid_playback_plugin->fheight=-1;
  vid_playback_plugin->can_return=cap&2;
  if (!(vid_playback_plugin->local_display=!(cap&8))) {
    // no display
    vid_playback_plugin->fwidth=vid_playback_plugin->fheight=0;
  }
  if (cap&4) {
    // fixed fps
    if (g_module_symbol (GM,"get_fps_list",(gpointer)&get_fps_list)) {
      

      // TODO - call get_fps_list - show choice to user
      // for no_display, prompt also for (fixed) screen size
      // usually we don't wanna start a new stream when we switch clips

      mainw->fixed_fps=24.0;
      if (!g_module_symbol (GM,"set_fps",(gpointer)&set_fps)) {
	g_free (vid_playback_plugin);
	vid_playback_plugin=NULL;
	g_module_close (GM);
	d_print (_("Unable to set framerate of plugin\n"));
	return FALSE;
      }
      if (!((*set_fps) (mainw->fixed_fps))) {
	d_print (_("Unable to set framerate of plugin\n"));
      }
    }
  }

  (*vid_playback_plugin->set_keyfun)(&pl_key_function);

  g_snprintf (vid_playback_plugin->name,256,"%s",name);
  vid_playback_plugin->gmodule=GM;
  cached_key=cached_mod=0;
  msg=g_strdup_printf(_("*** Using %s plugin for fs playback, agreed to use palette type %d. ***\n"),name,mainw->ext_palette);
  d_print (msg);
  g_free (msg);
  return TRUE;
}


void vid_playback_plugin_exit (void) {
  // external plugin
  if (!(mainw->uyvy_frame==NULL)) {
    g_free(mainw->uyvy_frame);
    mainw->uyvy_frame=NULL;
  }
  if (!(mainw->yuv_frame==NULL)) {
    g_free(mainw->yuv_frame);
    mainw->yuv_frame=NULL;
  }
  (*vid_playback_plugin->exit_screen)(mainw->ptr_x,mainw->ptr_y);
  mainw->ext_playback=FALSE;
  mainw->ext_keyboard=FALSE;
}

gboolean 
check_encoder_restrictions (gboolean get_extension) {
  // TODO - allow auto-resampling - resizing before encoding
  gchar **checks;
  gchar **array=NULL;
  gchar **array2;
  gint pieces,numtok;
  gboolean calc_aspect=FALSE;
  gchar aspect_buffer[512];
  gint hblock=1,vblock=1;
  int i,r;
  GList *ofmt_all=NULL;
  gboolean sizer=FALSE;

  // for auto resizing/resampling
  gdouble best_fps=0.;
  gint best_arate=0;
  gint width=cfile->hsize;
  gint height=cfile->vsize;

  gdouble best_fps_delta=0.;
  gint best_arate_delta=0;
  gboolean allow_aspect_override=FALSE;

#ifdef AUTO_RESAMPLE_RESIZE
  gint best_fps_num=0,best_fps_denom=0;
#else
  gboolean must_override=FALSE;
#endif
  // TODO - allow lists for size

  g_snprintf (prefs->encoder.of_restrict,5,"none");
  if (!((ofmt_all=plugin_request_by_line(PLUGIN_ENCODERS,prefs->encoder.name,"get_formats"))==NULL)) {
    // get any restrictions for the current format
    for (i=0;i<g_list_length(ofmt_all);i++) {
      if ((numtok=get_token_count (g_list_nth_data (ofmt_all,i),'|'))>2) {
	array=g_strsplit (g_list_nth_data (ofmt_all,i),"|",-1);
	if (!strcmp(array[0],prefs->encoder.of_name)) {
	  if (numtok>4) {
	    g_snprintf(prefs->encoder.of_def_ext,16,"%s",array[4]);
	  }
	  else {
	    memset (prefs->encoder.of_def_ext,0,1);
	  }
	  if (numtok>3) {
	    g_snprintf(prefs->encoder.of_restrict,128,"%s",array[3]);
	  }
	  else {
	    g_snprintf(prefs->encoder.of_restrict,128,"none");
	  }
	  prefs->encoder.of_allowed_acodecs=atoi (array[2]);
	  g_list_free (ofmt_all);
	  break;
	}
      }
    }
  }

  if (get_extension) return TRUE; // just wanted file extension

  
  if (mainw->save_with_sound&&prefs->encoder.audio_codec!=AUDIO_CODEC_NONE) {
    if (!(prefs->encoder.of_allowed_acodecs&(1<<prefs->encoder.audio_codec))) {
      do_encoder_acodec_error();
      return FALSE;
    }
  }

  if (!strcmp(prefs->encoder.of_restrict,"none")) {
    return TRUE;
  }

  pieces=get_token_count (prefs->encoder.of_restrict,',');
  checks=g_strsplit(prefs->encoder.of_restrict,",",pieces);

  for (r=0;r<pieces;r++) {
    // check each restriction in turn

    if (!strncmp (checks[r],"fps=",4)) {
      gdouble allowed_fps;
      gint mbest_num=0,mbest_denom=0;
      gint numparts;
      gchar *fixer;

      best_fps_delta=1000000000.;
      array=g_strsplit(checks[r],"=",2);
      numtok=get_token_count (array[1],';');
      array2=g_strsplit(array[1],";",numtok);
      for (i=0;i<numtok;i++) {
	mbest_num=mbest_denom=0;
	if ((numparts=get_token_count (array2[i],':'))>1) {
#ifdef AUTO_RESAMPLE_RESIZE
	  gchar **array3=g_strsplit(array2[i],":",2);
	  mbest_num=atoi (array3[0]);
	  mbest_denom=atoi (array3[1]);
	  g_strfreev (array3);
#endif
	  if (mbest_denom==0) continue;
	  allowed_fps=(mbest_num*1.)/(mbest_denom*1.);
	}
	else allowed_fps=g_strtod (array2[i],NULL);

	// convert to 8dp
	fixer=g_strdup_printf ("%.8f %.8f",allowed_fps,cfile->fps);
	g_free (fixer);

	if (allowed_fps>=cfile->fps) {
	  if (allowed_fps-cfile->fps<best_fps_delta) {
	    best_fps_delta=allowed_fps-cfile->fps;
#ifdef AUTO_RESAMPLE_RESIZE
	    if (mbest_denom>0) {
	      best_fps_num=mbest_num;
	      best_fps_denom=mbest_denom;
	      best_fps=0.;
	      cfile->ratio_fps=TRUE;
	    }
	    else {
	      best_fps_num=best_fps_denom=0;
#endif
	      best_fps=allowed_fps;
	      cfile->ratio_fps=FALSE;
#ifdef AUTO_RESAMPLE_RESIZE
	    }
#endif
	  }
	}
	else if ((best_fps_denom==0&&allowed_fps>best_fps)||(best_fps_denom>0&&allowed_fps>(best_fps_num*1.)/(best_fps_denom*1.))) {
	  if (mbest_denom>0) {
	    best_fps_num=mbest_num;
	    best_fps_denom=mbest_denom;
	    best_fps=0.;
	    cfile->ratio_fps=TRUE;
	  }
	  else {
	    best_fps=allowed_fps;
	    best_fps_num=best_fps_denom=0;
	    cfile->ratio_fps=FALSE;
	  }
	}
      }
      g_strfreev(array2);
#ifndef AUTO_RESAMPLE_RESIZE
      if (best_fps_delta&&best_fps>0.) {
	do_encoder_fps_error(best_fps);
	g_strfreev(array);
	g_strfreev(checks);
	return FALSE;
      }
#endif
    }

    if (!strncmp (checks[r],"size=",5)) {
      // TODO - allow list for size
      array=g_strsplit(checks[r],"=",2);
      array2=g_strsplit(array[1],"x",2);
      width=atoi (array2[0]);
      height=atoi (array2[1]);
      g_strfreev(array2);
      sizer=TRUE;
    }

    if (!strncmp (checks[r],"arate=",6)&&mainw->save_with_sound&&prefs->encoder.audio_codec!=AUDIO_CODEC_NONE&&(cfile->arate*cfile->achans*cfile->asampsize)) {
      // we only perform this test if we are encoding with audio
      // find next highest allowed rate from list,
      // if none are higher, use the highest
      gint allowed_arate;
      best_arate_delta=1000000000;

      array=g_strsplit(checks[r],"=",2);
      numtok=get_token_count (array[1],';');
      array2=g_strsplit(array[1],";",numtok);
      for (i=0;i<numtok;i++) {
	allowed_arate=atoi (array2[i]);
	if (allowed_arate>=cfile->arate) {
	  if (allowed_arate-cfile->arate<best_arate_delta) {
	    best_arate_delta=allowed_arate-cfile->arate;
	    best_arate=allowed_arate;
	  }
	}
	else if (allowed_arate>best_arate) best_arate=allowed_arate;
      }
      g_strfreev(array2);

      if (!capable->has_sox) {
	do_encoder_sox_error();
	g_strfreev(array);
	g_strfreev(checks);
	return FALSE;
      }
      
#ifndef AUTO_RESAMPLE_RESIZE
      if ((best_arate_delta&best_arate)>0) {
	do_encoder_arate_error(best_arate);
	g_strfreev(array);
	g_strfreev(checks);
	return FALSE;
	}
#endif
    }

    if (!strncmp (checks[r],"hblock=",7)) {
      // width must be a multiple of this
      array=g_strsplit(checks[r],"=",2);
      hblock=atoi (array[1]);
      width=(gint)(width/hblock+.5)*hblock;
    }

    if (!strncmp (checks[r],"vblock=",7)) {
      // height must be a multiple of this
      array=g_strsplit(checks[r],"=",2);
      vblock=atoi (array[1]);
      height=(gint)(height/vblock+.5)*vblock;
    }

    if (!strncmp (checks[r],"aspect=",7)) {
      // we calculate the nearest smaller frame size using aspect, 
      // hblock and vblock
      calc_aspect=TRUE;
      array=g_strsplit(checks[r],"=",2);
      g_snprintf (aspect_buffer,512,"%s",array[1]);
    }
    
    /// end restrictions
    if (!(array==NULL)) {
      g_strfreev(array);
      array=NULL;
    }
  }

  g_strfreev(checks);

  if (calc_aspect&&!sizer) {
    // we calculate this last, after getting hblock and vblock sizes
    gchar **array3;
    gdouble allowed_aspect;

    // aspect restricted encoders don't seem to like odd frame sizes
    if (hblock==1) hblock++;
    if (vblock==1) vblock++;

    numtok=get_token_count (aspect_buffer,';');
    array2=g_strsplit(aspect_buffer,";",numtok);
    
    height=width=0;

    for (i=0;i<numtok;i++) {
      array3=g_strsplit(array2[i],":",2);
      allowed_aspect=(gdouble)(atoi (array3[0]))/(gdouble)(atoi (array3[1]));
      g_strfreev(array3);
      minimise_aspect_delta (allowed_aspect,hblock,vblock,cfile->hsize,cfile->vsize,&width,&height);
    }
    g_strfreev(array2);

    if (((cfile->hsize*1.)/(hblock*1.))==(int)((cfile->hsize*1.)/(hblock*1.))&&((cfile->vsize*1.)/(vblock*1.))==(int)((cfile->vsize*1.)/(vblock*1.))) allow_aspect_override=TRUE;

#ifndef AUTO_RESAMPLE_RESIZE
    if ((width!=cfile->hsize||height!=cfile->vsize)&&width*height>0) {
      if (!capable->has_convert) {
	if (allow_aspect_override) must_override=TRUE;
	do_error_dialog (_ ("Unable to resize, please install imageMagick\n"));
	return FALSE;
      }
      if (!must_override) {
	if (!do_encoder_aspect_error(width,height,allow_aspect_override)) return FALSE;
      }
      width=cfile->hsize;
      height=cfile->vsize;
    }
#endif
  }

#ifndef AUTO_RESAMPLE_RESIZE
  if ((width!=cfile->hsize||height!=cfile->vsize)&&width*height>0) {
    if (!capable->has_convert) {
      do_error_dialog (_ ("Unable to resize, please install imageMagick\n"));
      return FALSE;
    }
    do_encoder_size_error(width,height);
    return FALSE;
  }
#endif
  
  
#ifdef AUTO_RESAMPLE_RESIZE
  
  if (sizer) allow_aspect_override=FALSE;
  if (((width!=cfile->hsize||height!=cfile->vsize)&&width*height>0)||(best_fps_delta>(.0005*prefs->ignore_tiny_fps_diffs)&&(best_fps>0.||best_fps_denom>0))||(best_arate_delta&best_arate)>0) {
    gboolean ofx1_bool=mainw->fx1_bool;
    mainw->fx1_bool=FALSE;
    if (!capable->has_convert) {
      if (allow_aspect_override) {
	width=cfile->hsize;
	height=cfile->vsize;
      }
      else {
	do_error_dialog (_ ("Unable to resize, please install imageMagick\n"));
	return FALSE;
      }
    }
    if (best_fps_delta<(.0005*prefs->ignore_tiny_fps_diffs)) best_fps_num=best_fps_denom=0;
    if (do_encoder_restrict_dialog (width,height,best_fps,best_fps_num,best_fps_denom,best_arate,allow_aspect_override)) {
      if (!mainw->fx1_bool) {
	width=cfile->hsize;
	height=cfile->vsize;
      }
      if (!auto_resample_resize (width,height,best_fps,best_fps_num,best_fps_denom,best_arate)) {
	mainw->fx1_bool=ofx1_bool;
	return FALSE;
      }
    }
    else {
      mainw->fx1_bool=ofx1_bool;
      return FALSE;
    }
  }
#endif
  return TRUE;
}


gboolean check_rfx_for_lives (lives_rendered_effect_t *rfx) {
  // check that an RFX is suitable for loading (cf. check_for_lives in effects-livido.c)
  if (rfx->num_in_channels==2&&rfx->props&RFX_PROPS_MAY_RESIZE) {
    d_print (g_strdup_printf (_ ("Failed to load %s, transitions may not resize.\n"),rfx->name));
    return FALSE;
  }
  return TRUE;
}


void render_fx_get_params (lives_rendered_effect_t *rfx, const gchar *plugin_name, gshort status) {
  // create lives_param_t array from plugin supplied values
  GList *parameter_list;
  int param_idx,i;
  lives_param_t *cparam;
  gchar **param_array;

  switch (status) {
  case RFX_STATUS_BUILTIN:
    parameter_list=plugin_request_by_line (PLUGIN_RENDERED_EFFECTS_BUILTIN,plugin_name,"get_parameters");
    break;
  case RFX_STATUS_CUSTOM:
    parameter_list=plugin_request_by_line (PLUGIN_RENDERED_EFFECTS_CUSTOM,plugin_name,"get_parameters");
    break;
  default:
    parameter_list=plugin_request_by_line (PLUGIN_RENDERED_EFFECTS_TEST,plugin_name,"get_parameters");
    break;
  }

  if (parameter_list==NULL) {
    rfx->num_params=0;
    rfx->params=NULL;
    return;
  }

  rfx->num_params=g_list_length (parameter_list);
  rfx->params=g_malloc (rfx->num_params*sizeof(lives_param_t));

  for (param_idx=0;param_idx<rfx->num_params;param_idx++) {
    // TODO - will need new g_strsplit which can handle | in strings
    // TODO - error check
      param_array=g_strsplit(g_list_nth_data(parameter_list,param_idx),"|",-1);
      cparam=&rfx->params[param_idx];
      cparam->name=g_strdup (param_array[0]);

#ifdef DEBUG_RENDER_FX_P
      g_print("Got parameter %s\n",cparam->name);
#endif
      cparam->group=atoi (param_array[1]);
      cparam->dp=0;

      if (!strncmp (param_array[2],"num",3)) {
	cparam->dp=atoi (param_array[2]+3);
	cparam->type=LIVES_PARAM_NUM;
      }
      else if (!strncmp (param_array[2],"bool",4)) {
	cparam->type=LIVES_PARAM_BOOL;
	cparam->dp=0;
      }
      else if (!strncmp (param_array[2],"colRGB24",8)) {
	cparam->type=LIVES_PARAM_COLRGB24;
	cparam->dp=0;
      }

      if (cparam->dp) {
	gfloat val=(gfloat)g_strtod (param_array[3],NULL);
	cparam->value=g_malloc(sizfloat);
	cparam->def=g_malloc(sizfloat);
	set_float_param(cparam->def,val);
	set_float_param(cparam->value,val);
	cparam->min=(gfloat)g_strtod (param_array[4],NULL);
	cparam->max=(gfloat)g_strtod (param_array[5],NULL);
      }
      else {
	if (cparam->type==LIVES_PARAM_COLRGB24) {
	  gint red=atoi (param_array[3]);
	  gint green=atoi (param_array[4]);
	  gint blue=atoi (param_array[5]);
	  cparam->value=g_malloc(sizint*3);
	  cparam->def=g_malloc(sizint*3);
	  set_colRGB24_param(cparam->def,red,green,blue);
	  set_colRGB24_param(cparam->value,red,green,blue);
	}
	else {
	  gint val=atoi (param_array[3]);
	  cparam->value=g_malloc(sizint);
	  cparam->def=g_malloc(sizint);
	  set_int_param(cparam->def,val);
	  set_int_param(cparam->value,val);
	  if (cparam->type==LIVES_PARAM_BOOL) {
	    cparam->min=0;
	    cparam->max=1;
	  }
	  else {
	    cparam->min=atoi (param_array[4]);
	    cparam->max=atoi (param_array[5]);
	  }
	}
      }
      for (i=0;i<MAX_PARAM_WIDGETS;i++) {
	cparam->widgets[i]=NULL;
      }
      cparam->onchange=FALSE;
      cparam->extra=NULL;
      g_strfreev (param_array);
  }
  g_list_free (parameter_list);
}


void sort_rfx_array (lives_rendered_effect_t *in, gint num) {
  // sort rfx array into UTF-8 order by menu entry
  int i;
  int start=1,min_val=0;
  gboolean used[num];
  gint sorted=1;
  gchar *min_string=NULL;
  lives_rendered_effect_t *rfx;

  for (i=0;i<num;i++) {
    used[i]=FALSE;
  }

  rfx=mainw->rendered_fx=(lives_rendered_effect_t *)g_malloc ((num+1)*sizeof(lives_rendered_effect_t));

  rfx->name=g_strdup (in[0].name);
  rfx->menu_text=g_strdup (in[0].menu_text);
  rfx->action_desc=g_strdup (in[0].action_desc);
  rfx->props=in[0].props;
  rfx->num_params=0;
  rfx->min_frames=1;
  rfx->params=NULL;
  rfx->extra=NULL;

  while (sorted<=num) {
    for (i=start;i<=num;i++) {
      if (!used[i-1]) {
	if (min_string==NULL) {
	  min_string=g_utf8_collate_key (in[i].menu_text,strlen (in[i].menu_text));
	  min_val=i;
	}
	else if (strcmp (min_string,g_utf8_collate_key(in[i].menu_text,strlen (in[i].menu_text)))==1) {
	  g_free (min_string);
	  min_string=g_utf8_collate_key (in[i].menu_text,strlen (in[i].menu_text));
	  min_val=i;
	}
      }
    }
    rfx_copy (&in[min_val],&mainw->rendered_fx[sorted++],FALSE);
    used[min_val-1]=TRUE;
    if (min_string!=NULL) g_free (min_string);
    min_string=NULL;
  }
  for (i=0;i<=num;i++) {
    rfx_free(&in[i]);
  }
}


void rfx_copy (lives_rendered_effect_t *src, lives_rendered_effect_t *dest, gboolean full) {
  // Warning, does not copy all fields (full will do that)
  dest->name=g_strdup (src->name);
  dest->menu_text=g_strdup (src->menu_text);
  dest->action_desc=g_strdup (src->action_desc);
  dest->min_frames=src->min_frames;
  dest->num_in_channels=src->num_in_channels;
  dest->status=src->status;
  dest->props=src->props;
  if (!full) return;
}


void rfx_free(lives_rendered_effect_t *rfx) {
  // should only be called from rebuild_rfx
  // or sort_rfx_array
  int i;

  g_free(rfx->name);
  g_free(rfx->menu_text);
  g_free(rfx->action_desc);
  for (i=0;i<rfx->num_params;i++) {
    g_free(rfx->params[i].name);
    g_free(rfx->params[i].def);
    g_free(rfx->params[i].value);
  }
  if (rfx->params!=NULL) {
    g_free(rfx->params);
  }
  if (rfx->extra!=NULL) {
    free(rfx->extra);
  }
}


void rfx_free_all (void) {
  int i;
  for (i=0;i<=mainw->num_rendered_effects_builtin+mainw->num_rendered_effects_custom+mainw->num_rendered_effects_test;i++) {
    rfx_free(&mainw->rendered_fx[i]);
  }
  g_free(mainw->rendered_fx);
}


void param_copy (lives_param_t *src, lives_param_t *dest, gboolean full) {
  dest->name=g_strdup (src->name);
  dest->group=src->group;
  dest->onchange=src->onchange;
  dest->type=src->type;
  dest->dp=src->dp;
  dest->min=src->min;
  dest->max=src->max;
  dest->extra=NULL;

  switch (dest->type) {
  case LIVES_PARAM_BOOL:
    dest->dp=0;
  case LIVES_PARAM_NUM:
    if (!dest->dp) {
      dest->def=g_malloc (sizint);
      memcpy (dest->def,src->def,sizint);
    }
    else {
      dest->def=g_malloc (sizfloat);
      memcpy (dest->def,src->def,sizfloat);
    }
    break;
  case LIVES_PARAM_COLRGB24:
    dest->def=g_malloc (3*sizint);
    memcpy (dest->def,src->def,3*sizint);
    break;
  }

  if (!full) return;
  // TODO - copy value, copy widgets

}


gint get_params_size (lives_rendered_effect_t *rfx) {
  int i;
  gint tot=0;

  for (i=0;i<rfx->num_params;i++) {
    switch (rfx->params[i].type) {
    case LIVES_PARAM_NUM:
    case LIVES_PARAM_BOOL:
      if (rfx->params[i].dp) tot+=sizfloat;
      else tot+=sizint;
      break;
    case LIVES_PARAM_COLRGB24:
      tot+=3*sizint;
      break;
    }
  }
  return tot;
}


void **store_rfx_params (lives_rendered_effect_t *rfx) {
  int i;
  void **store=(void **)g_malloc (rfx->num_params*sizeof(void *));

  for (i=0;i<rfx->num_params;i++) {
    switch (rfx->params[i].type) {
    case LIVES_PARAM_BOOL:
    case LIVES_PARAM_NUM:
      if (rfx->params[i].dp) {
	store[i]=g_malloc (sizfloat);
	memcpy (store[i],rfx->params[i].value,sizfloat);
      }
      else {
	store[i]=g_malloc (sizint);
	memcpy (store[i],rfx->params[i].value,sizint);
      }
      break;
    case LIVES_PARAM_COLRGB24:
      store[i]=g_malloc (3*sizint);
      memcpy (store[i],rfx->params[i].value,3*sizint);
      break;
    }
  }
  return store;
}

void set_rfx_params_from_store (lives_rendered_effect_t *rfx, void **param_store) {
    // check each parameter value against the stored value. If it has changed
    // then change the default. This gives a user controllable way of setting
    // defaults in the init trigger

  int i;

  lives_colRGB24_t rgbs,rgbv,rgbd;
  
  for (i=0;i<rfx->num_params;i++) {
    switch (rfx->params[i].type) {
    case LIVES_PARAM_BOOL:
    case LIVES_PARAM_NUM:
      if (rfx->params[i].dp) {
	if (get_float_param (param_store[i])!=get_float_param (rfx->params[i].value)) set_float_param (rfx->params[i].def,get_float_param (rfx->params[i].value));
      }
      else {
	if (get_int_param (param_store[i])!=get_int_param (rfx->params[i].value)) set_int_param (rfx->params[i].def,get_int_param (rfx->params[i].value));
      }
      break;
    case LIVES_PARAM_COLRGB24:
      get_colRGB24_param (param_store[i],&rgbs);
      get_colRGB24_param (rfx->params[i].value,&rgbv);
      get_colRGB24_param (rfx->params[i].def,&rgbd);
      if (rgbs.red!=rgbv.red) rgbd.red=rgbv.red;
      if (rgbs.green!=rgbv.green) rgbd.green=rgbv.green;
      if (rgbs.blue!=rgbv.blue) rgbd.blue=rgbv.blue;
      set_colRGB24_param (rfx->params[i].def,rgbd.red,rgbd.green,rgbd.blue);
    }
  } 
}

void rfx_params_store_free (lives_rendered_effect_t *rfx, void **store) {
  int i;
  for (i=0;i<rfx->num_params;i++) {
    g_free (store[i]);
  }
}


gboolean get_bool_param(void *value) {
  gboolean ret;
  memcpy(&ret,value,sizint);
  return ret;
}

gint get_int_param(void *value) {
  gint ret;
  memcpy(&ret,value,sizint);
  return ret;
}

gfloat get_float_param(void *value) {
  gfloat ret;
  memcpy(&ret,value,sizfloat);
  return ret;
}

void get_colRGB24_param(void *value, lives_colRGB24_t *rgb) {
  memcpy(rgb,value,sizint*3);
}

void get_colRGBA32_param(void *value, lives_colRGBA32_t *rgba) {
  memcpy(rgba,value,sizint*4);
}

void set_bool_param(void *value, const gboolean _const) {
  set_int_param(value,_const);
}

void set_int_param(void *value, const gint _const) {
  memcpy(value,&_const,sizint);
}
void set_float_param(void *value, const gfloat _const) {
  memcpy(value,&_const,sizfloat);

}

void set_colRGB24_param(void *value, gint red, gint green, gint blue) {
  lives_colRGB24_t *rgbp=(lives_colRGB24_t *)value;

  if (red<0) red=0;
  if (red>255) red=255;
  if (green<0) green=0;
  if (green>255) green=255;
  if (blue<0) blue=0;
  if (blue>255) blue=255;

  memcpy(&rgbp->red,&red,sizint);
  memcpy(&rgbp->green,&green,sizint);
  memcpy(&rgbp->blue,&blue,sizint);
}

void set_colRGBA32_param(void *value, gint red, gint green, gint blue, gint alpha) {
  lives_colRGBA32_t *rgbap=(lives_colRGBA32_t *)value;
  memcpy(&rgbap->red,&red,sizint);
  memcpy(&rgbap->green,&green,sizint);
  memcpy(&rgbap->blue,&blue,sizint);
  memcpy(&rgbap->alpha,&alpha,sizint);
}




///////////////////////////////////////////////////////////////



gint find_rfx_plugin_by_name (gchar *name, gshort status) {
  int i;
  for (i=1;i<mainw->num_rendered_effects_builtin+mainw->num_rendered_effects_custom+mainw->num_rendered_effects_test;i++) {
    if (mainw->rendered_fx[i].name!=NULL&&!strcmp (mainw->rendered_fx[i].name,name)&&mainw->rendered_fx[i].status==status) return (gint)i;
  }
  return -1;
}

