// plugins.h
// LiVES
// (c) G. Finch 2003 <salsaman@xs4all.nl>
// see file ../COPYING for licensing details
#include <gtk/gtk.h>
#include <gmodule.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>
#include "main.h"

#ifdef USE_LIVIDO
#ifndef __LIVIDO_H__
#include "plugins-livido.h"
#endif
#endif


GList *
get_plugin_list (gchar *plugin_type, gboolean allow_nonex);
#define PLUGIN_ENCODERS "encoders"
#define PLUGIN_VID_PLAYBACK "playback/video"


// smogrify handles the directory differently for themes
#define PLUGIN_THEMES "themes"


#ifdef USE_LIVIDO
#define PLUGIN_LIVIDO_FX_BUILTIN "effects/realtime/livido"

// current limit is 64 as we use a guint64 bit mask
// TODO - use array to multiply this
#define MAX_LIVIDO_PLUGINS 64
#endif

#define USE_EXT_RENDER


GList *
plugin_request (const gchar *plugin_type, const gchar *plugin_name, const gchar *request);

GList *
plugin_request_with_blanks (const gchar *plugin_type, const gchar *plugin_name, const gchar *request);

GList *
plugin_request_by_line (const gchar *plugin_type, const gchar *plugin_name, const gchar *request);

GList *
plugin_request_by_space (const gchar *plugin_type, const gchar *plugin_name, const gchar *request);

// TODO - make 1 function
GList *
plugin_request_common (const gchar *plugin_type, const gchar *plugin_name, const gchar *request, gchar *delim, gboolean allow_blanks);

GList *
get_plugin_result (gchar *, gchar *, gboolean); /* private function */


void 
do_plugin_error(const gchar *plugin_type, const gchar *plugin_name_or_null);


typedef struct {
  // playback
  gchar name[64];
  GModule *gmodule;

  const gchar * (*version) (void);
  void (*set_keyfun)(void (*fn)(gboolean, guint16, guint16));
  guint (*set_palette)(const gchar *p_name);
  guint (*get_capabilities)(void);
  gboolean (*init_screen)(int width, int height, gboolean fullscreen);
  gboolean (*render_frame)(int hsize, int vsize, void *pixel_data, void *return_data);
  void (*free_frame)(void);
  gboolean (*send_keycodes)(void);
  void (*exit_screen)(gint16 mouse_x, gint16 mouse_y);

  gint fwidth,fheight;

  //capabilities
  gboolean can_resize;
  gboolean can_return;
  gboolean local_display;
} _vid_playback_plugin;


_vid_playback_plugin *vid_playback_plugin;


gboolean open_vid_playback_plugin (const gchar *name);
void vid_playback_plugin_exit (void);


// external rendered fx plugins (RFX plugins)
#define PLUGIN_RENDERED_EFFECTS_BUILTIN "effects/rendered"

// in the home directory
#define PLUGIN_RENDERED_EFFECTS_CUSTOM "lives/plugins/effects/rendered/custom"
#define PLUGIN_RENDERED_EFFECTS_TEST "lives/plugins/effects/rendered/test"

// rfx scripts for the SDK
#define PLUGIN_RENDERED_EFFECTS_BUILTIN_SCRIPTS "effects/RFXscripts"

// in the home directory
#define PLUGIN_RENDERED_EFFECTS_CUSTOM_SCRIPTS "lives/plugins/effects/RFXscripts/custom"
#define PLUGIN_RENDERED_EFFECTS_TEST_SCRIPTS "lives/plugins/effects/RFXscripts/test"


// max number of display widgets per parameter (currently 3 for RGB spinbuttons)
#define MAX_PARAM_WIDGETS 3


typedef struct {
  // livido style part
  gchar *name;
  gint group;
  gint type;

#define LIVES_PARAM_NUM 1
#define LIVES_PARAM_BOOL 2
#define LIVES_PARAM_COLRGB24 3
  
  gint dp;  //decimals, 0 for int and bool
  void *value;  // current values

  gfloat min;
  gfloat max;

  void *def; // default values

  void *extra;  // for future use

  //--------------------------------------------------
  // extras for LiVES

  // TODO - change to GtkWidget **widgets, terminated with a NULL
  GtkWidget *widgets[MAX_PARAM_WIDGETS]; // widgets which hold value/RGBA settings
  gboolean onchange; // is there a trigger ?
} lives_param_t;


typedef struct {
  gchar *name;  // the name of the executable (so we can run it !)
  gchar *menu_text;
  gchar *action_desc;
  gint min_frames;
  gint num_in_channels;
  gshort status;
#define RFX_STATUS_BUILTIN 0 // factory presets
#define RFX_STATUS_CUSTOM 1 // custom effects in the custom menu
#define RFX_STATUS_TEST 2 // test effects in the advanced menu
#define RFX_STATUS_ANY 3 // indicates free choice of statuses

  // these are only used when prompting for a name
#define RFX_STATUS_COPY 128 // indicates a copy operation to test
#define RFX_STATUS_RENAME 129 // indicates a copy operation to test

  guint32 props;
#define RFX_PROPS_SLOW        0x0001  // hint to GUI
#define RFX_PROPS_MAY_RESIZE  0x0002 // is a tool


#define RFX_PROPS_RESERVED1   0x1000
#define RFX_PROPS_RESERVED2   0x2000
#define RFX_PROPS_RESERVED3   0x4000
#define RFX_PROPS_AUTO_BUILT  0x8000

  GtkWidget *menuitem;  // the menu item which activates this effect
  gint num_params;
  lives_param_t *params;
  void *extra;  // for future use */
} lives_rendered_effect_t;

gboolean check_rfx_for_lives (lives_rendered_effect_t *);


void render_fx_get_params (lives_rendered_effect_t *, const gchar *plugin_name, gshort status);

void sort_rfx_array (lives_rendered_effect_t *in_array, gint num_elements);

gint find_rfx_plugin_by_name (gchar *name, gshort status);

void rfx_copy (lives_rendered_effect_t *src, lives_rendered_effect_t *dest, gboolean full);

void rfx_free(lives_rendered_effect_t *rfx);

void rfx_free_all (void);

void param_copy (lives_param_t *src, lives_param_t *dest, gboolean full);


typedef struct {
  gshort red;
  gshort green;
  gshort blue;
} lives_colRGB24_t;

typedef struct {
  gshort red;
  gshort green;
  gshort blue;
  gshort alpha;
} lives_colRGBA32_t;



// TODO - move into plugins-livido.c
gboolean get_bool_param(void *value);
gint get_int_param(void *value);
gfloat get_float_param(void *value);
void get_colRGB24_param(void *value, lives_colRGB24_t *rgb);
void get_colRGBA32_param(void *value, lives_colRGBA32_t *rgba);

void set_bool_param(void *value, const gboolean _const);
void set_int_param(void *value, const gint _const);
void set_float_param(void *value, const gfloat _const);
void set_colRGB24_param(void *value, gint red, gint green, gint blue);
void set_colRGBA32_param(void *value, gint red, gint green, gint blue, gint alpha);

// get the size of all parameters in bytes for an effect
gint get_params_size (lives_rendered_effect_t *);

// return an array of parameter values
void **store_rfx_params (lives_rendered_effect_t *);
void set_rfx_params_from_store (lives_rendered_effect_t *rfx, void **store);
void rfx_params_store_free (lives_rendered_effect_t *, void **store);




