// resample.c
// LiVES
// (c) G. Finch 2004 <salsaman@xs4all.nl>
// see file ../COPYING for licensing details


// functions for reordering, resampling video and audio

#include "main.h"
#include "resample.h"
#include "support.h"
#include "callbacks.h"

#ifdef AUTO_RESAMPLE_RESIZE
gboolean auto_resample_resize (gint width,gint height,gdouble fps,gint fps_num,gint fps_denom, gint arate) {
  // do a block atomic: resample audio, then resample video/resize or joint resample/resize

  gchar *com;
  gint current_file=mainw->current_file;
  gboolean audio_resampled=FALSE;
  gboolean video_resampled=FALSE;
  gboolean video_resized=FALSE;
  gint frames=cfile->frames;

  if (arate>0&&arate!=cfile->arate) {
    cfile->undo1_int=arate;
    cfile->undo2_int=cfile->achans;
    cfile->undo3_int=cfile->asampsize;
    cfile->undo4_int=cfile->arps;
    cfile->undo1_uint=cfile->signed_endian;
    on_resaudio_ok_clicked (NULL,NULL);
    if (mainw->error) return FALSE;
    audio_resampled=TRUE;
  }

  if (fps_denom>0) {
    fps=(fps_num*1.)/(fps_denom*1.);
  }
  if (fps>0.&&fps!=cfile->fps) {
    // FPS CHANGE...
    if ((width!=cfile->hsize||height!=cfile->vsize)&&width*height>0) {
      // CHANGING SIZE..
      if (fps>cfile->fps) {
	// we will have more frames...
	// ...do resize first
	cfile->ohsize=cfile->hsize;
	cfile->ovsize=cfile->vsize;
	com=g_strdup_printf ("smogrify resize_all %s %d %d %d",cfile->handle,cfile->frames,width,height);

	cfile->progress_start=1;
	cfile->progress_end=cfile->frames;
	
	unlink(cfile->info_file);
	system(com);
	g_free (com);
	if (!(do_progress_dialog(TRUE,TRUE,_ ("Resizing")))) {
	  cfile->undo_action=UNDO_ATOMIC_RESAMPLE_RESIZE;
	  if (!audio_resampled) {
	    cfile->undo1_int=arate;
	    cfile->undo2_int=cfile->achans;
	    cfile->undo3_int=cfile->asampsize;
	    cfile->undo4_int=cfile->arps;
	    cfile->undo1_uint=cfile->signed_endian;
	  }
	  cfile->hsize=width;
	  cfile->vsize=height;
	  
	  cfile->undo1_dbl=fps;
	  cfile->undo_start=1;
	  cfile->undo_end=frames;
	  on_undo_activate(NULL,NULL);
	  return FALSE;
	}
	    
	    
	cfile->hsize=width;
	cfile->vsize=height;
	
	cfile->undo1_dbl=fps;
	cfile->undo_start=1;
	cfile->undo_end=frames;

	// now resample

	// special "cheat" mode for LiVES
	prefs->reorder_leave_back=TRUE;
	    
	on_resample_vid_ok (NULL,NULL);

	prefs->reorder_leave_back=FALSE;
	cfile->undo_action=UNDO_ATOMIC_RESAMPLE_RESIZE;
	if (mainw->error) {
	  on_undo_activate(NULL,NULL);
	  return FALSE;
	}

	video_resized=TRUE;
	video_resampled=TRUE;
      }
      else {
	// fewer frames
	// do resample *with* resize
	cfile->ohsize=cfile->hsize;
	cfile->ovsize=cfile->vsize;
	cfile->undo1_dbl=fps;

	// another special "cheat" mode for LiVES
	prefs->reorder_width=width;
	prefs->reorder_height=height;

	on_resample_vid_ok (NULL,NULL);

	prefs->reorder_width=prefs->reorder_height=0;
	cfile->undo_action=UNDO_ATOMIC_RESAMPLE_RESIZE;
	cfile->hsize=width;
	cfile->vsize=height;
	if (mainw->error) {
	  on_undo_activate(NULL,NULL);
	  return FALSE;
	}
       	video_resampled=TRUE;
	video_resized=TRUE;
      }
    }
    else {
      //////////////////////////////////////////////////////////////////////////////////
      // NO SIZE CHANGE
      // just resample...here is how we do it, just set cfile->undo1_dbl=new_fps;
      // but beware of cfile->events
      cfile->undo1_dbl=fps;
      cfile->undo_start=1;
      cfile->undo_end=cfile->frames;
      
      on_resample_vid_ok (NULL,NULL);
      
      if (audio_resampled) cfile->undo_action=UNDO_ATOMIC_RESAMPLE_RESIZE;
      if (mainw->error) {
	on_undo_activate(NULL,NULL);
	return FALSE;
      }
      //////////////////////////////////////////////////////////////////////////////////////
      video_resampled=TRUE;
    }
  }
  else {
    // NO FPS CHANGE
    if ((width!=cfile->hsize||height!=cfile->vsize)&&width*height>0) {
      // no fps change - just a normal resize
      cfile->undo_start=1;
      cfile->undo_end=cfile->frames;

      com=g_strdup_printf ("smogrify resize_all %s %d %d %d",cfile->handle,cfile->frames,width,height);
    
      cfile->progress_start=1;
      cfile->progress_end=cfile->frames;

      cfile->ohsize=cfile->hsize;
      cfile->ovsize=cfile->vsize;
      
      unlink(cfile->info_file);
      system(com);
      g_free (com);
      if (!do_progress_dialog(TRUE,TRUE,_ ("Resizing"))) {
	cfile->hsize=width;
	cfile->vsize=height;
	if (audio_resampled) cfile->undo_action=UNDO_ATOMIC_RESAMPLE_RESIZE;
	else {
	  cfile->undo_action=UNDO_RESIZABLE;
	  set_undoable (_ ("Resize"),TRUE);
	}
	on_undo_activate (NULL,NULL);
	return FALSE;
      }
      cfile->hsize=width;
      cfile->vsize=height;
      if (audio_resampled) cfile->undo_action=UNDO_ATOMIC_RESAMPLE_RESIZE;
      else {
	cfile->undo_action=UNDO_RESIZABLE;
	set_undoable (_ ("Resize"),TRUE);
      }
      video_resized=TRUE;
      switch_to_file ((mainw->current_file=0),current_file);
    }
  }

  if (cfile->undo_action==UNDO_ATOMIC_RESAMPLE_RESIZE) {
    // just in case we missed anything...
    
    set_undoable (_ ("Resample/Resize"),TRUE);
    if (!audio_resampled) {
      cfile->undo1_int=arate;
      cfile->undo2_int=cfile->achans;
      cfile->undo3_int=cfile->asampsize;
      cfile->undo4_int=cfile->arps;
      cfile->undo1_uint=cfile->signed_endian;
    }
    if (!video_resampled) {
      cfile->undo1_dbl=fps;
    }
    if (!video_resized) {
      cfile->ohsize=cfile->hsize;
      cfile->ovsize=cfile->vsize;
    }
    cfile->undo_start=1;
    cfile->undo_end=frames;
  }
  return TRUE;
}
#endif



void
quantise_events (gint rec_start)
{ 
  // this function is called immediately after recording a performance

  // when we recorded frame changes, they could have happened at any time,
  // because when the user presses a key, the frame change is immediate
  // and not in time with a normal frame slot

  // therefore, we need quantise the frame changes so that they occur only on real frame timings,
  // otherwise we will have too many events, and the video will play too slow

  // what we will do is run through all the events, and then move them to the nearest actual frame timing
  // if there are multiple frame events for one real frame, we will use the nearest one to the frame slot

  //e.g: if we have the following before quantisation (with an fps of 25)

  //  index 'frame'     actual frame                 actual time        played at
  //            1           10                       0.02                0.00
  //            2           11                       0.03                0.04
  //            3           14                       0.037               0.08

  // we would end up with:

  //            1           10                       0.00               0.00
  //            2           14                       0.04               0.04

  // note that we also need to do this with other types of events - e.g. fps changes

  // However, as we go forward we must keep track of the fps changes so that we always know when the frames should fall.
  // we also need to fill in any gaps of zeros we find (can be caused by skipped frames); we will use 'last_frame' and 'last_fps' for this


  // Note on events:
  // Frames can also be skipped on playback; so we should repeat an event if we want to make sure the associated events are acted on

  // Here, we make runs of fps changes for that reason.



  // Note 2: we make 3 arrays.
  // 1) The first array which is recorded, and it goes from 0 to cfile->perf_end-1.
  //    cfile->perf_end here just holds the size of our array
  //    It has event values and relative time (since the start of play).

  // 2) We reorder this array to a new one, putting in the events on the frame changes, and filling in any zeros.
  //    It goes from 0 (frame 1) upwards and is the same size as array 1

  //  cfile->perf_end is now set to the frame number of the last event

  // 3) The third array is created, which goes from cfile->perf_start to cfile->perf_end


  // Note also that we actually store fps*1000


  int i,j,framec;
  event *temp_frames,*temp_fps;
  gint frame;

  long double reltime=cfile->ev_origticks,abstime,last_abstime=0l;
  long double offset_frames, delta;
  gint qframe=rec_start+1;  // our quantised frame number
  gint diff_frame,last_frame=0;
  gdouble curr_fps=cfile->fps;
  long double t_stamp=0.;
  gint fps=0;
  guint last_fps=0;
  int last_fps_index=0;
  gint real_perf_end=0;
  gboolean alloc=TRUE;
  gint event_space,event_overkill;
  gboolean backfill_to_start=FALSE;
  gboolean reset_qframe=FALSE;

  // for resampling, we want to backfill to the start
  if (rec_start<0) {
    rec_start=-rec_start;
    qframe=rec_start;
    backfill_to_start=TRUE;
  }

  // first we reference the current events
  temp_frames=cfile->events[EVENT_FRAMES];
  temp_fps=cfile->events[EVENT_FPS];

  cfile->events[EVENT_FRAMES]=NULL;
  cfile->events[EVENT_FPS]=NULL;

  // now make a new event space
  
  // TODO - use predetermined size
  event_overkill=(gint)(FPS_MAX*10/(cfile->fps+1));

  if (create_event_space((cfile->perf_end-cfile->perf_start+1)*event_overkill,CREATE_EV_FRAMES|(CREATE_EV_FPS*(temp_fps!=NULL)))==FALSE) {
    cfile->perf_counter=cfile->perf_start=cfile->perf_end=0;
    return;
  }
  j=cfile->ev_fps_start-1;
  
  for (i=cfile->perf_start-1;i<cfile->perf_end;i++) {
    // walk thru frame events
    frame=(temp_frames+i)->value;
    reltime=(temp_frames+i)->reltime;
    fps=0;

    // if we had an fps change before this frame change, deal with that first
    if (temp_fps!=NULL) {
      if (j<cfile->ev_fps_end) {
      // before or simultaneous with frame change
      // fps events with a reltime of zero are ignored
	if ((temp_fps+j)->reltime<=reltime&&(temp_fps+j)->reltime!=0l&&(temp_fps+j)->value) {
	  fps=(temp_fps+j)->value;
	  // reverse play is not needed now, since we have the frames recorded
	  if (fps<0) fps=-fps;
	  i--;
	  frame=0;
	  reltime=(temp_fps+j)->reltime;
	}
	// next fps change is after the frame change, come back to it later
	else {
	  if ((temp_fps+j)->reltime>0l&&(temp_fps+j)->value) {
	    j--;
	  }
	}
      }
      j++;
    }

    // get the absolute time of our event
    abstime=reltime-mainw->origticks;

    // find our quantised frame at the current fps
    offset_frames=(abstime-t_stamp)/1000000.*curr_fps;
    diff_frame=(gint)(offset_frames);
    delta=offset_frames-(long double)diff_frame;
    
    if (reset_qframe&&frame!=0) {
      qframe+=diff_frame;
      diff_frame=0;
      t_stamp=abstime;
      curr_fps=last_fps/1000.;
      reset_qframe=FALSE;
    }
    
    // round to nearest whole frame
    if (delta>.5) {
      delta=1.-delta;
      diff_frame++;
    }
    // ignore any events before perf_start (e.g. when we resample)
    if ((frame!=0||fps!=0)&&qframe+diff_frame>=cfile->perf_start) {
      real_perf_end=qframe+diff_frame;

      // OK, we have an event !
      diff_frame-=cfile->perf_start-1;
	  
      if (frame!=0) {
	last_abstime=abstime;
	// fill any zero gaps (can be caused by skipped frames in playback)
	if (last_frame!=0||backfill_to_start==TRUE) {
	  if (prefs->q_type==Q_FILL) {
	    // fill quantisation
	    for (framec=qframe+diff_frame-1;framec>0&&(cfile->events[EVENT_FRAMES]+framec-1)->value==0; framec--) {
	      (cfile->events[EVENT_FRAMES]+framec-1)->value=frame;
	    }
	  }
	  else {
	    // smooth quantisation
	    gint length=1,prev_frame=frame;
	    for (framec=qframe+diff_frame-1;framec>0&&(cfile->events[EVENT_FRAMES]+framec-1)->value==0; framec--) {
	      length++;
	    }
	    if (framec>0) {
	      prev_frame=(cfile->events[EVENT_FRAMES]+framec-1)->value;
	    }
	    for (framec=qframe+diff_frame-1;framec>0&&(cfile->events[EVENT_FRAMES]+framec-1)->value==0; framec--) {
	      (cfile->events[EVENT_FRAMES]+framec-1)->value=frame+(gint)((prev_frame-frame)/length*(framec-qframe-diff_frame+2));
	    }
	  }
	}
	else {
	  // adjust perf_start to the real perf_start
	  cfile->perf_start+=qframe+diff_frame-1;
	}
	// write new frame into our quantised time slot
	// we will cunningly use the reltime field now to record our delta time,
	// that way we can find the nearest event

	if ((cfile->events[EVENT_FRAMES]+qframe+diff_frame-1)->reltime>delta||(cfile->events[EVENT_FRAMES]+qframe+diff_frame-1)->reltime==0) {
	  (cfile->events[EVENT_FRAMES]+qframe+diff_frame-1)->value=last_frame=frame;
	  (cfile->events[EVENT_FRAMES]+qframe+diff_frame-1)->reltime=delta;
	}
      }

      if (fps!=0) {
	// write fps event
	(cfile->events[EVENT_FPS]+qframe+diff_frame-1)->value=fps;
	// we also need to fill in any zeros, as we might skip the frame with the actual fps change...
	if (last_fps!=0) {
	  for (framec=qframe+diff_frame-1;framec>last_fps_index&&framec>0&&(cfile->events[EVENT_FPS]+framec-1)->value==0; framec--) {
	    (cfile->events[EVENT_FPS]+framec-1)->value=last_fps;
	    (cfile->events[EVENT_FPS]+framec-1)->reltime=reltime;
	  }
	}
	last_fps=fps;
	last_fps_index=qframe+diff_frame;
	// fps change - make this our new baseline for frame counting, update the virtual fps
	// we attach any fps changes to the last frame event
	// this is not 100% accurate for variable fps
	reset_qframe=TRUE;
      }
    } // end event
  } // end for()

  if (last_fps==0) {
    // no fps changes were recorded, we can drop event_fps
    if (cfile->events[EVENT_FPS]!=NULL) {
      g_free(cfile->events[EVENT_FPS]);
      cfile->events[EVENT_FPS]=NULL;
    }
  }
  else {
    // fill to end with fps - the event_pending system will ensure our fps is reset to the default once we reach the performance end
    for (i=qframe+1;i<real_perf_end;i++) {
      (cfile->events[EVENT_FPS]+i-1)->value=last_fps;
    }
  }

  if (last_frame==0) {
    g_free(cfile->events[EVENT_FRAMES]);
    cfile->events[EVENT_FRAMES]=NULL;
  }


  // now we can free temp (our record array)
  g_free(temp_frames);
  g_free(temp_fps);

  if (last_fps==0&&last_frame==0) {
    d_print(_ ("nothing recorded.\n"));
    cfile->perf_start=cfile->perf_end=cfile->perf_counter=0;
    return;
  }

  // pass 1 done
  cfile->perf_end=real_perf_end;

  // finally, we will make our real event spaces, 

  // last_frame should never be zero, but just in case, we check
  if (last_frame) {
    temp_frames=cfile->events[EVENT_FRAMES];
    cfile->events[EVENT_FRAMES]=NULL;
  }
  if (last_fps) {
    temp_fps=cfile->events[EVENT_FPS];
    cfile->events[EVENT_FPS]=NULL;
  }

  event_space=cfile->perf_end-cfile->perf_start+1;

  // make the correctly sized event spaces
  if (last_frame) {
    // frames
    if (create_event_space (event_space,CREATE_EV_FRAMES)==FALSE) {
      g_free(temp_frames);
      cfile->events[EVENT_FPS]=NULL;
      if (last_fps&&temp_fps!=NULL) {
	g_free(temp_fps);
	cfile->events[EVENT_FPS]=NULL;
      }
      alloc=FALSE;
    }
  }

  if (alloc==TRUE) {
    // fps
    if (last_fps) {
      if (create_event_space (event_space,CREATE_EV_FPS)==FALSE) {
	g_free(temp_fps);
	cfile->events[EVENT_FPS]=NULL;
	if (!(cfile->events[EVENT_FRAMES]==NULL)) {
	  g_free(temp_frames);
	  cfile->events[EVENT_FRAMES]=NULL;
	}
	alloc=FALSE;
      }
    }
  }

  if (alloc==FALSE) {
    cfile->perf_start=cfile->perf_end=cfile->perf_counter=0;
    return;
  }

  for (i=0;i<event_space;i++) {
    if (last_frame) {
      (cfile->events[EVENT_FRAMES]+i)->value=(temp_frames+i)->value;
    }
    if (last_fps) {
      (cfile->events[EVENT_FPS]+i)->value=(temp_fps+i)->value;
      (cfile->events[EVENT_FPS]+i)->reltime=(temp_fps+i)->reltime;
    }
  }

  if (last_frame) {
    g_free(temp_frames);
  }
  if (last_fps) {
    g_free(temp_fps);
  }
}





void
on_reorder_activate                (GtkMenuItem     *menuitem,
				    gpointer         user_data)
{
  gchar *msg;

  if (menuitem!=NULL&&cfile->undo_action==UNDO_RESAMPLE) {
    cfile->undo_action=UNDO_REORDER;
    if (mainw->current_file>0) {
      d_print(_ ("Reordering frames..."));
    }
    else {
      d_print(_ ("Resampling clipboard video..."));
    }
  }
  cfile->old_frames=cfile->frames;

  if ((cfile->frames=reorder_frames())<0) {
    // reordering error
    if (!(cfile->undo_action==UNDO_RESAMPLE)) {
     cfile->frames=-cfile->frames;
    }
    return;
  }

  if (cfile->start>cfile->frames) {
    cfile->start=cfile->frames;
  }

  if (cfile->end>cfile->frames) {
    cfile->end=cfile->frames;
  }

  if (!(cfile->events[EVENT_FRAMES]==NULL)) {
    g_free(cfile->events[EVENT_FRAMES]);
    cfile->events[EVENT_FRAMES]=NULL;
  }
  gtk_widget_set_sensitive(mainw->reorder,FALSE);

  // if we have no fps events, we now have no performance data
  if (cfile->events[EVENT_FPS]==NULL) {
     cfile->perf_start=cfile->perf_end=cfile->perf_counter=0;
  }
  
  cfile->undo_action=UNDO_REORDER;
  set_undoable ("Reorder",TRUE);
  cfile->changed=TRUE;

  switch_to_file(mainw->current_file,mainw->current_file);
  if (mainw->current_file>0) {
    d_print(_ ("done\n"));
    msg=g_strdup_printf(_ ("Length of video is now %d frames.\n"),cfile->frames);
  }
  else {
    msg=g_strdup_printf(_ ("Clipboard was resampled to %d frames.\n"),cfile->frames);
  }

  d_print(msg);
  g_free(msg);

}




void
on_resample_audio_activate (GtkMenuItem     *menuitem,
			    gpointer         user_data)
{
   // show the playback rate - real audio rate is cfile->arps
  mainw->fx1_val=cfile->arate;
  mainw->fx2_val=cfile->achans;
  mainw->fx3_val=cfile->asampsize;
  resaudw=create_resaudw(1);
  gtk_widget_show (resaudw->dialog);

}

void
on_resaudio_ok_clicked                      (GtkButton *button,
					     GtkEntry *entry)
{
  gchar *com,*msg;  
  gint arate,achans,asampsize,arps;
  int asigned=1,aendian=1;
  gint cur_signed,cur_endian;
  gboolean noswitch=mainw->noswitch;

  if (!(button==NULL)) {
    arps=arate=(gint)atoi (gtk_entry_get_text(GTK_ENTRY(resaudw->entry_arate)));
    achans=(gint)atoi (gtk_entry_get_text(GTK_ENTRY(resaudw->entry_achans)));
    asampsize=(gint)atoi (gtk_entry_get_text(GTK_ENTRY(resaudw->entry_asamps)));
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(resaudw->rb_unsigned))==TRUE) {
      asigned=0;
    }
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(resaudw->rb_bigend))==TRUE) {
      aendian=0;
    }

    gtk_widget_destroy (resaudw->dialog);
    mainw->noswitch=TRUE;
    while (g_main_context_iteration(NULL,FALSE));
    mainw->noswitch=noswitch;
    g_free (resaudw);
    
    if (arate<=0) {
      do_error_dialog (_ ("\n\nNew rate must be greater than 0\n"));
      return;
    }
  }
  else {
    // called from on_redo or other places
    arate=cfile->undo1_int;
    achans=cfile->undo2_int;
    asampsize=cfile->undo3_int;
    arps=cfile->undo4_int;
    asigned=!(cfile->undo1_uint&AFORM_UNSIGNED);
    aendian=!(cfile->undo1_uint&AFORM_BIG_ENDIAN);
  }

  // store old values for undo/redo
  cfile->undo1_int=cfile->arate;
  cfile->undo2_int=cfile->achans;
  cfile->undo3_int=cfile->asampsize;
  cfile->undo4_int=cfile->arps;
  cfile->undo1_uint=cfile->signed_endian;

  cur_signed=!(cfile->signed_endian&AFORM_UNSIGNED);
  cur_endian=!(cfile->signed_endian&AFORM_BIG_ENDIAN);

  if (!(arate==cfile->arate&&arps==cfile->arps&&achans==cfile->achans&&asampsize==cfile->asampsize&&asigned==cur_signed&&aendian==cur_endian)) {
    if (cfile->arps!=cfile->arate) {
      gdouble audio_stretch=(gdouble)cfile->arps/(gdouble)cfile->arate;
     // pb rate != real rate - stretch to pb rate and resample 
      unlink (cfile->info_file);
      com=g_strdup_printf ("smogrify resample_audio %s %d %d %d %d %d %d %d %d %d %d %.4f",cfile->handle,cfile->arps,cfile->achans,cfile->asampsize,cur_signed,cur_endian,arps,cfile->achans,cfile->asampsize,cur_signed,cur_endian,audio_stretch);
      system (com);
      do_progress_dialog (TRUE,FALSE,_ ("Resampling audio")); // TODO - allow cancel ??
      g_free (com);
      cfile->arate=cfile->arps=arps;
    }

    if (cfile->arate!=arate) {
      unlink (cfile->info_file);
      com=g_strdup_printf ("smogrify resample_audio %s %d %d %d %d %d %d %d %d %d %d",cfile->handle,cfile->arps,cfile->achans,cfile->asampsize,cur_signed,cur_endian,arps,achans,asampsize,asigned,aendian);
      system (com);
      do_progress_dialog (TRUE,FALSE,_ ("Resampling audio"));
      g_free (com);
    }
  }

  cfile->arate=arate;
  cfile->achans=achans;
  cfile->asampsize=asampsize;
  cfile->arps=arps;
  cfile->signed_endian=get_signed_endian (asigned, aendian);
  cfile->changed=TRUE;

  cfile->undo_action=UNDO_AUDIO_RESAMPLE;
  mainw->error=FALSE;
  reget_afilesize(cfile);

  if (cfile->afilesize==0) {
    do_error_dialog (_ ("LiVES was unable to resample the audio as requested.\n"));
    on_undo_activate (NULL,NULL);
    set_undoable ("Resample Audio",FALSE);
    mainw->error=TRUE;
    return;
  }
  set_undoable ("Resample Audio",!prefs->conserve_space);

  switch_to_file(mainw->current_file,mainw->current_file);

  msg=g_strdup_printf (_ ("Audio was resampled to %d Hz, %d channels, %d bit"),arate,achans,asampsize);
  d_print (msg);
  g_free (msg);
  if (cur_signed!=asigned) {
    if (asigned==1) {
      d_print (_ (", signed"));
    }
    else {
      d_print (_ (", unsigned"));
    }
  }
  if (cur_endian!=aendian) {
    if (aendian==1) {
      d_print (_ (", little-endian"));
    }
    else {
      d_print (_ (", big-endian"));
    }
  }
  d_print ("\n");
}


void 
on_resaudw_asamps_changed (GtkWidget irrelevant, gpointer rubbish) {
  if (atoi (gtk_entry_get_text (GTK_ENTRY (resaudw->entry_asamps)))==8) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (resaudw->rb_unsigned), TRUE);
    gtk_widget_set_sensitive (resaudw->rb_signed,FALSE);
    gtk_widget_set_sensitive (resaudw->rb_unsigned,FALSE);
    gtk_widget_set_sensitive (resaudw->rb_bigend,FALSE);
    gtk_widget_set_sensitive (resaudw->rb_littleend,FALSE);
  }
  else {
    if (cfile->signed_endian&AFORM_UNSIGNED) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (resaudw->rb_unsigned), TRUE);
    }
    else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (resaudw->rb_signed), TRUE);
    }
    gtk_widget_set_sensitive (resaudw->rb_signed,TRUE);
    gtk_widget_set_sensitive (resaudw->rb_unsigned,TRUE);
    gtk_widget_set_sensitive (resaudw->rb_bigend,TRUE);
    gtk_widget_set_sensitive (resaudw->rb_littleend,TRUE);
  }
}



void
on_resample_video_activate (GtkMenuItem     *menuitem,
			    gpointer         user_data)
{
  // change speed from the menu
  GtkWidget *resample_vid;

  if (!(cfile->events[EVENT_FRAMES]==NULL)) {
    do_error_dialog (_ ("\n\nLiVES cannot resample with frame changes recorded.\nPlease Reorder the frames with Tools|Reorder Frames,\n or clear the performance by pressing 'r'.\n"));
    return;
  }

  if (!(cfile->events[EVENT_FPS]==NULL)) {
    if (do_warning_dialog (_ ("\n\nWarning - resampling with variable frame rates.\nThis feature is not well tested.\n\nDo you wish to continue ?\n"))==FALSE) {
      return;
    }
  }

  resample_vid=create_new_pb_speed(2);
  mainw->fx1_val=cfile->fps;
  gtk_widget_show(resample_vid);
}



void
on_resample_vid_ok                  (GtkButton       *button,
				     GtkEntry         *entry)
{
  int i;
  gint old_frames;
  gint ostart=cfile->start;
  gint oend=cfile->end;
  gdouble oundo1_dbl=cfile->undo1_dbl;
  gchar *msg;
  gdouble time=0.;
  gint value=1,length=0;
  event *cur_event,*start_event;
  gdouble delta_time;
  gchar *test_fps_string1;
  gchar *test_fps_string2;

  mainw->error=FALSE;

  // we need this because the size of our fps events can now vary from frames
  cfile->ev_fps_start=1;
  cfile->ev_fps_end=cfile->perf_end-cfile->perf_start+1;

  if (!(button==NULL)) {
    gtk_widget_destroy(gtk_widget_get_toplevel(GTK_WIDGET(button)));
    if (mainw->fx1_val==0.) mainw->fx1_val=1.;
  }
  else {
    mainw->fx1_val=cfile->undo1_dbl;
  }

  test_fps_string1=g_strdup_printf ("%.3f00000",mainw->fx1_val);
  test_fps_string2=g_strdup_printf ("%.8f",mainw->fx1_val);

  if (strcmp (test_fps_string1,test_fps_string2)) {
    // got a ratio
    cfile->ratio_fps=TRUE;
  }
  else {
    cfile->ratio_fps=FALSE;
  }
  g_free (test_fps_string1);
  g_free (test_fps_string2);

  if (mainw->fx1_val==cfile->fps&&cfile->events[EVENT_FPS]==NULL) return;
  delta_time=1./mainw->fx1_val;

  // TODO - do_threaded_dialog

  if (cfile->events[EVENT_FPS]!=NULL) {
    for (i=0;value>0;length=i++) {
      value=calc_variable_frame (time,CALC_LAST);
      time+=delta_time;
    }
  }
  else {
    length=(gint) (cfile->frames/cfile->fps*mainw->fx1_val+.49999);
  }
  if (length<1) length=1;
  time=0.;
  if (length>1) {
    time=.00001;
  }

  if (create_event_space(length,CREATE_EV_FRAMES)==FALSE) {
    do_memory_error_dialog();
    if (cfile->events[EVENT_FPS]==NULL) {
      cfile->perf_counter=cfile->perf_start=cfile->perf_end=0;
    }
    mainw->error=TRUE;
    return;
  }

  old_frames=cfile->frames;
  cfile->undo1_dbl=cfile->fps;

  save_event_frames();
  save_event_fps();

  start_event=cur_event=cfile->events[EVENT_FRAMES];
  value=1;
  
  // stop calc_variable_frame from thinking there is a frame array
  cfile->events[EVENT_FRAMES]=NULL;

  // work out which frame should be at each tick
  for (i=0;value>0&&i<length;i++) {
    if ((value=calc_variable_frame (time,CALC_LAST))>0) {
      cur_event->value=value;
      cur_event->reltime=time*1000000.+cfile->ev_origticks;
      time+=delta_time;
      cur_event++;
    }
  }

  cfile->events[EVENT_FRAMES]=start_event;

  if (cfile->events[EVENT_FPS]==NULL) {
    if (create_event_space (1,CREATE_EV_FPS)==FALSE) {
      do_memory_error_dialog();
      load_event_frames();
      load_event_fps();
      if (cfile->events[EVENT_FRAMES]==NULL&&cfile->events[EVENT_FPS]==NULL) {
	cfile->perf_counter=cfile->perf_start=cfile->perf_end=0;
      }
      cfile->undo1_dbl=oundo1_dbl;
      mainw->error=TRUE;
      return;
    }
  }

  cfile->perf_start=1;
  cfile->perf_end=i;
  mainw->origticks=cfile->ev_origticks;
  cfile->pb_fps=cfile->fps=mainw->fx1_val;

  quantise_events(-1);

  if (cfile->events[EVENT_FPS]!=NULL) {
    g_free (cfile->events[EVENT_FPS]);
    cfile->events[EVENT_FPS]=NULL;
  }

  // TODO - end_threaded_dialog

  if (cfile->events[EVENT_FRAMES]==NULL) {
    do_memory_error_dialog();
    load_event_frames();
    load_event_fps();
    if (cfile->events[EVENT_FRAMES]==NULL&&cfile->events[EVENT_FPS]==NULL) {
      cfile->perf_counter=cfile->perf_start=cfile->perf_end=0;
    }
    cfile->undo1_dbl=oundo1_dbl;
    mainw->error=TRUE;
    return;
  }


  if (cfile->ratio_fps) {
    // got a ratio
    msg=g_strdup_printf(_ ("Resampling video at %.8f frames per second..."),cfile->fps);
  }
  else {
    msg=g_strdup_printf(_ ("Resampling video at %.3f frames per second..."),cfile->fps);
  }

  if (mainw->current_file>0) {
    d_print(msg);
  }
  g_free(msg);

  cfile->frames=0;
  cfile->start=(gint)((cfile->start-1)/cfile->undo1_dbl*cfile->fps+1.);
  if ((cfile->end=(gint)(cfile->end/cfile->undo1_dbl*cfile->fps+.49999))<cfile->start) cfile->end=cfile->start;

  cfile->undo_action=UNDO_RESAMPLE;
  on_reorder_activate(NULL,NULL);
  cfile->undo_action=UNDO_RESAMPLE;

  if (cfile->frames<=0) {
    // reordering error...
    if (cfile->events[EVENT_FRAMES]!=NULL) {
      g_free (cfile->events[EVENT_FRAMES]);
      cfile->events[EVENT_FRAMES]=NULL;
    }
    cfile->frames=-cfile->frames;
    load_event_frames();
    load_event_fps();
    if (cfile->events[EVENT_FRAMES]==NULL&&cfile->events[EVENT_FPS]==NULL) {
      cfile->perf_counter=cfile->perf_start=cfile->perf_end=0;
    }
    cfile->start=ostart;
    cfile->end=oend;
    cfile->pb_fps=cfile->fps=cfile->undo1_dbl;
    cfile->undo1_dbl=oundo1_dbl;
    sensitize();
    mainw->error=TRUE;
    do_error_dialog ("Reordering error !\n");
    return;
  }

  set_undoable ("Resample",TRUE);
  get_play_times();
  cfile->old_frames=old_frames;
}



///////// GUI stuff /////////////////////////////////////////////////////



_resaudw *
create_resaudw (gshort type)
{
  // type 1 == resample
  // type 2 == insert silence

  GtkWidget *dialog_vbox15;
  GtkWidget *vbox21;
  GtkWidget *frame6;
  GtkWidget *hbox24;
  GtkWidget *label94;
  GtkWidget *combo_entry2;
  GtkWidget *label95;
  GtkWidget *combo_entry3;
  GtkWidget *label96;
  GtkWidget *combo_entry1;
  GtkWidget *vseparator3;
  GtkWidget *vbox22;
  GtkWidget *radiobutton26;
  GSList *radiobutton26_group = NULL;
  GtkWidget *radiobutton27;
  GtkWidget *vseparator4;
  GtkWidget *vbox23;
  GtkWidget *radiobutton28;
  GSList *radiobutton28_group = NULL;
  GtkWidget *radiobutton29;
  GtkWidget *label92;
  GtkWidget *frame7;
  GtkWidget *hbox25;
  GtkWidget *label98;
  GtkWidget *combo4;
  GtkWidget *label99;
  GtkWidget *combo5;
  GtkWidget *label100;
  GtkWidget *combo6;
  GtkWidget *vseparator5;
  GtkWidget *vbox24;
  GSList *radiobutton30_group = NULL;
  GtkWidget *vseparator6;
  GtkWidget *vbox25;
  GSList *radiobutton32_group = NULL;
  GtkWidget *label93;
  GtkWidget *dialog_action_area15;
  GtkWidget *cancelbutton13;
  GtkWidget *okbutton12;
  GList *channels = NULL;
  GList *sampsize = NULL;
  GList *rate = NULL;
  gchar *tmp;

  _resaudw *resaudw=(_resaudw*)(g_malloc(sizeof(_resaudw)));

  channels = g_list_append (channels, "1");
  channels = g_list_append (channels, "2");

  sampsize = g_list_append (sampsize, "8");
  sampsize = g_list_append (sampsize, "16");

  rate = g_list_append (rate, "5512");
  rate = g_list_append (rate, "8000");
  rate = g_list_append (rate, "11025");
  rate = g_list_append (rate, "22050");
  rate = g_list_append (rate, "32000");
  rate = g_list_append (rate, "44100");
  rate = g_list_append (rate, "48000");
  rate = g_list_append (rate, "88200");
  rate = g_list_append (rate, "96000");
  rate = g_list_append (rate, "128000");


  resaudw->dialog = gtk_dialog_new ();
  gtk_container_set_border_width (GTK_CONTAINER (resaudw->dialog), 10);
  if (type==1) {
    gtk_window_set_title (GTK_WINDOW (resaudw->dialog), _("LiVES: - Resample Audio"));
  }
  if (type==2) {
    gtk_window_set_title (GTK_WINDOW (resaudw->dialog), _("LiVES: - Insert Silence"));
  }
  gtk_window_set_position (GTK_WINDOW (resaudw->dialog), GTK_WIN_POS_CENTER);
  gtk_window_set_modal (GTK_WINDOW (resaudw->dialog), TRUE);
  gtk_window_set_transient_for(GTK_WINDOW(resaudw->dialog),GTK_WINDOW(mainw->LiVES));

  dialog_vbox15 = GTK_DIALOG (resaudw->dialog)->vbox;
  gtk_widget_show (dialog_vbox15);
  if (palette->style&STYLE_3||palette->style==STYLE_PLAIN) {
    gtk_widget_modify_bg (resaudw->dialog, GTK_STATE_NORMAL, &palette->normal_back);
  }
  else {
    gtk_widget_modify_bg (resaudw->dialog, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }


  vbox21 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox21);
  gtk_box_pack_start (GTK_BOX (dialog_vbox15), vbox21, TRUE, TRUE, 0);

  frame6 = gtk_frame_new (NULL);
 
  if (type==1) {

    gtk_widget_show (frame6);
    gtk_box_pack_start (GTK_BOX (vbox21), frame6, TRUE, TRUE, 0);
    
    if (palette->style&STYLE_3||palette->style==STYLE_PLAIN) {
      gtk_widget_modify_bg (frame6, GTK_STATE_NORMAL, &palette->normal_back);
    }
    else {
      gtk_widget_modify_bg (frame6, GTK_STATE_NORMAL, &palette->menu_and_bars);
    }
    
    hbox24 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox24);
    gtk_container_add (GTK_CONTAINER (frame6), hbox24);
    gtk_container_set_border_width (GTK_CONTAINER (hbox24), 10);
    
    label94 = gtk_label_new (_("Rate (Hz) "));
    gtk_widget_show (label94);
    gtk_box_pack_start (GTK_BOX (hbox24), label94, FALSE, FALSE, 0);
    gtk_label_set_justify (GTK_LABEL (label94), GTK_JUSTIFY_LEFT);
    
    combo_entry2 = gtk_entry_new();
    gtk_widget_show (combo_entry2);
    gtk_box_pack_start (GTK_BOX (hbox24), combo_entry2, TRUE, TRUE, 1);

    gtk_editable_set_editable (GTK_EDITABLE (combo_entry2), FALSE);
    gtk_entry_set_width_chars (GTK_ENTRY (combo_entry2), 8);
    
    tmp=g_strdup_printf ("%d",(gint)mainw->fx1_val);
    gtk_entry_set_text (GTK_ENTRY (combo_entry2),tmp);
    g_free (tmp);
    GTK_WIDGET_UNSET_FLAGS (combo_entry2, GTK_CAN_FOCUS);
    
    label95 = gtk_label_new (_("         Channels "));
    gtk_widget_show (label95);
    gtk_box_pack_start (GTK_BOX (hbox24), label95, FALSE, FALSE, 0);
    gtk_label_set_justify (GTK_LABEL (label95), GTK_JUSTIFY_LEFT);
    
    combo_entry3 = gtk_entry_new();
    gtk_widget_show (combo_entry3);
    gtk_box_pack_start (GTK_BOX (hbox24), combo_entry3, TRUE, TRUE, 1);
    
    gtk_editable_set_editable (GTK_EDITABLE (combo_entry3), FALSE);
    gtk_entry_set_width_chars (GTK_ENTRY (combo_entry3), 3);
    
    tmp=g_strdup_printf ("%d",(gint)mainw->fx2_val);
    gtk_entry_set_text (GTK_ENTRY (combo_entry3),tmp);
    g_free (tmp);
    GTK_WIDGET_UNSET_FLAGS (combo_entry3, GTK_CAN_FOCUS);
    
    label96 = gtk_label_new (_("        Sample Size "));
    gtk_widget_show (label96);
    gtk_box_pack_start (GTK_BOX (hbox24), label96, FALSE, FALSE, 0);
    gtk_label_set_justify (GTK_LABEL (label96), GTK_JUSTIFY_LEFT);

    combo_entry1 = gtk_entry_new();
    gtk_widget_show (combo_entry1);
    gtk_box_pack_start (GTK_BOX (hbox24), combo_entry1, TRUE, TRUE, 1);

    gtk_editable_set_editable (GTK_EDITABLE (combo_entry1), FALSE);
    gtk_entry_set_width_chars (GTK_ENTRY (combo_entry1), 3);

    tmp=g_strdup_printf ("%d",(gint)mainw->fx3_val);
    gtk_entry_set_text (GTK_ENTRY (combo_entry1),tmp);
    g_free (tmp);
    GTK_WIDGET_UNSET_FLAGS (combo_entry1, GTK_CAN_FOCUS);

    vseparator3 = gtk_vseparator_new ();
    gtk_widget_show (vseparator3);
    gtk_box_pack_start (GTK_BOX (hbox24), vseparator3, FALSE, FALSE, 12);

    vbox22 = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox22);
    gtk_box_pack_start (GTK_BOX (hbox24), vbox22, TRUE, TRUE, 0);

    radiobutton26 = gtk_radio_button_new_with_mnemonic (NULL, _("Signed"));
    gtk_widget_show (radiobutton26);
    gtk_box_pack_start (GTK_BOX (vbox22), radiobutton26, TRUE, FALSE, 0);
    gtk_radio_button_set_group (GTK_RADIO_BUTTON (radiobutton26), radiobutton26_group);
    radiobutton26_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton26));
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radiobutton26), TRUE);

    radiobutton27 = gtk_radio_button_new_with_mnemonic (NULL, _("Unsigned"));
    gtk_widget_show (radiobutton27);
    gtk_box_pack_start (GTK_BOX (vbox22), radiobutton27, TRUE, FALSE, 0);
    gtk_radio_button_set_group (GTK_RADIO_BUTTON (radiobutton27), radiobutton26_group);
    radiobutton26_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton27));

    if ((gint)mainw->fx4_val&AFORM_UNSIGNED) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radiobutton27), TRUE);
    }
    else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radiobutton26), TRUE);
    }

    gtk_widget_set_sensitive (radiobutton26, FALSE);
    gtk_widget_set_sensitive (radiobutton27, FALSE);

    vseparator4 = gtk_vseparator_new ();
    gtk_widget_show (vseparator4);
    gtk_box_pack_start (GTK_BOX (hbox24), vseparator4, FALSE, FALSE, 12);
    
    vbox23 = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox23);
    gtk_box_pack_start (GTK_BOX (hbox24), vbox23, TRUE, TRUE, 0);

    radiobutton28 = gtk_radio_button_new_with_mnemonic (NULL, _("Little Endian"));
    gtk_widget_show (radiobutton28);
    gtk_box_pack_start (GTK_BOX (vbox23), radiobutton28, TRUE, FALSE, 0);
    gtk_radio_button_set_group (GTK_RADIO_BUTTON (radiobutton28), radiobutton28_group);
    radiobutton28_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton28));
    
    radiobutton29 = gtk_radio_button_new_with_mnemonic (NULL, _("Big Endian"));
    gtk_widget_show (radiobutton29);
    gtk_box_pack_start (GTK_BOX (vbox23), radiobutton29, TRUE, FALSE, 0);
    gtk_radio_button_set_group (GTK_RADIO_BUTTON (radiobutton29), radiobutton28_group);
    radiobutton28_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (radiobutton29));
    
    if ((gint)mainw->fx4_val&AFORM_BIG_ENDIAN) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radiobutton29), TRUE);
    }
    else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radiobutton28), TRUE);
    }

    gtk_widget_set_sensitive (radiobutton29, FALSE);
    gtk_widget_set_sensitive (radiobutton28, FALSE);

    label92 = gtk_label_new (_("Current"));
    
    gtk_widget_show (label92);
    gtk_frame_set_label_widget (GTK_FRAME (frame6), label92);
    gtk_label_set_justify (GTK_LABEL (label92), GTK_JUSTIFY_LEFT);
    
  }

  frame7 = gtk_frame_new (NULL);
  gtk_widget_show (frame7);
  gtk_box_pack_start (GTK_BOX (vbox21), frame7, TRUE, TRUE, 0);

  if (palette->style&STYLE_3||palette->style==STYLE_PLAIN) {
    gtk_widget_modify_bg (frame7, GTK_STATE_NORMAL, &palette->normal_back);
  }
  else {
    gtk_widget_modify_bg (frame7, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }

  hbox25 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox25);
  gtk_container_add (GTK_CONTAINER (frame7), hbox25);
  gtk_container_set_border_width (GTK_CONTAINER (hbox25), 10);

  label98 = gtk_label_new (_("Rate (Hz) "));
  gtk_widget_show (label98);
  gtk_box_pack_start (GTK_BOX (hbox25), label98, FALSE, FALSE, 0);
  gtk_label_set_justify (GTK_LABEL (label98), GTK_JUSTIFY_LEFT);

  combo4 = gtk_combo_new ();
  gtk_combo_set_popdown_strings (GTK_COMBO (combo4), rate);
  gtk_widget_show (combo4);
  gtk_box_pack_start (GTK_BOX (hbox25), combo4, TRUE, TRUE, 1);

  resaudw->entry_arate = GTK_COMBO (combo4)->entry;
  gtk_widget_show (resaudw->entry_arate);
  gtk_entry_set_width_chars (GTK_ENTRY (resaudw->entry_arate), 8);

  tmp=g_strdup_printf ("%d",(gint)mainw->fx1_val);
  gtk_entry_set_text (GTK_ENTRY (resaudw->entry_arate),tmp);
  g_free (tmp);

  label99 = gtk_label_new (_("    Channels "));
  gtk_widget_show (label99);
  gtk_box_pack_start (GTK_BOX (hbox25), label99, FALSE, FALSE, 0);
  gtk_label_set_justify (GTK_LABEL (label99), GTK_JUSTIFY_LEFT);

  combo5 = gtk_combo_new ();
  gtk_combo_set_popdown_strings (GTK_COMBO (combo5), channels);
  gtk_widget_show (combo5);
  gtk_box_pack_start (GTK_BOX (hbox25), combo5, FALSE, FALSE, 0);

  resaudw->entry_achans = GTK_COMBO (combo5)->entry;
  gtk_widget_show (resaudw->entry_achans);
  gtk_editable_set_editable (GTK_EDITABLE (resaudw->entry_achans), FALSE);
  gtk_entry_set_width_chars (GTK_ENTRY (resaudw->entry_achans), 3);

  tmp=g_strdup_printf ("%d",(gint)mainw->fx2_val);
  gtk_entry_set_text (GTK_ENTRY (resaudw->entry_achans),tmp);
  g_free (tmp);

  label100 = gtk_label_new (_("    Sample Size "));
  gtk_widget_show (label100);
  gtk_box_pack_start (GTK_BOX (hbox25), label100, FALSE, FALSE, 0);
  gtk_label_set_justify (GTK_LABEL (label100), GTK_JUSTIFY_LEFT);

  combo6 = gtk_combo_new ();
  gtk_combo_set_popdown_strings (GTK_COMBO (combo6), sampsize);
  gtk_widget_show (combo6);
  gtk_box_pack_start (GTK_BOX (hbox25), combo6, TRUE, TRUE, 0);

  resaudw->entry_asamps = GTK_COMBO (combo6)->entry;
  gtk_widget_show (resaudw->entry_asamps);
  gtk_entry_set_max_length (GTK_ENTRY (resaudw->entry_asamps), 2);
  gtk_editable_set_editable (GTK_EDITABLE (resaudw->entry_asamps), FALSE);
  gtk_entry_set_width_chars (GTK_ENTRY (resaudw->entry_asamps), 3);

  tmp=g_strdup_printf ("%d",(gint)mainw->fx3_val);
  gtk_entry_set_text (GTK_ENTRY (resaudw->entry_asamps),tmp);
  g_free (tmp);

  vseparator5 = gtk_vseparator_new ();
  gtk_widget_show (vseparator5);
  gtk_box_pack_start (GTK_BOX (hbox25), vseparator5, FALSE, FALSE, 12);

  vbox24 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox24);
  gtk_box_pack_start (GTK_BOX (hbox25), vbox24, TRUE, TRUE, 0);

  resaudw->rb_signed = gtk_radio_button_new_with_mnemonic (NULL, _("Signed"));
  gtk_widget_show (resaudw->rb_signed);
  gtk_box_pack_start (GTK_BOX (vbox24), resaudw->rb_signed, TRUE, FALSE, 0);
  gtk_radio_button_set_group (GTK_RADIO_BUTTON (resaudw->rb_signed), radiobutton30_group);
  radiobutton30_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (resaudw->rb_signed));
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (resaudw->rb_signed), TRUE);

  resaudw->rb_unsigned = gtk_radio_button_new_with_mnemonic (NULL, _("Unsigned"));
  gtk_widget_show (resaudw->rb_unsigned);
  gtk_box_pack_start (GTK_BOX (vbox24), resaudw->rb_unsigned, TRUE, FALSE, 0);
  gtk_radio_button_set_group (GTK_RADIO_BUTTON (resaudw->rb_unsigned), radiobutton30_group);
  radiobutton30_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (resaudw->rb_unsigned));

  if ((gint)mainw->fx4_val&AFORM_UNSIGNED) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (resaudw->rb_unsigned), TRUE);
  }
  else {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (resaudw->rb_signed), TRUE);
  }

  if (cfile->asampsize==8) {
    gtk_widget_set_sensitive (resaudw->rb_signed, FALSE);
    gtk_widget_set_sensitive (resaudw->rb_unsigned, FALSE);
  }

  vseparator6 = gtk_vseparator_new ();
  gtk_widget_show (vseparator6);
  gtk_box_pack_start (GTK_BOX (hbox25), vseparator6, FALSE, FALSE, 12);

  vbox25 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox25);
  gtk_box_pack_start (GTK_BOX (hbox25), vbox25, TRUE, TRUE, 0);

  resaudw->rb_littleend = gtk_radio_button_new_with_mnemonic (NULL, _("Little Endian"));
  gtk_widget_show (resaudw->rb_littleend);
  gtk_box_pack_start (GTK_BOX (vbox25), resaudw->rb_littleend, TRUE, FALSE, 0);
  gtk_radio_button_set_group (GTK_RADIO_BUTTON (resaudw->rb_littleend), radiobutton32_group);
  radiobutton32_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (resaudw->rb_littleend));
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (resaudw->rb_littleend), TRUE);

  resaudw->rb_bigend = gtk_radio_button_new_with_mnemonic (NULL, _("Big Endian"));
  gtk_widget_show (resaudw->rb_bigend);
  gtk_box_pack_start (GTK_BOX (vbox25), resaudw->rb_bigend, TRUE, FALSE, 0);
  gtk_radio_button_set_group (GTK_RADIO_BUTTON (resaudw->rb_bigend), radiobutton32_group);
  radiobutton32_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (resaudw->rb_bigend));

  if ((gint)mainw->fx4_val&AFORM_BIG_ENDIAN) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (resaudw->rb_bigend), TRUE);
  }
  else {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (resaudw->rb_littleend), TRUE);
  }

  if (cfile->asampsize==8) {
    gtk_widget_set_sensitive (resaudw->rb_littleend, FALSE);
    gtk_widget_set_sensitive (resaudw->rb_bigend, FALSE);
    }

  label93 = gtk_label_new (_("New"));
  gtk_widget_show (label93);
  gtk_frame_set_label_widget (GTK_FRAME (frame7), label93);
  gtk_label_set_justify (GTK_LABEL (label93), GTK_JUSTIFY_LEFT);

  dialog_action_area15 = GTK_DIALOG (resaudw->dialog)->action_area;
  gtk_widget_show (dialog_action_area15);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area15), GTK_BUTTONBOX_END);

  cancelbutton13 = gtk_button_new_from_stock ("gtk-cancel");
  gtk_widget_show (cancelbutton13);
  gtk_dialog_add_action_widget (GTK_DIALOG (resaudw->dialog), cancelbutton13, GTK_RESPONSE_CANCEL);
  GTK_WIDGET_SET_FLAGS (cancelbutton13, GTK_CAN_DEFAULT);

  okbutton12 = gtk_button_new_from_stock ("gtk-ok");
  gtk_widget_show (okbutton12);
  gtk_dialog_add_action_widget (GTK_DIALOG (resaudw->dialog), okbutton12, GTK_RESPONSE_OK);
  GTK_WIDGET_SET_FLAGS (okbutton12, GTK_CAN_DEFAULT);
  gtk_widget_grab_default (okbutton12);

  g_signal_connect (GTK_OBJECT(resaudw->entry_asamps), "changed",
		    G_CALLBACK (on_resaudw_asamps_changed),
		    NULL);

  g_signal_connect (GTK_OBJECT (cancelbutton13), "clicked",
		    G_CALLBACK (on_cancel_button1_clicked),
		    resaudw);

  if (type==1) {
    g_signal_connect (GTK_OBJECT (okbutton12), "clicked",
		      G_CALLBACK (on_resaudio_ok_clicked),
		      NULL);
  }
  if (type==2) {
    g_signal_connect (GTK_OBJECT (okbutton12), "clicked",
		      G_CALLBACK (on_ins_silence_details_clicked),
		      NULL);
  }

  g_signal_connect (GTK_OBJECT (resaudw->dialog), "delete_event",
                      G_CALLBACK (return_true),
                      NULL);

  g_list_free (channels);
  g_list_free (sampsize);
  g_list_free (rate);

  return resaudw;
}



GtkWidget*
create_new_pb_speed (gshort type)
{
  GtkWidget *new_pb_speed;
  GtkWidget *dialog_vbox6;
  GtkWidget *vbox6;
  GtkWidget *pb_speed_text;
  GtkObject *spinbutton_pb_speed_adj;
  GtkWidget *spinbutton_pb_speed;
  GtkWidget *dialog_action_area6;
  GtkWidget *cancelbutton4;
  GtkWidget *change_pb_ok;
  GtkWidget *change_audio_speed;
  gchar label_text[256];

  new_pb_speed = gtk_dialog_new ();
  gtk_container_set_border_width (GTK_CONTAINER (new_pb_speed), 10);
  if (type==1) {
    gtk_window_set_title (GTK_WINDOW (new_pb_speed), _("LiVES: - Change playback speed"));
  }
  else if (type==2) {
    gtk_window_set_title (GTK_WINDOW (new_pb_speed), _("LiVES: - Resample Video"));
  }
  gtk_window_set_position (GTK_WINDOW (new_pb_speed), GTK_WIN_POS_CENTER);
  gtk_window_set_modal (GTK_WINDOW (new_pb_speed), TRUE);
  gtk_window_set_default_size (GTK_WINDOW (new_pb_speed), 300, 200);
  gtk_window_set_transient_for(GTK_WINDOW(new_pb_speed),GTK_WINDOW(mainw->LiVES));

  gtk_widget_modify_bg(new_pb_speed, GTK_STATE_NORMAL, &palette->normal_back);
  if (palette->style&STYLE_1) {
    gtk_dialog_set_has_separator(GTK_DIALOG(new_pb_speed),FALSE);
  }

  dialog_vbox6 = GTK_DIALOG (new_pb_speed)->vbox;
  gtk_widget_show (dialog_vbox6);

  vbox6 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox6);
  gtk_box_pack_start (GTK_BOX (dialog_vbox6), vbox6, TRUE, TRUE, 0);

  if (type==1) {
    g_snprintf(label_text,256,_("\n\nCurrent playback speed is %.3f frames per second.\n\nPlease enter the desired playback speed\nin frames per second"),cfile->fps);
  }
  else if (type==2) {
    g_snprintf(label_text,256,_("\n\nCurrent playback speed is %.3f frames per second.\n\nPlease enter the resampled rate\nin frames per second"),cfile->fps);
  }
  pb_speed_text = gtk_label_new (label_text);
  gtk_widget_show (pb_speed_text);
  gtk_box_pack_start (GTK_BOX (vbox6), pb_speed_text, TRUE, TRUE, 0);
  gtk_widget_modify_fg(pb_speed_text, GTK_STATE_NORMAL, &palette->normal_fore);

  spinbutton_pb_speed_adj = gtk_adjustment_new (cfile->fps, 1, FPS_MAX, 0.01, .1, .5);
  spinbutton_pb_speed = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton_pb_speed_adj), 1, 3);
  gtk_widget_show (spinbutton_pb_speed);
  gtk_box_pack_start (GTK_BOX (vbox6), spinbutton_pb_speed, FALSE, FALSE, 8);
  gtk_widget_set_size_request (spinbutton_pb_speed, 10, -1);
  gtk_entry_set_activates_default (GTK_ENTRY ((GtkEntry *)&(GTK_SPIN_BUTTON (spinbutton_pb_speed)->entry)), TRUE);

  change_audio_speed = gtk_check_button_new_with_mnemonic (_("Change the audio speed as well"));
  if (type==1&&cfile->achans>0) {
    gtk_widget_show (change_audio_speed);
  }
  gtk_box_pack_start (GTK_BOX (vbox6), change_audio_speed, FALSE, FALSE, 0);

  dialog_action_area6 = GTK_DIALOG (new_pb_speed)->action_area;
  gtk_widget_show (dialog_action_area6);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area6), GTK_BUTTONBOX_END);

  cancelbutton4 = gtk_button_new_from_stock ("gtk-cancel");
  gtk_widget_show (cancelbutton4);
  gtk_dialog_add_action_widget (GTK_DIALOG (new_pb_speed), cancelbutton4, GTK_RESPONSE_CANCEL);
  GTK_WIDGET_SET_FLAGS (cancelbutton4, GTK_CAN_DEFAULT);

  change_pb_ok = gtk_button_new_from_stock ("gtk-ok");
  gtk_widget_show (change_pb_ok);
  gtk_dialog_add_action_widget (GTK_DIALOG (new_pb_speed), change_pb_ok, GTK_RESPONSE_OK);
  GTK_WIDGET_SET_FLAGS (change_pb_ok, GTK_CAN_DEFAULT);
  gtk_widget_grab_default (change_pb_ok);

  g_signal_connect (GTK_OBJECT (change_audio_speed), "toggled",
		    G_CALLBACK (on_boolean_toggled),
		    GINT_TO_POINTER (1));
  g_signal_connect (GTK_OBJECT (cancelbutton4), "clicked",
		    G_CALLBACK (on_cancel_button1_clicked),
		    NULL);
  if (type==1) {
    g_signal_connect (GTK_OBJECT (change_pb_ok), "clicked",
		      G_CALLBACK (on_change_speed_ok_clicked),
		      NULL);
  }
  else if (type==2) {
    g_signal_connect (GTK_OBJECT (change_pb_ok), "clicked",
		      G_CALLBACK (on_resample_vid_ok),
		      NULL);
  }
  g_signal_connect_after (GTK_OBJECT (spinbutton_pb_speed), "value_changed",
			  G_CALLBACK (on_spin_value_changed),
			  GINT_TO_POINTER (1));


  return new_pb_speed;
}



void
on_change_speed_activate                (GtkMenuItem     *menuitem,
					 gpointer         user_data)
{
  // change speed from the menu
  GtkWidget *new_pb_speed=create_new_pb_speed(1);
  mainw->fx1_bool=FALSE;
  mainw->fx1_val=cfile->fps;
  gtk_widget_show(new_pb_speed);
}



void
on_change_speed_ok_clicked                (GtkButton *button,
					   gpointer         user_data)
{
  gdouble arate=cfile->arate/cfile->fps;
  gchar *msg;

  // change playback rate
  if (!(button==NULL)) {
    gtk_widget_destroy(gtk_widget_get_toplevel(GTK_WIDGET(button)));
  }
  else {
    mainw->fx1_bool=!(cfile->undo1_int==cfile->arate);
    mainw->fx1_val=cfile->undo1_dbl;
  }

  set_undoable ("Speed Change",TRUE);
  cfile->undo1_dbl=cfile->fps;
  cfile->undo1_int=cfile->arate;
  cfile->undo_action=UNDO_CHANGE_SPEED;

  if (mainw->fx1_val==0.) mainw->fx1_val=1.;
  cfile->pb_fps=cfile->fps=mainw->fx1_val;
  if (mainw->fx1_bool==TRUE) {
    cfile->arate=(gint)(arate*cfile->fps+.5);
    msg=g_strdup_printf (_ ("Changed playback speed to %.3f frames per second and audio to %d Hz.\n"),cfile->fps,cfile->arate);
  }
  else {
    msg=g_strdup_printf (_ ("Changed playback speed to %.3f frames per second.\n"),cfile->fps);
  }
  d_print (msg);
  g_free (msg);

  // resize the timer bars, etc
  set_sel_label();
  get_play_times();
}
