// saveplay.c
// LiVES (lives-exe)
// (c) G. Finch 2003
// Released under the GPL 2.0
// see file ../COPYING for licensing details.

#include <unistd.h>
#include <stdlib.h>

#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "main.h"
#include "callbacks.h"
#include "support.h"


#ifdef USE_LIVIDO
extern void livido_init_all(void);
extern void livido_generator_end (livido_instance_t *src);
extern void livido_bg_generator_end (livido_instance_t *src);
extern gboolean livido_playback_gen_start(void);
#endif

void
read_file_details(const gchar *file_name) {
  // get preliminary details

  FILE *infofile;
  gchar *com=g_strdup_printf("smogrify get_details %s \"%s\"",cfile->handle,file_name);
  system(com);
  g_free(com);

  clear_mainw_msg();

  while (!(infofile=fopen(cfile->info_file,"r"))) {
    g_usleep(prefs->sleep_time);
  }

  fgets(mainw->msg,512,infofile);
  fclose(infofile);
}


void 
deduce_file(gchar *file_name, gdouble start, gint end) {
  // this is a utility function to deduce whether we are dealing with a file, 
  // a selection, a backup, or a location

  if (g_strrstr(file_name,"://")!=NULL&&strncmp (file_name,"dvd://",6)) {
    mainw->opening_loc=TRUE;
    open_file(file_name);
    mainw->opening_loc=FALSE;
  }
  else {
    if (!(strcmp(file_name+strlen(file_name)-4,".lv1"))) {
      restore_file(file_name);
    }
    else {
      open_file_sel(file_name,start,end);
    }
  }
}


void open_file (const gchar *file_name) {
  // this function should be called to open a whole file
  open_file_sel(file_name,0.,0);
}

void open_file_sel(const gchar *file_name, gdouble start, gint frames) {
  gchar *com=g_strdup("dummy");
  gint withsound=1;
  gint old_file=mainw->current_file;
  gint new_file=old_file;
  gchar *fname=g_strdup(file_name);
  gint achans,arate,asampsize;
  gint current_file;
  gchar msg[256],loc[256];

  if (old_file==-1||cfile->opening==FALSE) {
    new_file=mainw->first_free_file;

    if (get_new_handle(new_file,fname)==FALSE) {
      g_free(fname);
      return;
    }
    g_free(fname);
    
    if (frames==0) {
      com=g_strdup_printf(_ ("Opening %s"),file_name);
    }
    else {
      com=g_strdup_printf(_ ("Opening %s start time %.2f sec. frames %d"),file_name,start,frames);
    }
    d_print(com);
    g_free(com);
    
    if (mainw->save_with_sound==FALSE) {
      d_print(_ (" without sound"));
      withsound=0;
    }
    
    mainw->noswitch=TRUE;
    mainw->current_file=new_file;
    mainw->reinit_effects=TRUE;

    // get the file size, etc. (frames is just a guess here)
    read_file_details(file_name);
    unlink (cfile->info_file);
    system ("sync;sync;sync");

    // we must set this before calling add_file_info
    cfile->opening=TRUE;
    add_file_info (cfile->handle);
    if (frames>0&&cfile->frames>frames) cfile->end=cfile->undo_end=cfile->frames=frames;

    // be careful, here we switch from mainw->opening_loc to cfile->opening_loc
    if (mainw->opening_loc==TRUE) {
      cfile->opening_loc=TRUE;
      mainw->opening_loc=FALSE;
    }
    
    if (cfile->f_size>prefs->warn_file_size*1000000.&&(mainw->is_ready==TRUE)&&!frames) {
      gchar *warn=g_strdup_printf(_ ("\nLiVES is not currently optimised for larger file sizes.\nYou are advised (for now) to start with a smaller file, or to use the 'Open File Selection' option.\n(Filesize=%.2fMB)\n\nAre you sure you wish to continue ?"),cfile->f_size/1000000.);
      if (!do_warning_dialog_with_check(warn,WARN_MASK_FSIZE)) {
	g_free(warn);
	close_current_file();
	if (old_file>0&&mainw->current_file!=old_file) {
	  switch_to_file (mainw->current_file,old_file);
	}
	mainw->noswitch=FALSE;
	return;
      }
      g_free(warn);
      d_print(_ (" - please be patient."));
    }
    
    d_print("\n");
#if defined DEBUG
    g_print("open_file: dpd in\n");
#endif

    // set undo_start and undo_end for preview
    cfile->undo_start=1;
    cfile->undo_end=cfile->frames;

    if (cfile->achans>0) {
      cfile->opening_audio=TRUE;
    }

    // these will get reset as we have no audio file yet, so preserve them
    achans=cfile->achans;
    arate=cfile->arate;
    asampsize=cfile->asampsize;
    cfile->old_frames=cfile->frames;
    cfile->frames=0;

    // we need this FALSE here, otherwise we will switch straight back here...
    cfile->opening=FALSE;

    // force a resize
    current_file=mainw->current_file;

    if (mainw->playing_file>-1) {
      do_quick_switch (current_file);
    }
    else {
      switch_to_file((mainw->current_file=old_file),current_file);
    }

    cfile->opening=TRUE;
    cfile->achans=achans;
    cfile->arate=arate;
    cfile->asampsize=asampsize;
    cfile->frames=cfile->old_frames;

    if (cfile->frames<=0) {
      cfile->undo_end=cfile->frames=123456789;
    }
    if (cfile->hsize*cfile->vsize==0) {
      cfile->frames=0;
    }

    get_play_times();
    if (!(mainw->toy_type==TOY_TV)) {
      add_to_winmenu();
      set_main_title(cfile->file_name,0);
    }

    if (mainw->file_open_params==NULL) mainw->file_open_params=g_strdup("");

    com=g_strdup_printf("smogrify open %s \"%s\" %d %.2f %d \"%s\"",cfile->handle,g_filename_from_utf8 (file_name,-1,NULL,NULL,NULL),withsound,start,frames,mainw->file_open_params);
    unlink (cfile->info_file);
    system(com);
    g_free(com);
    mainw->noswitch=FALSE;

    if (mainw->toy_type==TOY_TV) {
      // for LiVES TV we do an auto-preview
      mainw->play_start=cfile->start=cfile->undo_start;
      mainw->play_end=cfile->end=cfile->undo_end;
      mainw->preview=TRUE;
      do {
	desensitize();
	procw_desensitize();
	on_playsel_activate (NULL, NULL);
      } while (mainw->cancelled==100);
      mainw->preview=FALSE;
      mainw->toy_type=TOY_NONE;
      return;
    }
  }
  // 'entry point' when we switch back

  // spin until loading is complete
  // afterwards, mainw->msg will contain file details

  // (also check for cancel)

  if (!do_progress_dialog(mainw->toy_type!=TOY_TV,TRUE,_ ("Opening"))) {
    // user switched to another clip...we will just return...maybe the user will come back here later...
    if (mainw->cancelled==3) {
      mainw->cancelled=0;
      return;
    }

    // cancelled
    // clean up our temp files
    com=g_strdup_printf("smogrify stop %s",cfile->handle);
    system(com);
    g_free(com);
    close_current_file();
    if (old_file>0&&mainw->current_file!=old_file) {
      switch_to_file (mainw->current_file,old_file);
    }
    return;
  }
  cfile->opening=cfile->opening_audio=FALSE;

#if defined DEBUG
  g_print("Out of dpd\n");
#endif

  // mainw->error is TRUE if we could not open the file
  if (mainw->error==TRUE) {
    do_blocking_error_dialog(mainw->msg);
    d_print(_ ("Failed\n"));
    close_current_file();
    if (old_file>0&&mainw->current_file!=old_file) {
      switch_to_file (mainw->current_file,old_file);
    }
    return;
  }

  if (cfile->opening_loc==TRUE) {
    cfile->changed=TRUE;
    cfile->opening_loc=FALSE;
  }

  // now file should be loaded...get full details
  cfile->is_loaded=TRUE;
  add_file_info(cfile->handle);
  if (cfile->frames<=0) {
    if (!cfile->afilesize) {
      // we got neither video nor audio...
      g_snprintf (msg,256,_ ("\n\nLiVES was unable to extract either video or audio.\n"));
      get_location ("mplayer",loc,256);
      if (strcmp (prefs->video_open_command,loc)) {
	g_strappend (msg,256,_ ("\n\nPlease check the setting of Video open command in\nTools|Preferences|Decoding\n"));
      }

      do_error_dialog(msg);
      d_print(_ ("Failed.\n"));
      close_current_file();
      if (old_file>0&&mainw->current_file!=old_file) {
	switch_to_file (mainw->current_file,old_file);
      }
      return;
    }
    cfile->frames=0;
  }

  // update widgets
  current_file=mainw->current_file;
  switch_to_file((mainw->current_file=0),current_file);

  if (prefs->show_recent==TRUE) {
    // TODO - add the file_open_params
    add_to_recent(file_name,start,frames);
  }
  g_free (mainw->file_open_params);
  mainw->file_open_params=NULL;

  if (!strcmp(cfile->type,"Frames")||!strcmp(cfile->type,"jpeg")||!strcmp(cfile->type,"Audio")) {
    cfile->is_untitled=TRUE;
    gtk_widget_set_sensitive(mainw->save,FALSE);
  }
}



void 
open_fw_device (void) {
  // TODO - prompt for max number of frames
  gchar *com;

  gint old_file=mainw->current_file;
  gint new_file=mainw->first_free_file;
  gint current_file;

  if (get_new_handle(new_file,"")==FALSE) {
    return;
  }
 mainw->current_file=new_file;
 
 d_print(_ ("Loading from firewire..."));
 
 if (prefs->is_pal) {
   g_snprintf(mainw->msg,512,"completed|%s|0|dv|720|576|8|25.00|0|44100|0|16|1|1|0",cfile->handle);
 }
 else {
   g_snprintf(mainw->msg,512,"completed|%s|0|dv|720|480|8|29.97|0|44100|0|16|1|1|0",cfile->handle);
 }

 add_file_info(cfile->handle);

 // set undo_start and undo_end for preview
 cfile->undo_start=1;
 cfile->undo_end=cfile->frames=123456789;

 current_file=mainw->current_file;
 switch_to_file ((mainw->current_file=old_file),current_file);
 cfile->opening=TRUE;
 mainw->dvgrab_preview=TRUE;

 get_play_times();
 add_to_winmenu();
 set_main_title(cfile->file_name,0);
 
 com=g_strdup_printf("smogrify dvgrab %s",cfile->handle);
 unlink (cfile->info_file);
 system(com);
 g_free(com);

  if (do_progress_dialog(TRUE,TRUE,_ ("Loading from firewire device"))==FALSE) {
    // cancelled
    // clean up our temp files
    com=g_strdup_printf("smogrify stop %s",cfile->handle);
    system(com);
    g_free(com);
    close_current_file();
    if (old_file>0&&mainw->current_file!=old_file) {
      switch_to_file (mainw->current_file,old_file);
    }
    return;
  }
  cfile->opening=cfile->opening_audio=FALSE;

  // mainw->error is TRUE if we could not load
  if (mainw->error==TRUE) {
    do_error_dialog(mainw->msg);
    d_print(_ ("failed.\n"));
    close_current_file();
    if (old_file>0&&mainw->current_file!=old_file) {
      switch_to_file (mainw->current_file,old_file);
    }
    return;
  }

  com=g_strdup_printf("smogrify dvcomplete %s",cfile->handle);
  unlink (cfile->info_file);
  system(com);
  g_free(com);
  
  do_progress_dialog(TRUE,TRUE,_ ("Checking frames"));

  cfile->changed=TRUE;

  // now file should be loaded...get full details
  cfile->is_loaded=TRUE;
  add_file_info(cfile->handle);

  if (cfile->frames<=0&&!cfile->afilesize) {
      // we got neither video nor audio...
      do_error_dialog(_ ("\n\nLiVES was unable to extract either video or audio from this file.\n"));
      d_print(_ ("failed.\n"));
      close_current_file();
      if (old_file>0&&mainw->current_file!=old_file) {
	switch_to_file (mainw->current_file,old_file);
      }
      return;
    }

  // update widgets
  switch_to_file((mainw->current_file=old_file),current_file);
  gtk_widget_set_sensitive(mainw->save,FALSE);
}

void
get_handle_from_info_file(void) {
  // called from get_new_handle to get the 'real' file handle
  // because until we know the handle we can't use the normal info file yet
  FILE *infofile;

  clear_mainw_msg();

  while (!(infofile=fopen(mainw->first_info_file,"r"))) {
    g_usleep(prefs->sleep_time);
  }

  fgets(mainw->msg,512,infofile);
  fclose(infofile);
  unlink(mainw->first_info_file);

  cfile=(file *)(g_malloc(sizeof(file)));
  g_snprintf(cfile->handle,256,mainw->msg);
  cfile->clip_type=CLIP_TYPE_DISK; // the default
}



void 
save_file (gboolean existing, gchar *n_file_name) {
  // if existing is TRUE, we are saving under the existing file_name
  gint arate=cfile->arate;
  gchar *mesg,*bit;
  gchar *com;
  gchar *full_file_name=NULL;
  gboolean save_perf=FALSE;
  gint asigned=!(cfile->signed_endian&AFORM_UNSIGNED);
  gboolean not_cancelled;

  gint current_file=mainw->current_file;
  gdouble aud_start=0,aud_end=0;
  gchar *fps_string;

  if (cfile->arate*cfile->achans) {
    aud_start=calc_variable_time (cfile->start)*cfile->arps/cfile->arate;
    aud_end=calc_variable_time (cfile->end+1)*cfile->arps/cfile->arate;
  }

  // new handling for save selection:
  // open a new clip, symlink images 1 - n to the encoded frames
  // encode the new clip
  // close the new clip


  if (cfile->events[EVENT_FRAMES]!=NULL) {
    do_error_dialog (_ ("Sorry, cannot save with reordered frames.\nPlease reorder frames from the Tools menu, \nor clear the recording by pressing 'r' twice.\n"));
    return;
  }

  // get file extension
  check_encoder_restrictions (TRUE);

  //append default extension (if necessary)
  if (!strlen (prefs->encoder.of_def_ext)) {
    // encoder adds its own extension
    if (rindex(n_file_name,'.')!=NULL) {
      memset (rindex (n_file_name,'.'),0,1);
    }
  }
  else {
    if (rindex(n_file_name,'.')==NULL) {
      full_file_name=g_strconcat(n_file_name,".",prefs->encoder.of_def_ext,NULL);
    }
  }

  if (full_file_name==NULL) {
    full_file_name=g_strdup (n_file_name);
  }

  if (existing==FALSE) {
    if (check_file(full_file_name)==FALSE) return;
    cfile->orig_file_name=FALSE;
    if (!strlen (cfile->comment)) {
      g_snprintf (cfile->comment,251,"Created with LiVES");
    }
    if (do_comments_dialog()==FALSE) {
	g_free(full_file_name);
	return;
    }
  }
  else if (cfile->orig_file_name==TRUE) {
    gchar *warn=g_strdup(_ ("Saving your video could lead to a loss of quality !\nYou are strongly advised to 'Save As' to a new file.\n\nDo you still wish to continue ?"));
    if (!do_warning_dialog_with_check(warn,WARN_MASK_SAVE_QUALITY)) {
	g_free(warn);
	g_free(full_file_name);
	return;
    }
    g_free(warn);
  }
  

  if (!mainw->save_all) {
    // we are saving a selection
    gint new_file;

    if ((new_file=mainw->first_free_file)==-1) {
      too_many_files();
      return;
    }

    // create new clip
    if (!get_new_handle(new_file,g_strdup_printf (_ ("selection")))) {
      return;
    }

    mainw->current_file=new_file;
    cfile->hsize=mainw->files[current_file]->hsize;
    cfile->vsize=mainw->files[current_file]->vsize;
    cfile->progress_start=cfile->start=1;
    cfile->progress_end=cfile->frames=cfile->end=mainw->files[current_file]->end-mainw->files[current_file]->start+1;
    cfile->fps=mainw->files[current_file]->fps;
    cfile->arps=mainw->files[current_file]->arps;
    cfile->arate=mainw->files[current_file]->arate;
    cfile->achans=mainw->files[current_file]->achans;
    cfile->asampsize=mainw->files[current_file]->asampsize;

    com=g_strdup_printf ("smogrify link_frames %s %s %d %d",cfile->handle,mainw->files[current_file]->handle,mainw->files[current_file]->start,mainw->files[current_file]->end);

    unlink(cfile->info_file);
    system(com);
    g_free(com);

    if (!(do_progress_dialog(TRUE,TRUE,_ ("Linking selection")))) {
      system (g_strdup_printf("smogrify close %s",cfile->handle));
      g_free (cfile);
      cfile=NULL;
      mainw->current_file=current_file;
      sensitize();
      d_print ("cancelled.\n");
      return;
    }
  }

#ifdef AUTO_RESAMPLE_RESIZE
  if (!check_encoder_restrictions(FALSE)) {
    if (!mainw->save_all) {
      system (g_strdup_printf("smogrify close %s",cfile->handle));
      g_free (cfile);
      cfile=NULL;
      switch_to_file (mainw->current_file,current_file);
    }
    d_print ("cancelled.\n");
    return;
  }
#endif

  if (mainw->save_with_sound==FALSE||prefs->encoder.of_allowed_acodecs==0) {
    bit=g_strdup(_ (" (with no sound)\n"));
    arate=0;
  }
  else {
    bit=g_strdup("\n");
  }

  if (!mainw->save_all) {
    mesg=g_strdup_printf(_ ("Saving frames %d to %d%s as \"%s\" : encoder = %s : format = %s..."),mainw->files[current_file]->start,mainw->files[current_file]->end,bit,full_file_name,prefs->encoder.name,prefs->encoder.of_name);
  } // end selection
  else {
    mesg=g_strdup_printf(_ ("Saving frames 1 to %d%s as \"%s\" : encoder %s : format = %s..."),cfile->frames,bit,full_file_name,prefs->encoder.name,prefs->encoder.of_name);
  }
  g_free (bit);
  
  d_print (mesg);
  g_free (mesg);

  if (!cfile->ratio_fps) {
    fps_string=g_strdup_printf ("%.3f",cfile->fps);
  }
  else {
    fps_string=g_strdup_printf ("%.8f",cfile->fps);
  }

  if (prefs->encoder.capabilities&ENCODER_NON_NATIVE) {
    com=g_strdup_printf("smogrify save %s \"%s%s%s/%s\" %s \"%s\" %d %d %d %d %d %d %.4f %.4f",cfile->handle,prefs->prefix_dir,PLUGIN_DIR,PLUGIN_ENCODERS,prefs->encoder.name,fps_string,full_file_name,1,cfile->frames,arate,cfile->achans,cfile->asampsize,asigned,aud_start,aud_end);
  }
  else {
    // for native plugins we go via the plugin
    com=g_strdup_printf("%s%s%s/%s save %s \"\" %s \"%s\" %d %d %d %d %d %d %.4f %.4f",prefs->prefix_dir,PLUGIN_DIR,PLUGIN_ENCODERS,prefs->encoder.name,cfile->handle,fps_string,full_file_name,1,cfile->frames,arate,cfile->achans,cfile->asampsize,asigned,aud_start,aud_end);
  }
  g_free (fps_string);

  unlink(cfile->info_file);
  save_file_comments();
  system(com);
  g_free(com);
  not_cancelled=do_progress_dialog(TRUE,TRUE,_ ("Saving"));
  mesg=g_strdup (mainw->msg);
  if (prefs->encoder.capabilities&ENCODER_NON_NATIVE) {
    com=g_strdup_printf("smogrify plugin_clear %s %d %d %s%s %s %s",cfile->handle,1,cfile->frames,prefs->prefix_dir,PLUGIN_DIR,PLUGIN_ENCODERS,prefs->encoder.name);
  }
  else {
    com=g_strdup_printf("%s%s%s/%s plugin_clear %s %d %d \"\" %s \"\"",prefs->prefix_dir,PLUGIN_DIR,PLUGIN_ENCODERS,prefs->encoder.name,cfile->handle,1,cfile->frames,PLUGIN_ENCODERS);
  }

  system(com);
  g_free(com);

  if (not_cancelled) {
    if (mainw->error) {
      if (save_perf) {
	deorder_frames(cfile->frames);
	d_print(_ ("deordered frames..."));
	restore_event_frames();
	restore_event_fps();
      }
      do_error_dialog(mesg);
      g_free (mesg);
      d_print(_ ("error.\n"));
      g_free(full_file_name);
      if (!mainw->save_all) {
	system (g_strdup_printf("smogrify close %s",cfile->handle));
	g_free (cfile);
	cfile=NULL;
	mainw->current_file=current_file;
	sensitize(); // must do this ourselves as we had no menuentry
      }
      return;
    }
    g_free (mesg);
    
    if (!g_file_test (full_file_name, G_FILE_TEST_EXISTS)) {
      do_error_dialog(_ ("\n\nEncoder error - output file was not created !\n"));
      d_print (_ ("failed.\n"));
      g_free(full_file_name);
      if (save_perf==TRUE) {
	deorder_frames(cfile->frames);
	d_print(_ ("deordered frames..."));
	restore_event_frames();
	restore_event_fps();
      }
      if (!mainw->save_all) {
	system (g_strdup_printf("smogrify close %s",cfile->handle));
	g_free (cfile);
	cfile=NULL;
	mainw->current_file=current_file;
	sensitize(); // must do this ourselves as we had no menuentry
      }
      return;
    }

    if (mainw->save_all) {
      g_snprintf(cfile->file_name,255,"%s",full_file_name);
      cfile->changed=FALSE;

      // save was successful
      cfile->f_size=sget_file_size (full_file_name);

      if (cfile->is_untitled) {
	cfile->is_untitled=FALSE;
	gtk_widget_set_sensitive(mainw->save,TRUE);
      }
      if (!cfile->was_renamed) {
	set_menu_text(cfile->menuentry,full_file_name,FALSE);
	g_snprintf(cfile->name,255,"%s",full_file_name);
      }
      set_main_title(cfile->name,0);
      add_to_recent (full_file_name,0.,0);

      if (save_perf) {
	if (not_cancelled) {
	  d_print(_ ("saved..."));
	}
	deorder_frames(cfile->frames);
	d_print(_ ("deordered frames..."));
	restore_event_frames();
	restore_event_fps();
      }
    }
    else {
      system (g_strdup_printf("smogrify close %s",cfile->handle));
      g_free (cfile);
      cfile=NULL;
      switch_to_file (mainw->current_file,current_file);
    }
  }
  g_free(full_file_name);
  if (not_cancelled) {
    d_print(_ ("done.\n"));
  }
}



void play_file (void) {
  // play the current clip from 'mainw->play_start' to 'mainw->play_end'
  gint arate;
  gchar *oinfo_file;
  gboolean osepwin=mainw->sep_win;
  gint event_overkill;

  gchar *com;
  gchar *com2=g_strdup (" ");
  gchar *com4=g_strdup (" ");
  gchar *com3=g_strdup (" ");
  gchar *stopcom;
  gchar *msg;

  unsigned int wid;
  gint asigned=!(cfile->signed_endian&AFORM_UNSIGNED);
  gint aendian=!(cfile->signed_endian&AFORM_BIG_ENDIAN);

  gboolean oloop=mainw->loop;
  gint current_file=mainw->current_file;
  gint audio_end=0;

  // flag for smogrify
  short play_vid=1;

  gint rec_frames;
  gint operf_counter=cfile->perf_counter;
  gint loop=(gint)mainw->loop;

  GClosure *freeze_closure;


  mainw->must_resize=FALSE;
  mainw->ext_playback=FALSE;

  mainw->reinit_effects=TRUE;

  mainw->last_played_file=mainw->last_played_frame=mainw->last_played_blend_file=mainw->last_played_blend_frame=-1;

  if (mainw->pre_src_file==-2) mainw->pre_src_file=mainw->current_file;

  // enable the freeze button
  gtk_accel_group_connect (GTK_ACCEL_GROUP (mainw->accel_group), GDK_BackSpace, GDK_CONTROL_MASK, 0, (freeze_closure=g_cclosure_new (G_CALLBACK (freeze_callback),NULL,NULL)));

  if (prefs->video_player==VID_PLAYER_MPLAYER&&mainw->fs==TRUE) {
    mainw->sep_win=TRUE;
  }

  // recording performance data
  if (operf_counter!=0&&(prefs->video_player==VID_PLAYER_MPLAYER||mainw->preview==TRUE)) {
    // ignore record/play performance if we are using mplayer or playing a preview
    if (prefs->video_player!=VID_PLAYER_INT&&operf_counter>0) {
      d_print(_ ("Unable to record without the internal player. Please switch to the internal player in Preferences|Playback to record.\n"));
    }
    cfile->perf_counter=0;
  }

  // set performance at right place
  if (cfile->perf_counter>0) {
    if (cfile->clip_type==CLIP_TYPE_DISK) {
      // record
      d_print(_ ("Recording performance..."));
      if (cfile->total_time==0) {
	get_total_time(cfile);
      }
      
      if (cfile->achans==0) {
	msg=g_strdup_printf (_ ("(no audio, recording length set to %.2f sec)..."),cfile->total_time+10);
	d_print(msg);
	g_free(msg);
      }
      
      if (cfile->undo_action==UNDO_RESAMPLE||cfile->undo_action==UNDO_REORDER) {
	gtk_widget_set_sensitive (mainw->undo,FALSE);
	gtk_widget_set_sensitive (mainw->redo,FALSE);
	cfile->undoable=cfile->redoable=FALSE;
      }
      
      // create event arrays for 'frame number' and 'fps'
      // we currently create 10 events per frame - TODO - use a buffer of
      // predetermined  length for this, make it initially owned by mainw,
      // that way we can record clip switches
      // Then after quantising, we should either: 
      // copy to cfile->events or render to a new clip and copy to its cfile->events
      
      
      rec_frames=cfile->total_time+10*cfile->fps;
      if (mainw->playing_sel==TRUE&&mainw->loop==FALSE) {
	rec_frames=(cfile->end-cfile->start+10*cfile->fps);
      }
    }
    else {
      rec_frames=MAX_RECORD_FRAMES;
    }

    event_overkill=(gint)(FPS_MAX*10/(cfile->fps+1));
    if (create_event_space(rec_frames*event_overkill,CREATE_EV_FRAMES|CREATE_EV_FPS)==TRUE) {
      if (mainw->playing_sel==FALSE&&prefs->loop_recording==TRUE) {
	mainw->loop=TRUE;
      }
      cfile->perf_end=rec_frames*event_overkill;
      cfile->perf_counter=cfile->perf_start=mainw->play_start;
    }
    else {
      g_signal_handler_block(mainw->record_perf,mainw->record_perf_func);
      gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mainw->record_perf),FALSE);
      g_signal_handler_unblock(mainw->record_perf,mainw->record_perf_func);
      // not enough memory
      perf_mem_warning();
      return;
    }
  }

  // note, here our start is in frames, in save_file it is in seconds !
  // TODO - check if we can change it to seconds here too

  mainw->audio_start=mainw->audio_end=0;

  if (cfile->perf_counter<0) {
    // play performance data
    // TODO - check this, am not able to right now as i have a broken soundcard :-(
    cfile->perf_counter=-mainw->play_start;
     if (cfile->perf_end>cfile->frames&&!mainw->playing_sel) {
       mainw->audio_end=(calc_variable_time(cfile->perf_end)*cfile->fps+1.)*cfile->arate/cfile->arps;
     }
  }
  
  if (mainw->audio_end==0) {
    mainw->audio_start=(calc_variable_time(mainw->play_start)*cfile->fps+1.)*cfile->arate/cfile->arps;
    mainw->audio_end=(calc_variable_time(mainw->play_end)*cfile->fps+1.)*cfile->arate/cfile->arps;
  }

  find_when_to_stop();

  if (!cfile->opening_audio&&!mainw->loop) {
    // if we are opening audio or looping we just play to the end of audio,
    // otherwise...
    audio_end=mainw->audio_end;
  }

  if (prefs->stop_screensaver==TRUE) {
    g_free (com2);
    com2=g_strdup ("xset s off; xset -dpms;");
  }
  if (prefs->pause_xmms==TRUE&&cfile->achans>0&&mainw->mute==FALSE) {
    g_free (com3);
    com3=g_strdup ("xmms -u;");
  }
  if (prefs->midisynch==TRUE&&mainw->preview==FALSE) {
    g_free (com4);
    com4=g_strdup  ("midistart");
  }
  com=g_strconcat (com2,com3,com4,NULL);
  if (strlen (com)) {
    system (com);
  }
  g_free (com); g_free (com2); g_free (com3); g_free (com4);
  com4=g_strdup (" ");

  if (mainw->preview==FALSE) {
    gtk_frame_set_label(GTK_FRAME(mainw->playframe),_ ("Play"));
  }
  else {
    gtk_frame_set_label(GTK_FRAME(mainw->playframe),_ ("Preview"));
  }

  if (palette->style&STYLE_1) {
    gtk_widget_modify_fg(gtk_frame_get_label_widget(GTK_FRAME(mainw->playframe)), GTK_STATE_NORMAL, &palette->normal_fore);
  }

  // blank the background if asked to
  if ((mainw->faded||(prefs->video_player!=VID_PLAYER_MPLAYER&&mainw->fs==TRUE&&mainw->sep_win==FALSE))&&(cfile->frames>0||mainw->foreign==TRUE)) {
    fade_background();
  }

  if ((mainw->sep_win==FALSE||(mainw->faded==FALSE&&prefs->video_player!=VID_PLAYER_MPLAYER&&(prefs->sepwin_type!=1)))&&(cfile->frames>0||mainw->foreign==TRUE)) {
    // show the frame in the main window
    gtk_widget_show(mainw->playframe);
  }

  // plug the plug into the playframe socket if we need to
  if (prefs->video_player!=VID_PLAYER_MPLAYER||mainw->sep_win==FALSE) {
    add_to_playframe();
  }
  
  arate=cfile->arate;
  if (mainw->mute==TRUE&&!cfile->opening_only_audio) arate=arate?-arate:-1;

  if (mainw->sep_win==TRUE) {
    wid=0;
  }
  else {
    wid=(unsigned int)mainw->xwin;
    if (prefs->video_player==VID_PLAYER_VEEJAY) {
      vj_client_send ("WIDx");
    }
  }
  
  cfile->pb_fps=cfile->fps;
  mainw->period=1000000./cfile->pb_fps;

#ifdef USE_LIVIDO
  if (mainw->num_tr_applied>0&&!mainw->preview&&mainw->blend_file>-1) {
#else
  if (mainw->rte&EFFECT_BLEND) {
#endif
    // reset frame counter for blend_file
    mainw->files[mainw->blend_file]->frameno=0;
  }
  mainw->playing_file=mainw->current_file;

  if (mainw->preview==FALSE||cfile->opening==FALSE) {
    desensitize();
  }
  
  enable_record();

  if (mainw->double_size==TRUE) {
    gtk_widget_hide(mainw->scrolledwindow);
  }

  gtk_widget_set_sensitive (mainw->stop, TRUE);
  gtk_widget_set_sensitive (mainw->m_stopbutton, TRUE);

  gtk_widget_set_sensitive (mainw->m_playbutton, FALSE);
  gtk_widget_set_sensitive (mainw->m_playselbutton, FALSE);
  gtk_widget_set_sensitive (mainw->m_rewindbutton, FALSE);
  gtk_widget_set_sensitive (mainw->m_mutebutton, FALSE);

  gtk_widget_set_sensitive (mainw->m_loopbutton, (!cfile->achans||mainw->mute||mainw->loop_cont)&&mainw->current_file>0);
  gtk_widget_set_sensitive (mainw->loop_continue, (!cfile->achans||mainw->mute||mainw->loop_cont)&&mainw->current_file>0);

  if (prefs->video_player!=VID_PLAYER_MPLAYER) {
    if (!cfile->frames) {
      if (mainw->preview_box!=NULL&&mainw->preview_box->parent!=NULL) {
	gtk_container_remove (GTK_CONTAINER (mainw->play_window), mainw->preview_box);
      }
    }
    else {
      // TODO - handle fixed size
      if (mainw->sep_win==TRUE) {
	// create a separate window for the internal player if requested

	if (prefs->sepwin_type==0) {
	  mainw->playing_file=mainw->current_file;
	  // needed
	  block_expose();
	  make_play_window();
	  unblock_expose();
	}
	else {
	  if (mainw->preview_box!=NULL&&mainw->preview_box->parent!=NULL) {
	    gtk_container_remove (GTK_CONTAINER (mainw->play_window), mainw->preview_box);
	  }
	  resize_play_window();
	  // needed
	  block_expose();
	  mainw->noswitch=TRUE;
	  while (g_main_context_iteration (NULL,FALSE));
	  mainw->noswitch=FALSE;
	  unblock_expose();
	}
      }
      if (!(mainw->play_window==NULL)) {
	hide_cursor (mainw->play_window->window);
	gtk_widget_set_app_paintable(mainw->play_window,TRUE);
	if (prefs->video_player==VID_PLAYER_VEEJAY) {
	  vj_client_send ("RESIZE");
	}
      }
    }
    
    if (mainw->foreign==FALSE&&mainw->sep_win==FALSE) {
      hide_cursor(mainw->playarea->window);
    }
    
    // pwidth and pheight are playback width and height
    if (mainw->sep_win==FALSE) {
      do {
	mainw->pwidth=mainw->playframe->allocation.width-H_RESIZE_ADJUST;
	mainw->pheight=mainw->playframe->allocation.height-V_RESIZE_ADJUST;
	if (mainw->pwidth*mainw->pheight==0) {
	  gtk_widget_queue_draw (mainw->playframe);
	  mainw->noswitch=TRUE;
	  while (g_main_context_iteration (NULL, FALSE));
	  mainw->noswitch=FALSE;
	}
      } while (mainw->pwidth*mainw->pheight==0);
      // double size
      if (mainw->double_size==TRUE) {
	on_double_size_activate (NULL,GINT_TO_POINTER (1));
      }
    }

    if (vid_playback_plugin!=NULL&&vid_playback_plugin->fheight>-1&&vid_playback_plugin->fwidth>-1) {

      // fixed o/p size for stream
      if (!(vid_playback_plugin->fwidth*vid_playback_plugin->fheight)) {
	vid_playback_plugin->fwidth=cfile->hsize;
	vid_playback_plugin->fheight=cfile->vsize;
      }
      mainw->pwidth=vid_playback_plugin->fwidth;
      mainw->pheight=vid_playback_plugin->fheight;
    }
    
    mainw->fixed_height=mainw->eventbox->allocation.height+mainw->menubar->allocation.height;

    if (mainw->fs==TRUE&&mainw->sep_win==FALSE&&cfile->frames>0) {
      fullscreen_internal();
    }

    // need to set cfile->event_pending, otherwise we may record a false fps change
    cfile->event_pending=EVENT_AUTO;
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(mainw->spinbutton_pb_fps),cfile->fps);
    
    // show the framebar
    if (prefs->show_framecount==TRUE&&mainw->fs==FALSE&&((mainw->preview==FALSE&&(cfile->frames>0||mainw->foreign==TRUE))||cfile->opening==TRUE)) {
      gtk_widget_show(mainw->framebar);
    }
    // flag to tell smogrify not to play video
    play_vid=-1;
  }

  if (cfile->frames==0) {
    play_vid=-1;
  }
  
  if (mainw->loop_cont==TRUE) {
    // tell audio to loop forever
    loop=-1;
  }

  cfile->play_paused=FALSE;
  mainw->actual_frame=0;

  if (loop==-1) {
    g_free (com4);
    com4=g_strdup_printf ("%s/%s/.stoploop",prefs->tmpdir,cfile->handle);
    unlink (com4);
    g_free (com4);
    com4=g_strdup_printf ("touch %s/%s/.stoploop 2>/dev/null;",prefs->tmpdir,cfile->handle);
  }

  if (prefs->video_player!=VID_PLAYER_MPLAYER&&prefs->audio_player!=AUD_PLAYER_MPLAYER) {
    com2=g_strdup_printf("smogrify stop_audio %s",cfile->handle);
  }
  else {
    com2=g_strdup_printf("smogrify stop %s",cfile->handle);
  }

  stopcom=g_strconcat (com4,com2,NULL);
  g_free (com4);

  // use a separate info_file for play
  oinfo_file=g_strdup(cfile->info_file);
  g_snprintf(cfile->info_file,256,"%s.play",oinfo_file);

  // PLAY
  unlink(cfile->info_file);
  mainw->playing_file=mainw->current_file;

  if (prefs->video_player==VID_PLAYER_MPLAYER&&mainw->fs==FALSE) {
    mainw->pwidth=cfile->hsize;
    mainw->pheight=cfile->vsize;
    if (mainw->double_size==TRUE) {
      on_double_size_activate (NULL,GINT_TO_POINTER (1));
    }
  }

  if (mainw->preview==FALSE&&(cfile->clip_type==CLIP_TYPE_DISK||cfile->achans)) {
    com=g_strdup_printf("smogrify play %s %.3f %d %d %d %d %u %d %d %d %d %d %d %d",cfile->handle,cfile->fps*play_vid,mainw->audio_start,audio_end,mainw->fs,loop,wid,mainw->pwidth,mainw->pheight,arate,cfile->achans,cfile->asampsize,asigned,aendian);
  } else {
    if (cfile->opening==FALSE&&cfile->clip_type==CLIP_TYPE_DISK) {
      com=g_strdup_printf("smogrify play_preview %s %.3f %d %d %d %d %d %u %d %d %d %d %d %d",cfile->handle,cfile->fps*play_vid,mainw->audio_start,audio_end,mainw->fs,0,wid,mainw->pwidth,mainw->pheight,arate,cfile->achans,cfile->asampsize,asigned,aendian);
    }
    else {
      com=g_strdup_printf("smogrify play_opening_preview %s %.3f %d %d %d %d %d %u %d %d %d %d %d %d",cfile->handle,cfile->fps*play_vid,mainw->audio_start,audio_end,mainw->fs,0,wid,mainw->pwidth,mainw->pheight,arate,cfile->achans,cfile->asampsize,asigned,aendian);
    }
  }

  system(com);
  //play until stopped or a stream finishes

#ifdef USE_LIVIDO
  if (mainw->foreign||livido_playback_gen_start()) {
#endif
    mainw->kb_timer=gtk_timeout_add (KEY_RPT_INTERVAL,&plugin_poll_keyboard,NULL);
    do_progress_dialog(FALSE,FALSE,NULL);
    gtk_timeout_remove (mainw->kb_timer);
    mainw->playing_file=-1;
#ifdef USE_LIVIDO
  }
#endif

  // play completed
  // kill sound(if still playing)
  system(stopcom);
  g_free(com);
  g_free (stopcom);
  mainw->actual_frame=0;

  if (prefs->video_player==VID_PLAYER_VEEJAY) {
    vj_client_send ("WID0");
  }

  // PLAY FINISHED...
  if (prefs->stop_screensaver==TRUE) system("xset s on; xset +dpms");
  if (prefs->pause_xmms==TRUE&&cfile->achans>0&&mainw->mute==FALSE) system("xmms -u");
  if (prefs->midisynch==TRUE) system ("midistop");

  if (mainw->ext_playback==TRUE) {
    (*vid_playback_plugin->exit_screen)(mainw->ptr_x,mainw->ptr_y);
    mainw->ext_playback=FALSE;
    mainw->ext_keyboard=FALSE;
  }
  // we could have started by playing a generator, which could've been closed
  if (mainw->files[current_file]==NULL) current_file=mainw->current_file;

  // wait for audio_ended...
  if (!(cfile->achans==0&&prefs->video_player!=VID_PLAYER_MPLAYER)) {
    wait_for_stop(com2);
  }
  g_free(com2);
  g_snprintf(cfile->info_file,256,"%s/%s/.status",prefs->tmpdir,cfile->handle);

  if (mainw->foreign==TRUE) {
    mainw->pwidth=mainw->playframe->allocation.width-H_RESIZE_ADJUST;
    mainw->pheight=mainw->playframe->allocation.height-V_RESIZE_ADJUST;

    cfile->hsize=mainw->pwidth;
    cfile->vsize=mainw->pheight;
    return;
  }

  gtk_widget_hide(mainw->playarea);
  if (!(mainw->dbl_size==NULL)) {
    gdk_pixbuf_unref(mainw->dbl_size);
    mainw->dbl_size=NULL;
  }

  // unblank the background
  if (mainw->faded||(prefs->video_player!=VID_PLAYER_MPLAYER&&mainw->fs==TRUE)) {
    unfade_background();
  }

  // resize out of double size
  if (mainw->double_size==TRUE&&mainw->fs==FALSE) {
    resize(1);
    if (palette->style&STYLE_1) {
      gtk_widget_show(mainw->sep_image);
    }
    gtk_widget_show(mainw->scrolledwindow);
  }

  // switch out of full screen mode
  if (mainw->fs==TRUE&&prefs->video_player!=VID_PLAYER_MPLAYER) {
    gtk_widget_show(mainw->frame1);
    gtk_widget_show(mainw->frame2);
    gtk_widget_show(mainw->eventbox3);
    gtk_widget_show(mainw->eventbox4);
    gtk_widget_show(mainw->sep_image);
    gtk_frame_set_label(GTK_FRAME(mainw->playframe),_ ("Preview"));
    gtk_container_set_border_width (GTK_CONTAINER (mainw->playframe), 10);
    resize(1);
    gtk_widget_show(mainw->t_bckground);
    gtk_widget_show(mainw->t_double);
  }

  if (prefs->video_player==VID_PLAYER_MPLAYER) {
    mainw->sep_win=osepwin;
  }

  if (mainw->eventbox->allocation.height+mainw->menubar->allocation.height>mainw->scr_height-2) {
    // the screen grew too much...remaximise it
    gtk_window_unmaximize (GTK_WINDOW(mainw->LiVES));
    mainw->noswitch=TRUE;
    while (g_main_context_iteration(NULL,FALSE));
    mainw->noswitch=FALSE;
    gtk_window_maximize (GTK_WINDOW(mainw->LiVES));
  }

  gtk_widget_hide(mainw->playframe);
  gtk_widget_show(mainw->frame1);
  gtk_widget_show(mainw->frame2);
  gtk_widget_show(mainw->eventbox3);
  gtk_widget_show(mainw->eventbox4);
  disable_record();

  gtk_container_set_border_width (GTK_CONTAINER (mainw->playframe), 10);
  
  mainw->loop=oloop;

  if (mainw->preview==FALSE||cfile->opening==FALSE) {
    sensitize();
  }
  if (cfile->opening==TRUE) {
    gtk_widget_set_sensitive (mainw->mute_audio, cfile->achans>0);
    gtk_widget_set_sensitive (mainw->loop_continue, TRUE);
    gtk_widget_set_sensitive (mainw->loop_video, cfile->achans>0&&cfile->frames>0);
  }
  gtk_widget_set_sensitive (mainw->stop, FALSE);
  gtk_widget_set_sensitive (mainw->m_stopbutton, FALSE);

  // end record performance
  if (cfile->perf_counter>0) {
    // record end
    g_signal_handler_block(mainw->record_perf,mainw->record_perf_func);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mainw->record_perf),FALSE);
    g_signal_handler_unblock(mainw->record_perf,mainw->record_perf_func);

    cfile->ev_fps_end=cfile->perf_end=cfile->perf_counter;
    cfile->ev_fps_start=cfile->perf_start;
    cfile->ev_origticks=mainw->origticks;

    // in utils.c
    quantise_events(mainw->play_start);
    if (cfile->undo_action==UNDO_RESAMPLE||cfile->undo_action==UNDO_REORDER) {
      set_undoable (NULL,FALSE);
    }
    if (cfile->events[EVENT_FRAMES]==NULL) {
      do_memory_error_dialog();
      // we can fail to allocate memory in quantise events, and sensitize will already have made this TRUE
      gtk_widget_set_sensitive(mainw->reorder,FALSE);
    }
    else {
      cfile->perf_counter=-1;
    }
    cfile->ev_fps_start=1;
    cfile->ev_fps_end=cfile->perf_end-cfile->perf_start+1;
    d_print(_ ("done.\n"));
  }

  // update screen for internal players
  if (prefs->video_player!=VID_PLAYER_MPLAYER) {

    gtk_widget_hide(mainw->framebar);
    gtk_entry_set_text(GTK_ENTRY(mainw->framecounter),"");
    gtk_image_set_from_pixbuf(GTK_IMAGE(mainw->image274),NULL);

    // kill the separate play window
    if (mainw->play_window!=NULL) {
      if (prefs->sepwin_type==0) {
	kill_play_window();
      }
      else {
	// or resize it back to single size
	if (!GTK_WIDGET_VISIBLE (mainw->play_window)) {
	  block_expose();
	  mainw->noswitch=TRUE;
	  while (g_main_context_iteration (NULL,FALSE));
	  mainw->noswitch=FALSE;
	  unblock_expose();
	}
	if (mainw->current_file>-1&&cfile->is_loaded) {
	  if (mainw->preview_box==NULL) {
	    // create the preview in the sepwin
	    make_preview_box();
	  }
	  if (mainw->current_file!=current_file) {
	    // now we have to guess how to center the play window
	    mainw->opwx=mainw->opwy=-1;
	    mainw->preview_frame=0;
	  }
	  load_preview_image(FALSE);
	  if (mainw->preview_box->parent==NULL) {
	    // and add it to the play window, and force a WM redraw
	    gtk_container_add (GTK_CONTAINER (mainw->play_window), mainw->preview_box);
	  }
	}
	if (!(mainw->play_window==NULL)) {
	  mainw->playing_file=-2;
	  resize_play_window();
	  // needed - block expose events to main window when we resize sepwin
	  if (cfile->is_loaded) {
	    block_expose();
	    mainw->noswitch=TRUE;
	    while (g_main_context_iteration (NULL,FALSE));
	    gtk_widget_queue_draw (mainw->preview_box);
	    gtk_widget_queue_resize (mainw->preview_box);
	    while (g_main_context_iteration (NULL,FALSE));
	    mainw->noswitch=FALSE;
	    unblock_expose();
	  }
	  else {
	    // opening preview - put the blank frame back in...
	    g_signal_handler_unblock(mainw->play_window,mainw->pw_exp_func);
	    gtk_widget_queue_draw (mainw->play_window);
	    block_expose();
	    mainw->noswitch=TRUE;
	    while (g_main_context_iteration (NULL,FALSE));
	    mainw->noswitch=FALSE;
	    unblock_expose();
	  }
	  mainw->playing_file=-1;
	  unhide_cursor(mainw->play_window->window);
	  
	  // and now we can update the main window
	  gtk_widget_queue_draw (mainw->LiVES);
	  mainw->noswitch=TRUE;
	  while (g_main_context_iteration (NULL,FALSE));
	  mainw->noswitch=FALSE;
	  if (mainw->play_window!=NULL) {
	    gtk_window_present (GTK_WINDOW (mainw->play_window));
	  }
	}
      }
    }
    // free the last frame image
    if (!(mainw->frame_image==NULL)) {
      gdk_pixbuf_unref(mainw->frame_image);
      mainw->frame_image=NULL;
    }
  }

  invalidate_pixel_buffers(0);
    
#ifdef USE_LIVIDO
  if (mainw->blend_file!=-1&&mainw->blend_file!=mainw->current_file&&mainw->files[mainw->blend_file]!=NULL&&mainw->files[mainw->blend_file]->clip_type==CLIP_TYPE_GENERATOR) {
    livido_bg_generator_end (mainw->files[mainw->blend_file]->ext_src);
  }
#endif
  
  if (mainw->current_file!=current_file) {
#ifdef USE_LIVIDO
    if (mainw->current_file>0&&cfile->clip_type==CLIP_TYPE_GENERATOR) {
      livido_generator_end (cfile->ext_src);
    }
#endif
    switch_to_file (mainw->current_file,current_file);
  }
  else {
    get_play_times();
  }
  
  // disable the freeze key
  gtk_accel_group_disconnect (GTK_ACCEL_GROUP (mainw->accel_group), freeze_closure);
  
  gtk_widget_show(mainw->scrolledwindow);
  
  if (mainw->current_file>-1) {
    if (cfile->perf_counter==0&&cfile->perf_start>0) {
      cfile->perf_counter=operf_counter;
    }
    
    if (mainw->toy_type==TOY_RANDOM_FRAMES&&!cfile->opening) {
      load_start_image (cfile->start);
      load_end_image (cfile->end);
    }
  }
  if (prefs->show_player_stats&&prefs->video_player!=VID_PLAYER_MPLAYER) {
    if (mainw->fps_measure>0.) {
      msg=g_strdup_printf (_ ("Average FPS was %.4f\n"),mainw->fps_measure);
      d_print (msg);
      g_free (msg);
    }
  }
  reset_clip_menu();
  if (mainw->size_warn) {
    do_error_dialog (_ ("\n\nSome frames in this clip are wrongly sized.\nYou should click on Tools--->Resize All\nand resize all frames to the current size.\n"));
    mainw->size_warn=FALSE;
  }
  mainw->is_processing=mainw->preview;
  if (cfile->clip_type==CLIP_TYPE_DISK&&cfile->frames==0&&mainw->record_perf) {
    g_signal_handler_block(mainw->record_perf,mainw->record_perf_func);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mainw->record_perf),FALSE);
    g_signal_handler_unblock(mainw->record_perf,mainw->record_perf_func);
  }
  // TODO - can this be done earlier ?
  if (mainw->cancelled==5) on_quit_activate (NULL,NULL);
}
  

gboolean
get_temp_handle(gint index) {
  // we can call this to get a temp handle for returning info from the backend
  // this function is also called from get_new_handle to create a permanent handle
  // for an opened file

  // if a temp handle is required, pass in index as mainw->first_free_file, and
  // call 'smogrify close cfile->handle' on it after use, then restore mainw->current_file

  // returns FALSE if we couldn't write to tempdir

  // WARNING: this function changes mainw->current_file, unless it returns FALSE (could not create cfile)


  gchar *com;
  gint ret;

  if (index==-1) {
    too_many_files();
    return FALSE;
  }

  com=g_strdup_printf("smogrify new %d",getpid());
  ret=system(com);
  g_free(com);

  if (ret) {
    tempdir_warning();
    return FALSE;
  }

  mainw->current_file=index;

  //get handle from info file, we will also malloc a new "file" struct here
  get_handle_from_info_file();
  create_cfile();
  return TRUE;
}



void 
create_cfile(void) {
  // any cfile (clip) initialisation goes in here
  int i;

  cfile->menuentry=NULL;
  cfile->start=cfile->end=0;
  cfile->frames=0;
  g_snprintf(cfile->type,40,_ ("Unknown"));
  cfile->f_size=0;
  cfile->achans=0;
  cfile->arate=0;
  cfile->arps=0;
  cfile->afilesize=0;
  cfile->asampsize=0;
  cfile->undoable=FALSE;
  cfile->redoable=FALSE;
  cfile->changed=FALSE;
  cfile->bpp=0;
  cfile->fps=cfile->pb_fps=prefs->default_fps;
  cfile->perf_counter=cfile->perf_end=cfile->perf_start=0;
  cfile->ev_fps_start=cfile->ev_fps_end=0;
  cfile->ev_origticks=mainw->origticks;
  for (i=0;i<=LAST_EVENT;i++) {
    cfile->events[i]=NULL;
  }
  cfile->insert_start=cfile->insert_end=0;
  cfile->is_untitled=TRUE;
  cfile->was_renamed=FALSE;
  cfile->undo_action=UNDO_NONE;
  cfile->opening_audio=cfile->opening=cfile->opening_only_audio=FALSE;
  cfile->pointer_time=0;
  cfile->restoring=cfile->opening_loc=cfile->nopreview=cfile->is_loaded=FALSE;
  cfile->video_time=cfile->total_time=cfile->laudio_time=cfile->raudio_time=0.;
  cfile->freeze_fps=0;
  cfile->event_pending=EVENT_NONE;
  cfile->frameno=cfile->last_frameno=0;
  cfile->proc_ptr=NULL;
  cfile->perf_counter=cfile->perf_start=cfile->perf_end=0;
  cfile->progress_start=cfile->progress_end=0;
  cfile->play_paused=cfile->nokeep=FALSE;
  cfile->undo_start=cfile->undo_end=0;
  cfile->rowstride=0; // unknown
  cfile->ext_src=NULL;
  cfile->clip_type=CLIP_TYPE_DISK;
  cfile->ratio_fps=FALSE;

  memset(cfile->name,0,1);
  memset(cfile->file_name,0,1);

  memset (cfile->comment,0,256);
  memset (cfile->author,0,256);
  memset (cfile->title,0,256);

  cfile->signed_endian=AFORM_UNKNOWN;
  g_snprintf(cfile->undo_text,32,_ ("_Undo"));
  g_snprintf(cfile->redo_text,32,_ ("_Redo"));

  g_snprintf(cfile->info_file,256,"%s/%s/.status",prefs->tmpdir,cfile->handle);
}


gboolean
get_new_handle (gint index, gchar *name) {
  // here is where we first initialize for the clipboard
  // and for paste_as_new, and restore
  // pass in name as NULL or "" and it will be set with an untitled number

  // this function *does not* change mainw->current_file, or add to the menu
  // or update mainw->clips_available

  gint current_file=mainw->current_file;
  if (get_temp_handle(index)==FALSE) return FALSE;

  // note : don't need to update first_free_file for the clipboard 
  if (!(index==0)) {
    get_next_free_file();
  }

  if (name==NULL||!strlen(name)) {
    cfile->is_untitled=TRUE;
    name=g_strdup_printf(_ ("Untitled%d"),mainw->untitled_number++);
  }
  g_snprintf(cfile->file_name,256,"%s",name);
  g_snprintf(cfile->name,256,"%s",name);
  mainw->current_file=current_file;
  return TRUE;
}



void add_file_info(const gchar *check_handle) {
  // file information has been retrieved, set struct cfile with details
  // contained in mainw->msg. We do this twice, once before opening the file, once again after.
  // The first time, frames and afilesize may not be correct.
  gint pieces;
  gchar *mesg;
  gchar **array=g_strsplit(mainw->msg,"|",-1);
  gchar *test_fps_string1;
  gchar *test_fps_string2;

  // sanity check handle against status file
  // (this should never happen...)

  if (strcmp(check_handle,array[1])) {
    g_printerr("Handle!=statusfile ! Bailing !\n");
    g_signal_emit_by_name(mainw->LiVES,"delete_event");
  }

  cfile->frames=atoi(array[2]);
  g_snprintf(cfile->type,40,"%s",array[3]);
  cfile->hsize=atoi(array[4]);
  cfile->vsize=atoi(array[5]);
  cfile->bpp=atoi(array[6]);
  cfile->pb_fps=cfile->fps=g_strtod(array[7],NULL);
  test_fps_string1=g_strdup_printf ("%.3f00000",cfile->fps);
  test_fps_string2=g_strdup_printf ("%.8f",cfile->fps);

  if (strcmp (test_fps_string1,test_fps_string2)) {
    cfile->ratio_fps=TRUE;
  }
  else {
    cfile->ratio_fps=FALSE;
  }
  g_free (test_fps_string1);
  g_free (test_fps_string2);

  cfile->f_size=g_strtod(array[8],NULL);
  cfile->arate=atoi(array[9]);
  cfile->achans=atoi(array[10]);
  cfile->asampsize=atoi(array[11]);
  cfile->signed_endian=get_signed_endian(atoi (array[12]), atoi (array[13]));
  cfile->afilesize=atoi(array[14]);

  pieces=get_token_count (mainw->msg,'|');

  if (pieces>14&&array[15]!=NULL) {
    g_snprintf (cfile->title,256,"%s",g_strchomp (g_strchug ((array[15]))));
  }
  if (pieces>15&&array[16]!=NULL) {
    g_snprintf (cfile->author,256,"%s",g_strchomp (g_strchug ((array[16]))));
  }
  if (pieces>16&&array[17]!=NULL) {
    g_snprintf (cfile->comment,256,"%s",g_strchomp (g_strchug ((array[17]))));
  }

  g_strfreev(array);

  if (mainw->save_with_sound==FALSE) {
    cfile->arate=cfile->achans=cfile->asampsize=cfile->afilesize=0;
  }

  if (cfile->frames<=0) {
    if (!cfile->afilesize&&cfile->is_loaded) {
      // we got no video or audio...
      return;
    }
    cfile->start=cfile->end=cfile->undo_start=cfile->undo_end=0;
  }
  else {
  // start with all selected
    cfile->start=1;
    cfile->end=cfile->frames;
    cfile->undo_start=cfile->start;
    cfile->undo_end=cfile->end;
  }

  cfile->orig_file_name=TRUE;
  cfile->is_untitled=FALSE;

  cfile->arps=cfile->arate;

  // some files give us silly frame rates, even single frames...
  if (cfile->fps==0.||cfile->fps==1000.||(cfile->frames<2&&cfile->is_loaded==TRUE)) {
    if (!(cfile->afilesize*cfile->asampsize*cfile->arate*cfile->achans)||cfile->frames<2) {
      if (cfile->frames!=1) {
      mesg=g_strdup_printf(_ ("\nPlayback speed not found or invalid ! Using default fps of %.3f fps. \nDefault can be set in Tools | Preferences | Misc.\n"),prefs->default_fps);
      d_print(mesg);
      g_free(mesg);
      }
      cfile->pb_fps=cfile->fps=prefs->default_fps;
    }
    else {
      cfile->laudio_time=cfile->raudio_time=cfile->afilesize/cfile->asampsize*8./cfile->arate/cfile->achans;
      cfile->pb_fps=cfile->fps=1.*(gint)(cfile->frames/cfile->laudio_time);
      if (cfile->fps>FPS_MAX||cfile->fps<1.) {
	cfile->pb_fps=cfile->fps=prefs->default_fps;
      }
      mesg=g_strdup_printf(_ ("Playback speed was adjusted to %.3f frames per second to fit audio.\n"),cfile->fps);
      d_print(mesg);
      g_free(mesg);
    }
  }
  if (cfile->opening==TRUE) return;

  if (cfile->bpp==256) {
    mesg=g_strdup_printf(_ ("Frames=%d type=%s size=%dx%d *bpp=Greyscale* fps=%.3f\nAudio: "),cfile->frames,cfile->type,cfile->hsize,cfile->vsize,cfile->fps);
  }
  else {
    cfile->bpp=24; // assume RGB24
    mesg=g_strdup_printf(_ ("Frames=%d type=%s size=%dx%d bpp=%d fps=%.3f\nAudio: "),cfile->frames,cfile->type,cfile->hsize,cfile->vsize,cfile->bpp,cfile->fps);
  }
  d_print(mesg);
  g_free(mesg);
    
  if (cfile->achans==0) {
    mesg=g_strdup (_ ("none\n"));
  }
  else {
    mesg=g_strdup_printf(_ ("%d Hz %d channel(s) %d bps\n"),cfile->arate,cfile->achans,cfile->asampsize);
  }
  d_print(mesg);
  g_free(mesg);

  // get the comments
  if (strlen (cfile->comment)) {
    mesg=g_strdup_printf(_ (" - Comment: %s\n"),cfile->comment);
    d_print(mesg);
    g_free(mesg);
  }
}


void 
save_file_comments (void) {
  // save the comments etc for smogrify
  int comment_fd;
  gchar *comment_file=g_strdup_printf ("%s/%s/.comment",prefs->tmpdir,cfile->handle);
  unlink (comment_file);
  comment_fd=creat(comment_file,S_IRUSR|S_IWUSR);
  write(comment_fd,cfile->title,strlen (cfile->title));
  write(comment_fd,"||%",3);
  write(comment_fd,cfile->author,strlen (cfile->author));
  write(comment_fd,"||%",3);
  write(comment_fd,cfile->comment,strlen (cfile->comment));
  close (comment_fd);
  g_free (comment_file);
}





void
wait_for_stop (const gchar *stop_command) {
  FILE *infofile;

# define SECOND_STOP_TIME 0.1
# define STOP_GIVE_UP_TIME 1.0

  gdouble time_waited=0.;
  gboolean sent_second_stop=FALSE;
  
  // send another stop if necessary
  mainw->noswitch=TRUE;
  while (!(infofile=fopen(cfile->info_file,"r"))) {
    while (g_main_context_iteration(NULL,FALSE));
    g_usleep(prefs->sleep_time);
    time_waited+=1000000./prefs->sleep_time;
    if (time_waited>SECOND_STOP_TIME&&sent_second_stop==FALSE) {
      system(stop_command);
      sent_second_stop=TRUE;
    }
    
    if (time_waited>STOP_GIVE_UP_TIME) {
      // give up waiting, but send a last try...
      system(stop_command);
      break;
    }
  }
  mainw->noswitch=FALSE;
  if (infofile) fclose (infofile);
}


gboolean
save_frame(gint frame, const gchar *file_name) {
  gint result;
  gchar *com;
  gchar full_file_name[256];

  if (rindex(file_name,'.')==NULL) {
    g_snprintf(full_file_name,255,"%s.%s",file_name,prefs->image_ext);
   }
  else {
    g_snprintf(full_file_name,255,"%s",file_name);
  }

  if (check_file(full_file_name)==FALSE) return FALSE;

  com=g_strdup_printf(_ ("Saving frame %d as %s..."),frame,full_file_name);
  d_print(com);
  g_free(com);

  com=g_strdup_printf("smogrify save_frame %s %d \"%s\"",cfile->handle,frame,g_filename_from_utf8 (full_file_name,-1,NULL,NULL,NULL));
  result=system(com);
  g_free(com);

  if (result==256) {
    d_print(_ ("failed (permission denied)\n"));
    return FALSE;
  }
  if (result==0) {
    d_print(_ ("done\n"));
    return TRUE;
  }
  // some other error condition
  return FALSE;
}


void
backup_file(const gchar *file_name) {
  gchar *com;
  gchar title[256];
  gchar **array;
  gchar full_file_name[256];
  gchar *hdrfile;
  gint withsound=1;
  gboolean with_perf=FALSE;
  gint backup_start,backup_end;

  if (mainw->save_all==FALSE&&cfile->perf_counter<0) {
    do_error_dialog(_ ("\n\nThis file has performance data recorded. You must click on Tools|Reorder Frames before you can backup a selection.\n"));
    return;
  }

  if (rindex(file_name,'.')==NULL) {
    g_snprintf(full_file_name,255,"%s.lv1",file_name);
   }
  else {
    g_snprintf(full_file_name,255,"%s",file_name);
  }

  // check if file exists
  if (check_file(full_file_name)==FALSE) return;

  if (mainw->save_all==TRUE) {
    backup_start=1;
    backup_end=cfile->frames;
    if (backup_end<backup_start) {
      backup_start=backup_end;
    }
  }
  else {
    backup_start=cfile->start;
    backup_end=cfile->end;
  }

  if (cfile->perf_counter<0) {
    with_perf=do_warning_dialog(_ ("This file has performance data recorded. Please click 'OK' to back up the performance data, 'Cancel' to back up without."));
  }

  if (cfile->undo_action==UNDO_REORDER||cfile->undo_action==UNDO_RESAMPLE) {
    // old event.frames/event.fps will be deleted/overwritten, we need to save em
    backup_event_frames();
    backup_event_fps();
  }

  // create header files
  write_headers(cfile);

  // save events
  hdrfile=g_strdup_printf("%s/%s/event.frames",prefs->tmpdir,cfile->handle);
  unlink(hdrfile);
  g_free(hdrfile);
  hdrfile=g_strdup_printf("%s/%s/event.fps",prefs->tmpdir,cfile->handle);
  unlink(hdrfile);
  g_free(hdrfile);

  // save the performance data (unless cancelled)
  if (with_perf==TRUE) {
    // we should *always* call save_event_frames, even if there are no frame events, because that is where we record perf_start
    save_event_frames();
    if (cfile->events[EVENT_FPS]!=NULL) save_event_fps();
  }

  //...and backup
  get_menu_text(cfile->menuentry,title);
  com=g_strdup_printf(_ ("Backing up %s to %s"),title,full_file_name);
  d_print(com);
  g_free(com);

  if (mainw->save_with_sound==FALSE) {
    d_print(_ (" without sound"));
    withsound=0;
  }

  d_print("...");
  cfile->progress_start=1;
  cfile->progress_end=cfile->frames;

  com=g_strdup_printf("smogrify backup %s %d %d %d \"%s\"",cfile->handle,withsound,backup_start,backup_end,g_filename_from_utf8 (full_file_name,-1,NULL,NULL,NULL));
  unlink (cfile->info_file);
  cfile->nopreview=TRUE;
  system(com);

  if ((do_progress_dialog(TRUE,TRUE,_ ("Backing up"))==FALSE)||mainw->error==TRUE) {
    if (mainw->error==TRUE) {
      d_print (_ ("failed.\n"));
    }

    // cancelled - clear up files
    cfile->nopreview=FALSE;
    g_free (com);

    // using restore details in the 'wrong' way here...it will also clear files
    com=g_strdup_printf("smogrify restore_details %s",cfile->handle);
    unlink (cfile->info_file);
    system(com);
    // auto-d
    g_free(com);
    if (cfile->undo_action==UNDO_REORDER||cfile->undo_action==UNDO_RESAMPLE) {
      // event.frames,fps was deleted, so we need to restore them
      restore_event_frames();
      restore_event_fps();
    }
   return;
  }

  cfile->nopreview=FALSE;
  g_free(com);
  if (cfile->undo_action==UNDO_REORDER||cfile->undo_action==UNDO_RESAMPLE) {
    // event.frames,fps was deleted, so we need to restore them
    restore_event_frames();
    restore_event_fps();
  }

  if (mainw->error==TRUE) {
    do_error_dialog(mainw->msg);
    d_print(_ ("failed.\n"));
    return;
  }

  if (with_perf==TRUE) {
    d_print(_ ("performance data was backed up..."));
  }

  array=g_strsplit(mainw->msg,"|",3);
  cfile->f_size=atoi(array[1]);
  g_strfreev(array);

  g_snprintf(cfile->file_name,255,"%s",full_file_name);
  if (cfile->was_renamed==FALSE) {
    g_snprintf(cfile->name,255,"%s",full_file_name);
    set_main_title(cfile->name,0);
    set_menu_text(cfile->menuentry,full_file_name,FALSE);
  }
  add_to_recent (full_file_name,0.,0);

  cfile->changed=FALSE;
  // set is_untitled to stop users from saving with a .lv1 extension
  cfile->is_untitled=TRUE;
  gtk_widget_set_sensitive(mainw->save,FALSE);
  d_print(_ ("done.\n"));
}


void 
write_headers (file *file) {
  int header_fd;
  gchar *hdrfile;

  // save the file details
  hdrfile=g_strdup_printf("%s/%s/header",prefs->tmpdir,file->handle);
  header_fd=creat(hdrfile,S_IRUSR|S_IWUSR);
  write(header_fd,file,sizhead);
  write(header_fd,LiVES_VERSION,strlen(LiVES_VERSION));
  close(header_fd);
  g_free(hdrfile);

  // more file details (since version 0.7.5)
  hdrfile=g_strdup_printf("%s/%s/header2",prefs->tmpdir,file->handle);
  header_fd=creat(hdrfile,S_IRUSR|S_IWUSR);
  write(header_fd,&file->frames,sizint);
  write(header_fd,&file->title,256);
  write(header_fd,&file->author,256);
  write(header_fd,&file->comment,256);
  close(header_fd);
  g_free(hdrfile);
}


gboolean 
read_headers(const gchar *file_name) {
  // file_name is only used to get the file size on the disk
  FILE *infofile;
  gchar buff[1024];
  gint pieces;
  gchar **array;
  gchar *com;
  int header_fd;
  gchar *hdrfile;
  gchar version[32];
  gint header_size;
  gint version_hash;

  memset (version,0,32);
  memset (buff,0,1024);

  hdrfile=g_strdup_printf("%s/%s/header",prefs->tmpdir,cfile->handle);
  header_fd=open(hdrfile,O_RDONLY);
  header_size=get_file_size(header_fd);

  if (header_size<sizhead) {
    return FALSE;
  }
  else {
    read(header_fd,cfile,sizhead);
    if (header_size>sizhead) {
      if (header_size-sizhead>31) {
	read(header_fd,&version,31);
	version[31]='\0';
      }
      else {
	read(header_fd,&version,header_size-sizhead);
	version[header_size-sizhead]='\0';
      }
    }
  }
  close(header_fd);
  g_free(hdrfile);

  // handle version changes
  version_hash=verhash(version);
  if (version_hash<7001) {
    cfile->arps=cfile->arate;
    cfile->signed_endian=mainw->endian;
  }

  com=g_strdup_printf("smogrify restore_details %s \"%s\" %d",cfile->handle,g_filename_from_utf8 (file_name,-1,NULL,NULL,NULL),!strcmp (file_name,"."));
  system(com);
  g_free(com);
  
  while (!(infofile=fopen(cfile->info_file,"r"))) {
    g_usleep(prefs->sleep_time);
  }
  
  fgets(buff,1024,infofile);
  fclose(infofile);

  pieces=get_token_count (buff,'|');
  array=g_strsplit(buff,"|",pieces);
  cfile->f_size=g_strtod(array[1],NULL);
  cfile->afilesize=atoi(array[2]);
  cfile->frames=atoi(array[3]);
  
  if (pieces>4&&array[4]!=NULL) {
    g_snprintf (cfile->title,256,"%s",g_strchomp (g_strchug ((array[4]))));
  }
  if (pieces>5&&array[5]!=NULL) {
    g_snprintf (cfile->author,256,"%s",g_strchomp (g_strchug ((array[5]))));
  }
  if (pieces>6&&array[6]!=NULL) {
    g_snprintf (cfile->comment,256,"%s",g_strchomp (g_strchug ((array[6]))));
  }
  
  g_strfreev(array);
  
  if (cfile->perf_start>cfile->frames) {
    // perf_end may be > cfile->frames, perf_start may not
    return FALSE;
  }
  return TRUE;
}


void 
open_set_file (gchar *set_name, gint clipnum) {
  size_t nlen;
  int set_fd;
  int pb_fps;
  gchar name[256];

  gchar *setfile=g_strdup_printf("%s/%s/set.%s",prefs->tmpdir,cfile->handle,set_name);

  memset (name,0,256);

  if ((set_fd=open(setfile,O_RDONLY))>0) {
    // get perf_start
    if ((nlen=read(set_fd,&pb_fps,sizint))) {
      cfile->pb_fps=pb_fps/1000.;
      read(set_fd,&cfile->frameno,sizint);
      nlen=read(set_fd,name,256);
    }
    if (nlen==0) {
      g_snprintf (name,256,"set_clip %.3d",clipnum);
    }
    if (strcmp (name+strlen (name)-1,")")) {
      g_snprintf (cfile->name,256,"%s (%s)",name,set_name);
    }
    else {
      g_snprintf (cfile->name,256,"%s",name);
    }
    g_snprintf(cfile->file_name,256,"%s",cfile->name);
    close (set_fd);
  }
  g_free (setfile);
}



void
restore_file(const gchar *file_name) {
  gchar *com=g_strdup("dummy");
  gchar *mesg;
  gboolean is_OK=TRUE;
  gchar *fname=g_strdup(file_name);

  gint old_file=mainw->current_file,current_file;
  gint new_file=mainw->first_free_file;
  gboolean not_cancelled;

  new_file=mainw->first_free_file;

  // create a new file
  if (get_new_handle(new_file,fname)==FALSE) {
    return;
  }
  
  mesg=g_strdup_printf(_ ("Restoring %s..."),file_name);
  d_print(mesg);
  g_free(mesg);
  
  mainw->current_file=new_file;
  
  cfile->hsize=prefs->default_width;
  cfile->vsize=prefs->default_height;
  
  switch_to_file((mainw->current_file=old_file),new_file);
  set_main_title(cfile->file_name,0);
  
  com=g_strdup_printf("smogrify restore %s \"%s\"",cfile->handle,file_name);
  system(com);
  g_free(com);
  unlink (cfile->info_file);
  
  cfile->restoring=TRUE;
  not_cancelled=do_progress_dialog(TRUE,TRUE,_ ("Restoring"));
  cfile->restoring=FALSE;

  if (mainw->error==TRUE||not_cancelled==FALSE) {
    if (mainw->error==TRUE) {
      do_blocking_error_dialog (mainw->msg);
    }
    close_current_file();
    if (old_file>0&&mainw->current_file!=old_file) {
      switch_to_file (mainw->current_file,old_file);
    }
    return;
  }
  
  // EVENT_FRAMES
  if (load_event_frames()==TRUE) {
    // EVENT_FPS
    if (cfile->perf_start>0) {
      is_OK=load_event_fps();
    }
  }
  else {
    cfile->perf_end=cfile->perf_start=cfile->perf_counter=0;
  }

  if (is_OK==TRUE) {
    // call function to return rest of file details
    //fsize, afilesize and frames
    is_OK=read_headers(file_name);
  }


  if (is_OK==FALSE) {
    mesg=g_strdup_printf(_ ("\n\nThe file %s is corrupt.\nLiVES was unable to restore it.\n"),file_name);
    do_blocking_error_dialog(mesg);
    g_free(mesg);
    
    d_print(_ ("failed\n"));
    close_current_file();
    if (old_file>0&&mainw->current_file!=old_file) {
      switch_to_file (mainw->current_file,old_file);
    }
    return;
  }
  
  // add entry to window menu
  // TODO - do this earlier and allow switching during restore
  add_to_winmenu();

  if (prefs->show_recent==TRUE) {
    add_to_recent(file_name,0.,0);
  }

  if (cfile->frames>0) {
    cfile->start=1;
  }
  else {
    cfile->start=0;
  }
  cfile->end=cfile->frames;
  cfile->arps=cfile->arate;
  cfile->pb_fps=cfile->fps;
  cfile->opening=FALSE;
  cfile->proc_ptr=NULL;

  cfile->changed=FALSE;

  g_snprintf(cfile->type,40,"Frames");
  mesg=g_strdup_printf(_ ("Frames=%d type=%s size=%dx%d bpp=%d fps=%.3f\nAudio: "),cfile->frames,cfile->type,cfile->hsize,cfile->vsize,cfile->bpp,cfile->fps);
  d_print(mesg);
  g_free(mesg);

  if (cfile->afilesize==0) {
    cfile->achans=0;
    mesg=g_strdup (_ ("none\n"));
  }
  else {
    mesg=g_strdup_printf(_ ("%d Hz %d channel(s) %d bps\n"),cfile->arate,cfile->achans,cfile->asampsize);
  }
  d_print(mesg);
  g_free(mesg);

  cfile->is_loaded=TRUE;
  current_file=mainw->current_file;
  switch_to_file((mainw->current_file=old_file),mainw->current_file);
  if (cfile->perf_start>0) {
    mesg=g_strdup (_ ("Performance data was recovered.\n"));
    d_print(mesg);
    g_free(mesg);
    cfile->perf_counter=-1;
    if (prefs->video_player==VID_PLAYER_MPLAYER) {
      do_error_dialog(_ ("WARNING: This file has performance data associated with it.\nYou will not be able to see this performance data if you use mplayer to play back.\n\nYou can switch to the internal player using the Playback tab in Preferences."));
    }
  }
}



void
save_event_frames(void) {
    int header_fd,i;
    gchar *hdrfile=g_strdup_printf("%s/%s/event.frames",prefs->tmpdir,cfile->handle);

    if (!cfile->perf_start) {
      unlink (hdrfile);
      return;
    }

    header_fd=creat(hdrfile,S_IRUSR|S_IWUSR);
    write(header_fd,&cfile->perf_start,sizint);

    if (!(cfile->events[EVENT_FRAMES]==NULL)) {
      for (i=0;i<=cfile->perf_end-cfile->perf_start;i++) {
	write(header_fd,&((cfile->events[EVENT_FRAMES]+i)->value),sizint);
      }
    }

    close(header_fd);
    g_free(hdrfile);
}



gboolean
load_event_frames (void) {
  // cfile->events[EVENT_FRAMES] should be freed before calling

  int i;
  int header_fd;
  gint perf_length;

  gchar *hdrfile=g_strdup_printf("%s/%s/event.frames",prefs->tmpdir,cfile->handle);

  if ((header_fd=open(hdrfile,O_RDONLY))>0) {
    // get perf_start
    read(header_fd,&cfile->perf_start,sizint);
    if (cfile->perf_start>0) {
       // get length of event_frames
      if ((perf_length=(get_file_size(header_fd)-sizint)/sizint)>0) {
	// allocate new space
	if (!create_event_space (perf_length,CREATE_EV_FRAMES)) {
	  do_error_dialog(_ ("\n\nLiVES was unable to allocate enough memory for the performance data.\nPerformance data will not be restored.\n"));
	  g_free(hdrfile);
	  return FALSE;
	}
	memset(cfile->events[EVENT_FRAMES],0,perf_length*sizeof(event));
	cfile->perf_end=perf_length+cfile->perf_start-1;
	for (i=0;i<perf_length;i++) {
	  read(header_fd,&((cfile->events[EVENT_FRAMES]+i)->value),sizint);
	}
      }
      else {
	// perf_start only
	if (cfile->events[EVENT_FRAMES]!=NULL) {
	  g_free (cfile->events[EVENT_FRAMES]);
	  cfile->events[EVENT_FRAMES]=NULL;
	}
      }
    }
    close(header_fd);
  }
  else {
    if (cfile->events[EVENT_FRAMES]!=NULL) {
      g_free (cfile->events[EVENT_FRAMES]);
      cfile->events[EVENT_FRAMES]=NULL;
    }
  }
  g_free(hdrfile);

  return TRUE;
}



void
save_event_fps(void) {
    int i;
    int header_fd;
    gchar *hdrfile=g_strdup_printf("%s/%s/event.fps",prefs->tmpdir,cfile->handle);

    if (cfile->events[EVENT_FPS]==NULL) {
      unlink (hdrfile);
      return;
    }

    header_fd=creat(hdrfile,S_IRUSR|S_IWUSR);
    for (i=0;i<=cfile->perf_end-cfile->perf_start;i++) {
      write(header_fd,&((cfile->events[EVENT_FPS]+i)->value),sizint);
    }
    close(header_fd);
    g_free(hdrfile);
}





gboolean
load_event_fps (void) {
  // cfile->events[EVENT_FPS] should be freed before calling
  int i;
  int header_fd;
  gint perf_length;
  gchar *hdrfile=g_strdup_printf("%s/%s/event.fps",prefs->tmpdir,cfile->handle);

  if ((header_fd=open(hdrfile,O_RDONLY))>0) {
    perf_length=get_file_size(header_fd)/sizint;
    if (cfile->perf_end>0) {
      if (cfile->perf_end!=(perf_length+cfile->perf_start-1)) {
	// uh-oh, the lengths of our event_frames and event_fps should match if both exist
	close(header_fd);
	g_free(hdrfile);
	cfile->perf_end=perf_length+cfile->perf_start-1;
	return TRUE;
      }
    }
    else {
      // there were no event_frames
      cfile->perf_end=perf_length+cfile->perf_start-1;
    }

    if (!create_event_space (perf_length,CREATE_EV_FPS)) {
      do_error_dialog(_ ("\n\nLiVES was unable to allocate enough memory for the performance data.\nPerformance data will not be restored.\n"));
      cfile->perf_end=cfile->perf_start=cfile->perf_counter=0;
      if (!(cfile->events[EVENT_FRAMES]==NULL)) {
	g_free(cfile->events[EVENT_FRAMES]);
	cfile->events[EVENT_FRAMES]=NULL;
      }
      // we return TRUE here, so the caller should check cfile->perf_end or cfile->perf_start
      g_free(hdrfile);
      return TRUE;
    }
    memset(cfile->events[EVENT_FPS],0,perf_length*sizeof(event));
    for (i=0;i<perf_length;i++) {
      read(header_fd,&((cfile->events[EVENT_FPS]+i)->value),sizint);
    }
    close(header_fd);
  }
  else {
    if (cfile->events[EVENT_FPS]!=NULL) {
      g_free (cfile->events[EVENT_FPS]);
      cfile->events[EVENT_FPS]=NULL;
    }
  }
  g_free(hdrfile);
  return TRUE;
}



void 
backup_event_frames(void) {
  gchar *fromfile=g_strdup_printf("%s/%s/event.frames",prefs->tmpdir,cfile->handle);
  gchar *tofile=g_strdup_printf("%s/%s/backup.frames",prefs->tmpdir,cfile->handle);
  rename(fromfile,tofile);
  g_free(fromfile);
  g_free(tofile);
}


void 
restore_event_frames(void) {
  gchar *tofile=g_strdup_printf("%s/%s/event.frames",prefs->tmpdir,cfile->handle);
  gchar *fromfile=g_strdup_printf("%s/%s/backup.frames",prefs->tmpdir,cfile->handle);
  rename(fromfile,tofile);
  g_free(fromfile);
  g_free(tofile);
}

void 
backup_event_fps(void) {
  gchar *fromfile=g_strdup_printf("%s/%s/event.fps",prefs->tmpdir,cfile->handle);
  gchar *tofile=g_strdup_printf("%s/%s/backup.fps",prefs->tmpdir,cfile->handle);
  rename(fromfile,tofile);
  g_free(fromfile);
  g_free(tofile);
}


void 
restore_event_fps(void) {
  gchar *tofile=g_strdup_printf("%s/%s/event.fps",prefs->tmpdir,cfile->handle);
  gchar *fromfile=g_strdup_printf("%s/%s/backup.fps",prefs->tmpdir,cfile->handle);
  rename(fromfile,tofile);
  g_free(fromfile);
  g_free(tofile);
}

gint
reorder_frames(void) {
  int new_frames=cfile->old_frames;
  gchar **array;
  gchar *com;

  if (prefs->reorder_width*prefs->reorder_height==0) com=g_strdup_printf("smogrify reorder %s %d 0 0 %d",cfile->handle,!mainw->endian,prefs->reorder_leave_back);
  else com=g_strdup_printf("smogrify reorder %s %d %d %d 0",cfile->handle,!mainw->endian,prefs->reorder_width,prefs->reorder_height);

  save_event_frames();

  unlink(cfile->info_file);
  // TODO - maybe use auto_dialog
  system(com);
  if (cfile->undo_action==UNDO_RESAMPLE) {
    if (mainw->current_file>0) {
      do_progress_dialog(TRUE,FALSE,_ ("Resampling video"));
    }
    else {
      do_progress_dialog(TRUE,FALSE,_ ("Resampling clipboard video"));
    }
  }
  else {
    do_progress_dialog(TRUE,FALSE,_ ("Reordering frames"));
  }
  g_free(com);
  
  if (mainw->error==TRUE) {
    do_error_dialog (_ ("\n\nLiVES was unable to reorder the frames."));
    deorder_frames(new_frames);
    new_frames=-new_frames;
  }
  else {
    array=g_strsplit(mainw->msg,"|",2);
  
    new_frames=atoi(array[1]);
    g_strfreev(array);
    
    if (cfile->frames>new_frames) {
      new_frames=cfile->frames;
    }
  }
  return new_frames;
}



gint
deorder_frames(gint old_frames) {
  gchar *com=g_strdup_printf("smogrify deorder %s %d %d %d",cfile->handle,cfile->perf_start,cfile->perf_end,old_frames);

  unlink(cfile->info_file);
  system(com);
  do_progress_dialog(TRUE,FALSE,_ ("Deordering frames"));
  g_free(com);

  return old_frames;
}


